
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const Connection_cxx_Version =
    "$Id: Connection.cxx,v 1.29 2001/01/23 22:22:34 cktam Exp $";
#include "MgAgent.hxx"
#include "Hardware.h"
#include "Gateway.h"
#include "Connection.h"
#include "rtpTypes.h"

extern Gateway* theGateway;

// strings used for debug purposes to display state info
char* ConnectionStateString[ConnectionOpen + 1] =
    {
        "Closed",
        "Half Open",
        "Open"
    };


//  Used by the Session Descriptor 
const char* SESSION_NAME = "Vovida Gateway";


Connection::Connection (MgAgent* agent,
                        Fifo < HardwareMsg > * outputQ,
                        string endpointId)
        : myLocalSD(),
        myRemoteSD()
{
    myAgent = agent;
    myHardwareQ = outputQ;
    myState = ConnectionClosed;
    myEndpointId = endpointId;

    // set persitent fields of local SD, non persistent fields are set
    // by method setMyLocalSD() for each session
    myLocalSD.setSessionName(SESSION_NAME);
    myLocalSD.setUserName("-");
    myLocalSD.setVersion(0);
    myLocalSD.setAddress(theGateway->getDomainName());

    SdpConnection* sessionConnection = new SdpConnection;
    sessionConnection->setUnicast (myLocalSD.getAddress().c_str());
    cout << "-------------------------------------------------------" << endl;
    cout << myLocalSD.getAddress() << endl;
    cout << "-------------------------------------------------------" << endl;
    myLocalSD.setConnection (*sessionConnection);

    SdpMedia* media = new SdpMedia ();
    media->setMediaType (MediaTypeAudio);
    media->setTransportType (TransportTypeRTP);

    // codec type:      rtpPayloadG729=18,
    //                  rtpPayloadL16_mono=11,
    //                  rtpPayloadPCMU=0,

    media->addFormat(rtpPayloadG729);

    myLocalSD.addMedia (media);
} 


Connection::~Connection (void)
{
}



MgcpResponseCode
Connection::open (CRCX crcx)
{
    MgcpResponseCode responseCode = TransactionExecuted;

    switch (myState)
    {

        case ConnectionClosed:
        {
            // allocate new connection info
            myId = theGateway->newConnectionId();

            setMyLocalSD();

            // store mandatory connection parameters
            myCallId = crcx->getCallId();
            myMode = crcx->getConnectionMode();

            // get remote SD
            myRemoteSD = crcx->getRemoteConnectionDescriptor();

            // request audio session
            requestAudioSession( crcx->getLocalConnectionOptions() );

            // set response parameters and send reply
            crcx->setConnectionId(myId);
            crcx->setLocalConnectionDescriptor(myLocalSD);
        }
        break;

        case ConnectionHalfOpen:
        case ConnectionOpen:
        {
            cpLog(LOG_ERR, "Connection already in use");
            responseCode = ReasonIncorrectConnectionId;
        }

        break;

    } 

    return responseCode;

}


MgcpResponseCode
Connection::modify (MDCX mdcx)
{
    MgcpResponseCode responseCode = TransactionExecuted;

    switch (myState)
    {

        case ConnectionClosed:
        {
            cpLog(LOG_ERR, "Connection closed, cannot modify");
            responseCode = ReasonIncorrectConnectionId;
        }

        break;

        case ConnectionHalfOpen:
        case ConnectionOpen:
        {
            if (myState == ConnectionOpen)
            {
                // cannot modify audio session so must first ask hardware
                // to close existing one then request a new one
                closeAudioSession();
            }

            // complete connection using remote SD provided by agent
            myRemoteSD = mdcx->getRemoteConnectionDescriptor();

            requestAudioSession( mdcx->getLocalConnectionOptions());
        }
        break;

    } 

    return responseCode;

} 


MgcpResponseCode
Connection::close (DLCX dlcx)
{
    MgcpResponseCode responseCode = ConnectionDeleted;

    switch (myState)
    {

        case ConnectionClosed:
        {
        }
        break;

        case ConnectionHalfOpen:
        case ConnectionOpen:
        {
            closeAudioSession();
        }
        break;

    } 

    return responseCode;

} 


void
Connection::setMyLocalSD (void)
{
    myLocalSD.setSessionId(atoi(myId.c_str()));

    vector < SdpMedia* > mediaList = myLocalSD.getMediaList();
    if (mediaList.size() > 0)
    {
        mediaList[0]->setPort(theGateway->newRtpPort());
    }
    else
    {
        SdpMedia* media = new SdpMedia;
        media->setPort(theGateway->newRtpPort());
        myLocalSD.addMedia (media);
    }
} 


void
Connection::requestAudioSession (MgcpConnectionOptions options)
{
    HardwareMsg msg;

    // build msg to send to hardware to create audio session
    msg.type = HardwareAudioType;
    msg.request.type = AudioStart;
    strcpy(msg.request.mode, myMode.c_str());
    msg.endpointId = myEndpointId;

    //    msg.request.localPort = myLocalSD.getPort();
    vector < SdpMedia* > mediaList = myLocalSD.getMediaList();
    if (mediaList.size() > 0)
    {
        msg.request.localPort = mediaList[0]->getPort();
    }
    else
    {
        cpLog (LOG_ERR, "%s:%d: No Media in local SDP", __FILE__, __LINE__);
        return ;
    }

    if (options.getEchoCancellation())
    {
        cpLog (LOG_DEBUG, "echo cancellation is on\n");
        msg.request.echoCancellation = true;
    }
    else
    {
        cpLog (LOG_DEBUG, "echo cancellation is off\n");
        msg.request.echoCancellation = false;
    };

    if (myRemoteSD.isValidDescriptor())
    {
        // remote end of connection is known, create complete
        // connection by setting remote connection info
        //	msg.request.remotePort = myRemoteSD.getPort();
        vector < SdpMedia* > mediaList = myRemoteSD.getMediaList();
        if (mediaList.size() > 0)
        {
            // Get remote port.
            msg.request.remotePort = mediaList[0]->getPort();

            // Get codec type.
            vector < int > * remoteFormatList = mediaList[0]->getFormatList();
            assert(remoteFormatList);

            vector < int > ::iterator iter = remoteFormatList->begin();

            // Iterate through the list of codecs and use the last one
            // on the list.
            // 
            // NOTE:  This is not really codec negotiation.  Need to add
            // code here to do the real negotiation.

            while (iter != remoteFormatList->end() )
            {
                msg.request.codecType = *iter;
                ++ iter;
            }
        }
        else
        {
            cpLog (LOG_ERR, "%s:%d: No Media in remote SDP", __FILE__, __LINE__);
            return ;
        }
        strcpy(msg.request.remoteHost, myRemoteSD.getConnection()->getUnicast().c_str());
        cout << "===================================================" << endl;
        cout << msg.request.remoteHost << endl;
        cout << "===================================================" << endl;
        changeState(ConnectionOpen);
    }
    else
    {
        // don't know remote info yet, create half open connection
        // TBD, should mode be forced to receive only?
        msg.request.remotePort = 0;
        strcpy (msg.request.remoteHost, "");
        changeState(ConnectionHalfOpen);
        // TBD don't open session yet, need RTP API to be improved to
        // allow modify session; for now force return
        return ;
    }
    myHardwareQ->add(msg);

} 


void
Connection::closeAudioSession (void)
{
    HardwareMsg msg;

    // build msg to send to hardware to close audio session
    msg.type = HardwareAudioType;
    msg.request.type = AudioStop;
    msg.endpointId = myEndpointId;
    myHardwareQ->add(msg);
    changeState(ConnectionClosed);

} 

void
Connection::changeState (ConnectionState newState)
{
    cpLog(LOG_DEBUG, "%s: %s -> %s", myId.c_str(),
          ConnectionStateString[myState],
          ConnectionStateString[newState]);
    myState = newState;
} 
