#ifndef DIGITMAP_H_
#define DIGITMAP_H_

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const DigitMap_h_Version =
    "$Id: DigitMap.h,v 1.12 2000/12/18 23:42:25 bko Exp $";

#include <locale.h>
#include <sys/types.h>
#include <regex.h>
#include <stdio.h>
#include "Endpoint.h"

#define TRUE                1
#define FALSE               0
#define MAX_DIGIT_BUFFER    15       // Size of the storage buffer for
// collected digits.

//  The following constants are hard coded for now.  In the future,
//  there shouldn't be a limit as to how many patterns the DigitMap
//  class can store.

#define MAX_PATTERNS        2      // Will support up to this many patterns

#define MAX_PATTERN_ARRAY   50     // This is the buffer used to store the
// patterns.


// hardware events handled by state machine

enum StateEventStimulus
{
    StimulusNULL,
    StimulusNewMessage,
    StimulusResponse,
    StimulusOffHook,
    StimulusOnHook,
    StimulusDigit0,
    StimulusDigit1,
    StimulusDigit2,
    StimulusDigit3,
    StimulusDigit4,
    StimulusDigit5,
    StimulusDigit6,
    StimulusDigit7,
    StimulusDigit8,
    StimulusDigit9,
    StimulusDigitStar,
    StimulusDigitHash
};



// ********************************************************************
// class DigitMap()
// description:
// returns:
// ********************************************************************


/**  Class used for collecting digits based on a given pattern.
 */
class DigitMap
{
    private:

        /**  Uses regcomp to find a match between the input string and the pattern.
         **  Return:  status.
         */
        int matchStr (char* string,
                      char* pattern,
                      regex_t* re);

        /**  Used to count the number of digits found in pattern.
         **  This is needed to compare the number of digits that is entered and
         **  the number of digits in the pattern.
         **  Returns:  number of digits in the pattern.
         */
        int countDigits (char* pattern);

        /**  Collects and stores the digits.  Returns false to indicate it's done
         **  collecting digits.  Returns true to indicate it's still in digit
         **  collecting state.
         **  Return:  TRUE  if in digit collecting state.
         **           FALSE if finished collecting digits.
         */
        bool collect (HardwareEvent digitEvent,
                      char digitArry[],
                      bool digitState);

        /**  Returns the matching DTMF event based on the digit received.
         */
        DtmfReqEv digitToEvent (char digit);


        /**  Used to remember whether if it's in digit collecting state or not.
         */
        bool myDigitState;

        /**
         */
        bool myTimerState;


        /**  The array that is used to store the digits collected.
         **  CAUTION:  The array is finite.
         */
        char myDigitArry[MAX_DIGIT_BUFFER];

        /**  Used for storing the digitmap from the call agent.
         */
        string myDigitMap;

        /**
         */
        char* myPattern [MAX_PATTERNS];

        /**
         */
        char myPatternArray[MAX_PATTERN_ARRAY];

    public:
        /**  Creates the DigitMap object.  Initializes myDigitArry to NULL.
         */
        DigitMap();

        DigitMap(char digitArray[MAX_DIGIT_BUFFER]);

        /**  Does nothing for now.
         */
        ~DigitMap()
        {}
        ;

        /**  Used to clear the stored digits in myDigitArry.
         */
        void clearBuffer();

        /**  Sends the collected digits to the Call Agent using MgNotify.
         */
        void sendDigits(MgAgent* myAgent,
                        ReqEv* myDtmfReqEv[],
                        string myId);


        void sendTestDigits(MgAgent* myAgent,
                            ReqEv* myDtmfReqEv[],
                            string myId);

        /**  Collects the digit from the endpoint.
         **  Calls DigitMap::collect to collect the digits and compares it to the
         **  pattern.
         **
         **  If DigitMap::collect returns true, then it hasn't found a match and it
         **  is still in a digit collecting state.
         **
         **  If DigitMap::collect returns false, then it has found a match and 
         **  DigitMap::handledigits will send the digits to the Call Agent.
         **  
         **  Return:  TRUE  if in digit collecting state.
         **           FALSE if finished collecting digits.
         */
        bool handledigits(MgAgent* myAgent,
                          ReqEv* myDtmfReqEv[],
                          string myId,
                          HardwareEvent digit_event);

        /**  Sets the digitmap string.
         */
        void setDigitMap(string digitMap);


        /**  Handles the Timer event during digit collection.
         */
        void handleDigitTimer(MgAgent* myAgent,
                              ReqEv* myDtmfReqEv[],
                              string myId);

        /**  Returns the digit map.  This is used for audit endpoint.
         */
        string returnDigitMap();

};


#endif  // DIGITMAP_H_
