/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const Http_cxx_Version =
    "$Id: Http.cxx,v 1.6 2001/11/05 18:07:36 icahoon Exp $";


#include "Http.hxx"
#include "Verify.hxx"


using Vocal::Http::Type;
using Vocal::Http::Token;
using Vocal::Http::EXTENSION;
using Vocal::Http::CRLF;
using Vocal::Http::SP;
using Vocal::Http::WS;
using Vocal::Http::DOT;
using Vocal::Http::DIGIT;
using Vocal::Http::COLON;


const Type      EXTENSION = ~0;


const Token     CRLF("\r\n");
const Token     SP(" ");
const Token     WS("\t ");
const Token     DOT(".");
const Token     DIGIT("0123456789");
const Token     COLON(":");


Token
Vocal::Http::getToken(
    Token           &   current, 
    size_t          &   byteCount, 
    const Token     &   searchString, 
    bool                eolExpected = false
) 
throw ( Vocal::Http::Exception )
{
    // Alway stop at the end of the line.
    //
    Token   until(searchString);
    
    if ( &searchString != &CRLF )
    {
        until += CRLF;
        
        test_covered("0");
    }
    
    // Remove the leading white space, including CRLF.
    //
    byteCount += removeLeading(current, WS);
    
    size_t  position = current.find_first_of(until);

    // If we hit the end of the message and we expect the end of line,
    // just return the rest of the string, otherwise this is an incomplete
    // message.
    //
    if ( position == Token::npos )
    {
        if ( eolExpected )
        {
            byteCount += current.size();
            
            test_covered("1");
            
            return ( current );
        }
        
        test_covered("2");
        
        throw Vocal::Http::Exception(Exception::Incomplete);
    }

    bool  crlf_found = ( CRLF.find(current[position]) != Token::npos );
    
    // If we are looking at the end of the line, and we weren't expecting
    // it, the message is incomplete.
    //
    if ( eolExpected == false && crlf_found )
    {
        test_covered("3");
        
        throw Vocal::Http::Exception(Exception::Incomplete);
    }
    
    // We have a token, so break the token at position.
    //
    Token   token;
    
    if ( position != 0 )
    {
        token = current.substr(0, position);
    
        current.erase(0, position);

        byteCount += position;
        
        test_covered("4");
    }

    // If we hit the end of the line and we were expecting it, trim the CRLF.
    // If we didn't hit the end of the line, but were expecting it, that's ok.
    //
    if ( eolExpected == true && crlf_found )
    {
        byteCount += removeCRLF(current);
        
        test_covered("5");
    }
    
    test_covered("6");
    
    return ( token );
}


size_t
Vocal::Http::removeLeading(
    Token           &   current,
    const Token     &   searchString
)
{
    size_t  size = current.size();
    size_t  i = 0;
    bool    found = false;

    // Continue in the loop util we find a non search string character.
    //        
    for ( ; i < size; ++i )
    {
        if ( searchString.find(current[i]) == Token::npos )
        {
            test_covered("7");
            
            break;
        }
        found = true;
        
        test_covered("8");
    }
    
    if ( found )
    {
        current.erase(0, i);
        
        test_covered("9");
    }
    
    test_covered("10");

    return ( i );
}


size_t  
Vocal::Http::removeCRLF(Token & current)
{
    size_t  position = 0;

    if ( current.size() > 1 && current[0] == '\r' && current[1] == '\n' )
    {
        test_covered("11");
        
        position = 2;
    }
    else if ( current.size() > 0 && current[0] == '\n' )
    {
        test_covered("12");

        position = 1;
    }
    
    if ( position )
    {
        test_covered("13");

        current.erase(0, position);
    }

    return ( position );
}
