
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#include "LGOperator.hxx"
#include "OpRejected.hxx"
#include "OpAccepted.hxx"
#include "OpTimeOut.hxx"
#include "OpChallenged.hxx"
#include "SipProxyEvent.hxx"
#include "SipEvent.hxx"
#include "CallInfo.hxx"
#include "CallContainer.hxx"
#include "StatusMsg.hxx"
#include "AckMsg.hxx"
#include "LGTransient.hxx"
#include "LGStateMcData.hxx"
#include "symbols.hxx"
#include "SipCSeq.hxx"
#include "LGParam.hxx"

LGTransient::LGTransient()
    : opRejected(0), opAccepted(0), opTimeOut(0)
{}

LGTransient::~LGTransient()
{
    delete opRejected; delete opAccepted; delete opTimeOut;
}

bool
LGTransient::addOp(LGOperator* nextOp)
{
    if(dynamic_cast<OpRejected*>(nextOp)!= 0)
	opRejected = nextOp;
    else if(dynamic_cast<OpAccepted*>(nextOp)!= 0)
	opAccepted = nextOp;
    else if(dynamic_cast<OpTimeOut*>(nextOp)!=0)
	opTimeOut = nextOp;
    else if(dynamic_cast<OpChallenged*>(nextOp)!=0)
	opChallenged = nextOp;
    else
	assert(0);
    return true;
}

LGState*
LGTransient::process(const Sptr<SipProxyEvent>& event)
{
    assert(opRejected && opAccepted && opTimeOut && opChallenged);

    LGState *nextState = this;

    Sptr<SipEvent> sipEvent;
    if(sipEvent.dynamicCast(event)!=0)
    {
	// process sip event...

	Sptr<StatusMsg> response;
	response.dynamicCast(sipEvent->getSipMsg());
	if(!response)
	{
//	    assert(0);
	    cpLog(LOG_CRIT,"ERROR: REceived NON_Response message in Transient state!!!");
	    return nextState;
	}

	SipStatusLine statusLine = response->getStatusLine();
	if(statusLine.getStatusCode() < 200)
	{
	    // MAY BE 180, with possible ring time
	    // show, should consider call accepted!!!
	    got1xx(sipEvent);
	}
	else
	{
	    Sptr<CallInfo> callInfo = event->getCallInfo();
	    if(callInfo->getSipCommand()->getType() == SIP_INVITE)
	    {
		if(statusLine.getStatusCode() >= 300)
		{
		    sipEvent->getSipStack()->sendAsync(new AckMsg(*response),
						       LGParam::NextHopHost,
						       LGParam::NextHopPort);
		}
		else
		    sipEvent->getSipStack()->sendAsync(new AckMsg(*response));
	    }
	    cpLog(LOG_DEBUG,"gonna cancel Timer...");
	    if(callInfo->getTimer()!=0)callInfo->getTimer()->cancelTimer();
	    cpLog(LOG_DEBUG,"done cancelling timer.");

	    cpLog(LOG_INFO,"Got status code [%d]", statusLine.getStatusCode());
	    if(statusLine.getStatusCode() < 300)
	    {
		cpLog(LOG_DEBUG,"Call accepted");
		Sptr<LGStateMcData> lgData;
		if(lgData.dynamicCast(callInfo->getCallStateData())==0)
		{
		    cpLog(LOG_CRIT,"cannot access state data!!!");
		    assert(0);
		}
		lgData->response = response;
		nextState = opAccepted->process(event);
	    }
	    else if(statusLine.getStatusCode() == 407)
	    {
		if((response->getProxyAuthenticate().getAuthScheme()==
		   AUTH_DIGEST && (!response->getProxyAuthenticate().
		   getTokenValue("stale").length()||
		   response->getProxyAuthenticate().getTokenValue("stale")
				   =="true")) ||
		    response->getCSeq().getNextCSeq()<3)
		{
		    cpLog(LOG_DEBUG,"Request Challenged by Proxy!!!");
		    nextState = opChallenged->process(event);
		}
		else
		{
		    cpLog(LOG_DEBUG,"Challenge Rejected by Proxy");
		    nextState = opRejected->process(event);
		}
	    }
	    else if(statusLine.getStatusCode() == 401)
	    {
		if((response->getWwwAuthenticate().getAuthScheme()==
		     AUTH_DIGEST && (!response->getWwwAuthenticate().
		     getTokenValue("stale").length()||
		     response->getWwwAuthenticate().getTokenValue("stale")
				     =="true")) ||
		    response->getCSeq().getNextCSeq()<3)
		{
		    cpLog(LOG_DEBUG,"Request Challenged by Endpoint!!!");
		    nextState = opChallenged->process(event);
		}
		else
		{
		    cpLog(LOG_DEBUG,"Challenge Rejected by Endpoint");
		    nextState = opRejected->process(event);
		}
	    }
	    else
	    {
		cpLog(LOG_DEBUG,"Call Rejected:%s",
		      sipEvent->getSipMsg()->encode().logData());
		nextState = opRejected->process(event);
	    }
	}
    }
    else
    {
	Sptr<TimerEvent> timerEvent;
	if(timerEvent.dynamicCast(event)!=0)
	{
	    // process the timeout...
	    if (!retry(timerEvent))
	    {
		cpLog(LOG_DEBUG,"Call TimeOut: %s",
		      timerEvent->getCallInfo()->getSipCommand()->
		      computeCallLeg().encode().logData());
		nextState = opTimeOut->process(event);
	    }
	    else
		cpLog(LOG_DEBUG,"gonna retry: %s",
		      timerEvent->getCallInfo()->getSipCommand()->
		      computeCallLeg().encode().logData());

	}
	else
	{
	    cpLog(LOG_CRIT,"received invalid event to process!!!");
	    assert(0);
	}
    }
    return nextState;
}

bool LGTransient::retry(Sptr<TimerEvent>& event)
{
    Sptr<CallInfo> callInfo = event->getCallInfo();
    Sptr<LGStateMcData> lgData;
    if(lgData.dynamicCast(callInfo->getCallStateData())==0)
    {
	assert(0);
    }
    if(LGParam::MaxRetry && (lgData->got1xx++ % LGParam::MaxRetry))
    {
	event->getSipStack()->sendAsync(callInfo->getSipCommand(),
					LGParam::NextHopHost,
					LGParam::NextHopPort);
	int ms = 0;
	if(LGParam::Function == LGParam::ORIGINATING)
	    ms = LGParam::InviteTimeOut;
	else
	    ms = LGParam::RequestTimeOut;
	Sptr<TimerEvent> retEvent = new TimerEvent(event->getFifo());
	retEvent->setCallInfo(callInfo, event->getCallContainer());
	retEvent->startTimer(retEvent, ms);
	callInfo->setTimer(retEvent);
    }
    else
	return false;
    return true;
}

void LGTransient::got1xx(Sptr<SipEvent>& event)
{
    Sptr<CallInfo> callInfo = event->getCallInfo();
    Sptr<LGStateMcData> lgData;
    if(lgData.dynamicCast(callInfo->getCallStateData())==0)
    {
	assert(0);
    }
    lgData->got1xx += (lgData->got1xx ? 0:1);
}
