
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const COPSClientAcceptMsg_cxx_Version =
    "$Id: COPSClientAcceptMsg.cxx,v 1.8 2000/12/18 23:42:53 bko Exp $";
#include <assert.h>

#include "COPSPepId.hxx"
#include "COPSIpv4LastPdpAddr.hxx"
#include "COPSIpv6LastPdpAddr.hxx"
#include "COPSClientSI.hxx"
#include "COPSIntegrity.hxx"
#include "COPSClientAcceptMsg.hxx"
#include "cpLog.h"

COPSClientAcceptMsg::COPSClientAcceptMsg()
        : _kaTimer(0),
        _acctTimer(0),
        _integrity(0)
{}


COPSClientAcceptMsg::COPSClientAcceptMsg(const char* data)
        : _kaTimer(0),
        _acctTimer(0),
        _integrity(0)
{
    parse(data);
}

void
COPSClientAcceptMsg::parse(const COPSHeader& hdr, const char* data) throw (COPSBadDataException&)
{
    assert(hdr.getOpCode() == COPS_OP_CAT);
    _hdr = new COPSHeader(hdr);
    parse(data);
    setMsgLength();
}

COPSClientAcceptMsg::~COPSClientAcceptMsg()
{}


void
COPSClientAcceptMsg::parse(const char* data) throw (COPSBadDataException&)
{
    COPSMsg::parseHeader(data);

    while (_dataStart < _dataLength)
    {
        COPSObjHeader objHdr;
        objHdr.parse(data + _dataStart);
        switch (objHdr.getCNum())
        {
            case COPS_KA:
            {
                _kaTimer = new COPSKATimer(data + _dataStart);
                _dataStart += _kaTimer->getDataLength();
            }
            break;
            case COPS_ACCT_TIMER:
            {
                _acctTimer = new COPSAcctTimer(data + _dataStart);
                _dataStart += _acctTimer->getDataLength();
            }
            break;
            case COPS_MSG_INTEGRITY:
            {
                _integrity = new COPSIntegrity(data + _dataStart);
                _dataStart += _integrity->getDataLength();
            }
            break;
            default:
            {
                cpLog(LOG_ALERT, "Bad Message format, unknown object type");
                throw COPSBadDataException("Bad Message format", __FILE__, __LINE__, 0);
            }
            break;
        }
    }
    checkSanity();
}

void
COPSClientAcceptMsg::checkSanity() const throw (COPSBadDataException&)
{
    if (!_hdr || !_kaTimer )
    {
        cpLog(LOG_ALERT, "Bad Message format");
        throw COPSBadDataException("Bad message format",
                                   __FILE__, __LINE__, 0);
    }
}

void
COPSClientAcceptMsg::dump(ostream& os)const
{
    if (_hdr.getPtr()) os << *_hdr << endl;
    if (_kaTimer.getPtr()) os << *_kaTimer << endl;
    if (_acctTimer.getPtr()) os << *_acctTimer << endl;
    if (_integrity.getPtr()) os << *_integrity << endl;
}

COPSClientAcceptMsg&
COPSClientAcceptMsg::operator<<(const COPSTimer& obj)
{
    //Message integrity object should be the very last one
    //If it is already added
    assert(_integrity == 0);

    if (obj.isKATimer())
    {
        assert(!_kaTimer);
        const COPSKATimer* tPtr = dynamic_cast < const COPSKATimer* > (&obj);
        assert(tPtr);
        _kaTimer = new COPSKATimer(*tPtr);
    }
    else
    {
        assert(!_acctTimer);
        const COPSAcctTimer* tPtr = dynamic_cast < const COPSAcctTimer* > (&obj);
        assert(tPtr);
        _acctTimer = new COPSAcctTimer(*tPtr);
    }
    setMsgLength();
    return (*this);
}

COPSClientAcceptMsg&
COPSClientAcceptMsg::operator<<(const COPSIntegrity& obj)
{
    assert(!_integrity);
    assert(obj.isMessageIntegrity());
    _integrity = new COPSIntegrity(obj);
    setMsgLength();
    return (*this);
}

COPSClientAcceptMsg&
COPSClientAcceptMsg::operator<<(const COPSHeader& obj)
{
    assert(!_hdr);
    assert(obj.getOpCode() == COPS_OP_CAT);
    _hdr = new COPSHeader(obj);
    setMsgLength();
    return (*this);
}

void
COPSClientAcceptMsg::setMsgLength()
{
    u_int16_t len = 0;
    if (_kaTimer.getPtr()) len += _kaTimer->getDataLength();
    if (_acctTimer.getPtr()) len += _acctTimer->getDataLength();
    if (_integrity.getPtr()) len += _integrity->getDataLength();
    _hdr->setMsgLength(len);
}

void
COPSClientAcceptMsg::writeData(int id) const throw(VNetworkException&)
{
    checkSanity();
    if (_hdr.getPtr()) _hdr->writeData(id);
    if (_kaTimer.getPtr()) _kaTimer->writeData(id);
    if (_acctTimer.getPtr()) _acctTimer->writeData(id);
    if (_integrity.getPtr()) _integrity->writeData(id);
}
