
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const COPSConnection_cxx_Version =
    "$Id: COPSConnection.cxx,v 1.9 2001/06/29 21:02:40 bko Exp $";
//Standard Include
#include <stdio.h>
#include <strstream.h>
#include <unistd.h>


//User defined
#include "VNetworkException.hxx"
#include "VEnvVar.hxx"
#include "NetworkAddress.h"
#include "Connection.hxx"
#include "cpLog.h"

#include "COPSKAMsg.hxx"
#include "COPSConnection.hxx"
#include "COPSClientOpenMsg.hxx"
#include "COPSClientCloseMsg.hxx"
#include "COPSTransceiver.hxx"
#include "Tcp_ClientSocket.hxx"


COPSConnection::COPSConnection(const string& hostName, int port)
    : _kaTimer(10),
      _currTime(0),
      _hostName(hostName),
      _port(port),
      _connected(false),
      _retryConnect(false),
      _retryCount(0),
      _retryTimer(20)
{
    connect();
}

COPSConnection::COPSConnection(const COPSClientOpenMsg& msg, const string& hostName, int port)
    : _kaTimer(10),
      _currTime(0),
      _hostName(hostName),
      _port(port),
      _connected(false),
      _retryConnect(false),
      _retryCount(0),
      _retryTimer(20)
{
    _openMsg = new COPSClientOpenMsg(msg);
    connect();
}


void*
COPSConnection::run(void* args)
{
    COPSConnection* self = static_cast < COPSConnection* > (args);

    //Loop through for Incoming messages
    for (; ; )
    {
        self->process();
        sleep(1);
    }
    return (0);
}

void
COPSConnection::process()
{
    _currTime++;
    if(_connected)
    {
        if(_currTime > _kaTimer)
        {
            //Did not receive KeepAlive message
            cpLog(LOG_INFO, "Did not receive KeepAlive message, closing connection");
            close();
        }
        else if(_currTime == _kaTimer)
        {
            ///Send KeepAlive message
            sendKeepAlive();
        }
    }
    else
    {
        _retryCount++;
        if(_retryConnect && (_retryCount >= _retryTimer))
        {
            connect();
        }
    }
}

void
COPSConnection::sendKeepAlive()
{
    if(!_connected) return;
    _currTime = 0;
    cpLog(LOG_DEBUG, "Sending keep-alive");
    COPSHeader hdr(COPS_OP_KA);
    COPSKAMsg msg;
    msg << hdr ;
    COPSTransceiver tr;
    try
    {
        tr.sendMsg(msg, _connection.getConnId());
        Sptr<COPSMsg> kMsg = readMsg(3);
        if(kMsg.getPtr() && kMsg->getHeader().isAKeepAlive())
        {
            _currTime = 0;
        }
    }
    catch (VNetworkException& e)
    {
        cpLog(LOG_ALERT, "Failed to write to connection:%s", _connection.getDescription().c_str());
        close();
    }
}

void
COPSConnection::close()
{
     cpLog(LOG_DEBUG, "COPSConnection::closing connection (%s)",
                      _connection.getDescription().c_str());
    if(_connection.isLive())
    {
        _connection.close();
    }
    _connected = false;
    _retryConnect = true;
    _retryCount = 0;
}

void
COPSConnection::connect()
{
    cpLog(LOG_INFO, "Thread (%d) Connecting to host %s:%d", 
                     pthread_self(), _hostName.c_str(), _port);
    if(_connected )
    {
        cpLog(LOG_INFO, "Already connected, ignoring connect request.");
        return;
    }

    char buf[4256];
    if(_openMsg.getPtr() == 0)
    {
        COPSHeader hdr(COPS_OP_OPN);

        COPSPepId pepId;
#if defined(__FreeBSD__)
        sprintf(buf, "Vovida-Pep-%d", reinterpret_cast < int > (pthread_self()));
#else
        sprintf(buf, "Vovida-Pep-%lu", pthread_self());
#endif
        COPSData d(buf);
        pepId.setData(d);

        _openMsg  = new COPSClientOpenMsg;
        *_openMsg << hdr ;
        *_openMsg << pepId ;
    }

    ostrstream ostr(buf, 4256);
    ostr << *_openMsg << ends;
    cpLog(LOG_DEBUG, "%s", ostr.str());

    cpLog(LOG_DEBUG, "PolicyServer:%s, port:%d",
          _hostName.c_str(), _port);

    NetworkAddress pdpAddress (_hostName, _port);
    try
    {
        TcpClientSocket cSock(pdpAddress, false);
        cSock.connect();
        _connection  = cSock.getConn();
        _openMsg->writeData(getSockId());

        Sptr<COPSMsg> recMsg = readMsg(10);
        if(recMsg.getPtr() == 0)
        {
            cpLog(LOG_INFO, "Failed to get connect");
            close();
        }
        else
        {
            if(recMsg->getHeader().isAClientAccept())
            {
                const COPSClientAcceptMsg* cMsg =
                    dynamic_cast < const COPSClientAcceptMsg* > (recMsg.getPtr());
                const Sptr < COPSKATimer > & kt = cMsg->getKATimer();

                assert(kt.getPtr() != 0);

                _kaTimer = kt->getTimerVal();
                //Set the retry timer to the twice the value of KA timer
                _retryTimer = _kaTimer * 2;
                _currTime =  0 ;
                _connected = true;
                _retryConnect = false;
            }
            else if(recMsg->getHeader().isAClientClose())
            {
                close();
            }
            else
            {
                cpLog(LOG_ALERT, "Wrong message type in response to Client Open message");
            }
        }
    }
    catch(VException & e)
    {
        cpLog(LOG_INFO, "Failed to connect to PDP server, will try again after %d seconds..", _retryTimer);
        _retryConnect = true;
        _retryCount = 0;
    }

    if(!_aThread)
    {
        //create a separate thread Handling keep alive 
        _aThread = new VThread;
        _aThread->spawn(COPSConnection::run, this);
    }
}


Sptr<COPSMsg>
COPSConnection::readMsg(int waitTime)
{
    Sptr<COPSMsg> msg;
    COPSTransceiver tr;
    if(_connection.isLive() && _connection.isReadReady(waitTime))
    {
        msg = tr.receiveMsg(getSockId());
    }
    return msg;
}


COPSConnection::~COPSConnection()
{
    _connection.close();
    if(_aThread.getPtr() != 0)
    {
        _aThread->exit();
    }
}

void
COPSConnection::sendMsg(const COPSMsg& msg)
{
    if(_connected)
    {
        msg.writeData(getSockId());
    }
    else
    {
        cpLog(LOG_ALERT, "Not connected to PDP, ignoring the message");
    }
}

