
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const COPSReqMsg_cxx_Version =
    "$Id: COPSReqMsg.cxx,v 1.9 2000/12/18 23:42:55 bko Exp $";
#include <assert.h>

#include "COPSIpv4InInterface.hxx"
#include "COPSIpv4OutInterface.hxx"
#include "COPSIpv6InInterface.hxx"
#include "COPSIpv6OutInterface.hxx"
#include "COPSHandle.hxx"
#include "COPSContext.hxx"
#include "COPSLPDPDecision.hxx"
#include "COPSClientSI.hxx"
#include "COPSIntegrity.hxx"
#include "COPSReqMsg.hxx"
#include "cpLog.h"


COPSReqMsg::COPSReqMsg()
        : _clientHandle(0),
        _context(0),
        _inInterface(0),
        _outInterface(0),
        _integrity(0),
        _lpdpContext(0)
{}


COPSReqMsg::COPSReqMsg(const char* data)
        : _clientHandle(0),
        _context(0),
        _inInterface(0),
        _outInterface(0),
        _integrity(0),
        _lpdpContext(0)
{
    parse(data);
}

void
COPSReqMsg::parse(const COPSHeader& hdr, const char* data) throw (COPSBadDataException&)
{
    _hdr = new COPSHeader(hdr);
    parse(data);
    setMsgLength();
}

COPSReqMsg::~COPSReqMsg()
{}


void
COPSReqMsg::parse(const char* data) throw (COPSBadDataException&)
{
    COPSMsg::parseHeader(data);
    while (_dataStart < _dataLength)
    {
        COPSObjHeader objHdr;
        objHdr.parse(data + _dataStart);
        switch (objHdr.getCNum())
        {
            case COPS_HANDLE:
            {
                _clientHandle = new COPSHandle(data + _dataStart);
                _dataStart += _clientHandle->getDataLength();
            }
            break;
            case COPS_CONTEXT:
            {
                if (!_context)
                {
                    //Message context
                    _context = new COPSContext(data + _dataStart);
                    _dataStart += _context->getDataLength();
                }
                else
                {
                    //lpdp context
                    _lpdpContext = new COPSContext(data + _dataStart);
                    _dataStart += _lpdpContext->getDataLength();
                }
            }
            break;
            case COPS_ININTF:
            {
                if (objHdr.getCType() == 1)
                {
                    _inInterface = new COPSIpv4InInterface(data + _dataStart);
                }
                else
                {
                    _inInterface = new COPSIpv6InInterface(data + _dataStart);
                }
                _dataStart += _inInterface->getDataLength();
            }
            break;
            case COPS_OUTINTF:
            {
                if (objHdr.getCType() == 1)
                {
                    _outInterface = new COPSIpv4OutInterface(data + _dataStart);
                }
                else
                {
                    _outInterface = new COPSIpv6OutInterface(data + _dataStart);
                }
                _dataStart += _outInterface->getDataLength();
            }
            break;
            case COPS_LPDP_DEC:
            {
                COPSLPDPDecision lpdp(data + _dataStart);
                _dataStart += lpdp.getDataLength();
                addLocalDecision(lpdp, _lpdpContext);
            }
            break;
            case COPS_CSI:
            {
                COPSClientSI csi(data + _dataStart);
                _dataStart += csi.getDataLength();
                _clientSI.push_back(csi);
            }
            break;
            case COPS_MSG_INTEGRITY:
            {
                _integrity = new COPSIntegrity(data + _dataStart);
                _dataStart += _integrity->getDataLength();
            }
            break;
            default:
            {
                cpLog(LOG_ALERT, "Bad Message format, unknown object type");
                throw COPSBadDataException("Bad Message format", __FILE__, __LINE__, 0);
            }
            break;
        }
    }
    checkSanity();
}

void
COPSReqMsg::checkSanity() const throw (COPSBadDataException&)
{
    if (!_hdr || !_clientHandle || !_context)
    {
        cpLog(LOG_ALERT, "Bad Message format");
        throw COPSBadDataException("Bad message format",
                                   __FILE__, __LINE__, 0);
    }
}

void
COPSReqMsg::dump(ostream& os)const
{
    if (_hdr.getPtr()) os << *_hdr << endl;
    if (_clientHandle.getPtr()) os << *_clientHandle << endl;
    if (_context.getPtr()) os << *_context << endl;

    for (vector < COPSClientSI > ::const_iterator sItr = _clientSI.begin();
            sItr != _clientSI.end(); sItr++)
    {
        if (_context->getRequestType() == COPSContext::CONFIG)
        {
            (*sItr).dumpPrObject(os);
        }
        else
        {
            os << (*sItr) << endl;
        }
    }

    //Display any local decisions
    DecisionMap::const_iterator itr;
    for (itr = _decisions.begin(); itr != _decisions.end(); itr++)
    {
        Sptr < COPSContext > cntxt = (*itr).first;
        const vector < COPSLPDPDecision > & decVec = (*itr).second;
        if (cntxt.getPtr()) os << *(cntxt.getPtr()) << endl;
        for (vector < COPSLPDPDecision > ::const_iterator vItr = decVec.begin();
                vItr != decVec.end(); vItr++)
        {
            os << (*vItr) << endl;
        }
    }

    if (_integrity.getPtr()) os << *_integrity << endl;
}

COPSReqMsg&
COPSReqMsg::operator<<(const COPSInterface& obj)
{
    assert(obj.isInInterface() || obj.isOutInterface());
    //Message integrity object should be the very last one
    //If it is already added
    assert(_integrity == 0);
    if (obj.isInInterface())
    {
        assert(!_inInterface);
        if (obj.isIpv4Address())
        {
            const COPSIpv4InInterface* tPtr = dynamic_cast < const COPSIpv4InInterface* > (&obj);
            assert(tPtr);
            _inInterface = new COPSIpv4InInterface(*tPtr);
        }
        else
        {
            const COPSIpv6InInterface* tPtr = dynamic_cast < const COPSIpv6InInterface* > (&obj);
            assert(tPtr);
            _inInterface = new COPSIpv6InInterface(*tPtr);
        }
    }
    else
    {
        assert(!_outInterface);
        if (obj.isIpv4Address())
        {
            const COPSIpv4OutInterface* tPtr = dynamic_cast < const COPSIpv4OutInterface* > (&obj);
            assert(tPtr);
            _outInterface = new COPSIpv4OutInterface(*tPtr);
        }
        else
        {
            const COPSIpv6OutInterface* tPtr = dynamic_cast < const COPSIpv6OutInterface* > (&obj);
            assert(tPtr);
            _outInterface = new COPSIpv6OutInterface(*tPtr);
        }
    }
    setMsgLength();
    return (*this);
}

COPSReqMsg&
COPSReqMsg::operator<<(const COPSClientSI& obj)
{
    assert(obj.isClientSI());
    _clientSI.push_back(obj);
    setMsgLength();
    return (*this);
}

void
COPSReqMsg::addLocalDecision(const COPSLPDPDecision& obj, Sptr < COPSContext > cntxt)
{
    COPSContext* tcntxt = 0;
    if (cntxt.getPtr() != 0)
    {
        tcntxt = new COPSContext(*cntxt);
    }

    assert(obj.isLocalDecision());
    if (obj.isFlagSet())
    {
        if (_decisions[tcntxt].size())
        {
            //Only one set of decision flags is allowed
            //for each context
            cpLog(LOG_ALERT, "Bad Message format, only one set of decision flags is allowed.");
            throw COPSBadDataException("Bad message format",
                                       __FILE__, __LINE__, 0);
        }
    }
    else
    {
        if (_decisions[tcntxt].size() == 0)
        {
            //The flags decision must precede any other
            //decision message, since the decision is not
            //flags throw exception
            cpLog(LOG_ALERT, "Bad Message format, flags decision must precede any other decision object.");
            throw COPSBadDataException("Bad message format",
                                       __FILE__, __LINE__, 0);
        }
    }
    _decisions[tcntxt].push_back(obj);
    setMsgLength();
}

COPSReqMsg&
COPSReqMsg::operator<<(const COPSIntegrity& obj)
{
    assert(obj.isMessageIntegrity());
    assert(!_integrity);
    _integrity = new COPSIntegrity(obj);
    setMsgLength();
    return (*this);
}

COPSReqMsg&
COPSReqMsg::operator<<(const COPSHeader& obj)
{
    assert(!_hdr);
    assert(obj.getOpCode() == COPS_OP_REQ);
    _hdr = new COPSHeader(obj);
    setMsgLength();
    return (*this);
}

COPSReqMsg&
COPSReqMsg::operator<<(const COPSHandle& obj)
{
    assert(!_clientHandle);
    _clientHandle = new COPSHandle(obj);
    setMsgLength();
    return (*this);
}

COPSReqMsg&
COPSReqMsg::operator<<(const COPSContext& obj)
{
    assert(!_context);
    _context = new COPSContext(obj);
    setMsgLength();
    return (*this);
}

void
COPSReqMsg::setMsgLength()
{
    u_int16_t len = 0;
    if (_clientHandle.getPtr()) len += _clientHandle->getDataLength();
    if (_context.getPtr()) len += _context->getDataLength();

    for (vector < COPSClientSI > ::const_iterator sItr = _clientSI.begin();
            sItr != _clientSI.end(); sItr++)
    {
        len += (*sItr).getDataLength();
    }

    //Display any local decisions
    DecisionMap::const_iterator itr;
    for (itr = _decisions.begin(); itr != _decisions.end(); itr++)
    {
        Sptr < COPSContext > cntxt = (*itr).first;
        const vector < COPSLPDPDecision > & decVec = (*itr).second;
        if (cntxt.getPtr()) len += cntxt->getDataLength();
        for (vector < COPSLPDPDecision > ::const_iterator vItr = decVec.begin();
                vItr != decVec.end(); vItr++)
        {
            len += (*vItr).getDataLength();
        }
    }

    if (_integrity.getPtr())
    {
        len += _integrity->getDataLength();
    }
    _hdr->setMsgLength(len);
}

void
COPSReqMsg::writeData(int id) const throw(VNetworkException&)
{
    checkSanity();
    if (_hdr.getPtr()) _hdr->writeData(id);
    if (_clientHandle.getPtr()) _clientHandle->writeData(id);
    if (_context.getPtr()) _context->writeData(id);

    for (vector < COPSClientSI > ::const_iterator sItr = _clientSI.begin();
            sItr != _clientSI.end(); sItr++)
    {
        (*sItr).writeData(id);
    }

    //Display any local decisions
    DecisionMap::const_iterator itr;
    for (itr = _decisions.begin(); itr != _decisions.end(); itr++)
    {
        Sptr < COPSContext > cntxt = (*itr).first;
        const vector < COPSLPDPDecision > & decVec = (*itr).second;
        if (cntxt.getPtr()) cntxt->writeData(id);
        for (vector < COPSLPDPDecision > ::const_iterator vItr = decVec.begin();
                vItr != decVec.end(); vItr++)
        {
            (*vItr).writeData(id);
        }
    }

    if (_integrity.getPtr()) _integrity->writeData(id);
}
