
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const COPSAgent_cxx_Version =
    "$Id: COPSAgent.cxx,v 1.20.10.1 2002/05/15 01:24:05 sprajpat Exp $";


#include "VocalCommon.hxx"
#include <strstream>
#include <cstdio>
#include <netinet/in.h>
#include <arpa/inet.h>


#include "COPSReqMsg.hxx"
#include "COPSDecisionMsg.hxx"
#include "COPSReportMsg.hxx"
#include "COPSAgent.hxx"
#include "COPSIpv4Interface.hxx"
#include "COPSIpv6Interface.hxx"
#include "Cops_def.hxx"
#include "PdpAgent.hxx"
#include "cpLog.h"
#include "PolicyTransaction.hxx"
#ifdef HAS_OSP
#include "PolicyTranslator.hxx"
#include "OSPException.hxx"
#endif

#ifdef HAS_RSVPLIB
#include "rsvp.h"
#endif

static const int COPS_BUFSIZE = 8096;
static const int BUFSIZE = 256;

#ifdef HAS_OSP
void
COPSOSPAgent::processRequest(const COPSReqMsg& msg)
throw (VBadDataException&, VNetworkException&)
{
    cpLog(LOG_DEBUG, "COPSOSPAgent::processRequest");
    PolicyTransaction polT;
    PolicyTranslator translator;
    try
    {
        translator.sendReq(msg, polT);
    }
    catch (PolicyBadDataException & e)
    {
        cpLog(LOG_DEBUG, "Cops->Osp req exception: %s",
              e.getDescription().c_str());
    }
    catch (OSPException& e)
    {
        cpLog(LOG_DEBUG, "Cops->Osp req exception: %s",
              e.getDescription().c_str());
    }

    //Wait for response
    //Create the decision Object and send it to the PEP
    sendOSPDecision(polT, msg);
}

void
COPSOSPAgent::sendOSPDecision(const PolicyTransaction& polT,
                              const COPSReqMsg& msg) const
{
    Sptr < COPSDecisionMsg > decisionMsg = new COPSDecisionMsg();
    COPSHeader hdr(COPS_OP_DEC, msg.getHeader().getClientType());

    COPSHandle handle;
    handle.setId(msg.getClientHandle().getId());

    COPSDecision dec1;
    dec1.setCmdCode(COPSDecision::DEC_INSTALL);

    COPSDecision dec2(DEC_STATELESS);
    COPSData data(polT.encode());
    
    dec2.setData(data);


    COPSContext cntxt(COPSContext::OUT, M_OSP);
    *decisionMsg << hdr ;
    *decisionMsg << handle ;

    decisionMsg->addDecision(dec1, cntxt);
    decisionMsg->addDecision(dec2, cntxt);

    char oBuf[COPS_BUFSIZE];
    ostrstream ostr(oBuf, COPS_BUFSIZE-1);
    ostr << "Sending DECISION:" << *decisionMsg << endl;
    ostr << ends;
    cpLog(LOG_DEBUG, "%s", ostr.str());
    try
    {
        decisionMsg->writeData(_connId);
        PdpAgent::instance().insertDecision(handle.getId(), decisionMsg);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to send the decision, reason:%s", e.getDescription().c_str());
    }
}


void
COPSOSPAgent::processReport(const COPSReportMsg& msg)
throw (VBadDataException&, VNetworkException&)
{
    PolicyTranslator translator;
    try
    {
        const vector < COPSClientSI > & csiv = msg.getClientSI();
        assert(csiv.size());
        PolicyTransaction polT;
        string dStr(csiv[0].getData().str());
        polT.decode(dStr);
        polT.print();
        if (msg.getReport().isSuccess())
        {
            translator.sendUsage(msg, polT);
        }
        else
        {
            translator.abandonTrans(string("Unknown"), polT);
        }
        //Remove the decision
        const COPSHandle& handle = msg.getClientHandle();
        PdpAgent::instance().removeDecision(handle.getId());
    }
    catch (PolicyBadDataException & e)
    {
        cpLog(LOG_DEBUG, "Cops->Osp report exception: %s",
              e.getDescription().c_str());
    }
    catch (OSPException& e)
    {
        cpLog(LOG_DEBUG, "Cops->Osp report exception: %s",
              e.getDescription().c_str());
    }
}

#endif



void
COPSRSVPAgent::processRequest(const COPSReqMsg& msg)
throw (VBadDataException&, VNetworkException&)
{
    cpLog(LOG_ALERT, "COPSRSVPAgent::processRequest");
    const COPSHeader& header = msg.getHeader();
    const COPSContext& inContext = msg.getContext();

#ifndef HAS_RSVPLIB
	// should not be instantiated in the first place:
    cpLog(LOG_ERR, "no RSVP implementation, cannot check QoS");
    return;
#endif

    Sptr < COPSDecisionMsg > decisionMsg = new COPSDecisionMsg();

    COPSHeader hdr(COPS_OP_DEC, header.getClientType());

    COPSHandle handle;
    handle.setId(msg.getClientHandle().getId());

    COPSDecision dec1;
    if (inContext.getMessageType() == M_PATH )
    {
        cpLog(LOG_DEBUG, "Checking for QoS");
        checkForQoS(msg, dec1);
    }
    else
    {
        dec1.setCmdCode(COPSDecision::DEC_INSTALL);
    }

    COPSContext cntxt(inContext.getRequestType(), inContext.getMessageType());
    *decisionMsg << hdr ;
    *decisionMsg << handle ;

    decisionMsg->addDecision(dec1, cntxt);

    char oBuf[COPS_BUFSIZE];
    ostrstream ostr(oBuf, COPS_BUFSIZE-1);
    ostr << "Sending DECISION:" << *decisionMsg << endl;
    ostr << ends;
    cpLog(LOG_DEBUG, "%s", ostr.str());

    try
    {
        decisionMsg->writeData(_connId);
        PdpAgent::instance().insertDecision(handle.getId(), decisionMsg);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to send the decision, reason:%s", e.getDescription().c_str());
    }
}

void
COPSRSVPAgent::checkForQoS(const COPSReqMsg& msg, COPSDecision& dec1)
{
#ifndef HAS_RSVPLIB
	    // should not be instantiated in the first place:
	    cpLog(LOG_ERR, "no RSVP implementation, cannot check QoS");
		return;
#else
    char buf[BUFSIZE];
    string srcIpPort;
    string destIpPort;

    // If RSVP library is present:
    const vector < COPSClientSI > & csiVec = msg.getClientSI();
    if (csiVec.size())
    {
        const COPSClientSI& csi = csiVec[0];

        const COPSData& cData = csi.getData();
        Object_header rsvpObjHeader;
        int idx = 0;
        memcpy(&rsvpObjHeader, cData.getData(), 4);
        if ((int)rsvpObjHeader.obj_ctype == 1)
        {
            SESSION_ipv4 rsvpSession;
            memcpy(&rsvpSession, cData.getData() + 4, sizeof(SESSION_ipv4));
            string destIp = inet_ntoa(rsvpSession.sess_destaddr);
            u_int16_t destPort = ntohs(rsvpSession.sess_destport);
            cpLog(LOG_DEBUG, "RSVP Dest Host:%s, port:%d", destIp.c_str(), destPort);
            sprintf(buf, "%s_%d", destIp.c_str(), destPort);
            destIpPort = buf;
        }
        else
        {
            //TODO - IPv6 Address
        }

        //HOP
        idx += ntohs(rsvpObjHeader.obj_length);
        memcpy(&rsvpObjHeader, cData.getData() + idx, 4);

        //TIME_VALUES
        idx += ntohs(rsvpObjHeader.obj_length);
        memcpy(&rsvpObjHeader, cData.getData() + idx, 4);

        //Sender TEMPLATE same as FILTER _SPEC
        idx += ntohs(rsvpObjHeader.obj_length);
        memcpy(&rsvpObjHeader, cData.getData() + idx, 4);
        if ((int)rsvpObjHeader.obj_ctype == 1)
        {
            FILTER_SPEC_ipv4 fSpec;
            memcpy(&fSpec, cData.getData() + idx + 4, sizeof(FILTER_SPEC_ipv4));
            string sIp = inet_ntoa(fSpec.filt_ipaddr);
            u_int16_t sport = ntohs(fSpec.filt_port);
            cpLog(LOG_DEBUG, "RSVP Source Host:%s, port:%d", sIp.c_str(), sport);
            sprintf(buf, "%s_%d", sIp.c_str(), sport);
            srcIpPort = buf;
        }
        else
        {
            //TODO - IPv6 Address
        }

    }

    if ((srcIpPort.size() == 0) || (destIpPort.size() == 0))
    {
        cpLog(LOG_ALERT, "Failed to get source/dest info from COPS message");
    }

    cpLog(LOG_DEBUG, "PATH %s -> %s", srcIpPort.c_str(), destIpPort.c_str());
    if (PdpAgent::instance().isQoSAllowed(srcIpPort, destIpPort))
    {
        dec1.setCmdCode(COPSDecision::DEC_INSTALL);
    }
    else
    {
        dec1.setCmdCode(COPSDecision::DEC_REMOVE);
    }
#endif
};


void
COPSRSVPAgent::processReport(const COPSReportMsg& msg)
throw (VBadDataException&, VNetworkException&)
{}


void
COPSQoSAgent::processRequest(const COPSReqMsg& msg)
throw (VBadDataException&, VNetworkException&)
{
    char buf[BUFSIZE*2];     // 512 bytes
    const vector < COPSClientSI > & csiVec = msg.getClientSI();
    assert (csiVec.size());

    string cData;

    if (csiVec[0].getData().str() == string(ENABLE_QOS))
    {
	/** 
          First CSI == ENABLE/DISABLE
          Second CSI == callId
          Third CSI == From Ip address
          Forth CSI == To Ip address
	 */
        if (csiVec.size() != 4)
        {
            cpLog(LOG_ALERT, "Incorrect number of CSI in QoS enable");
            return;
        }

        string callId = csiVec[1].getData().str();
        string fromIp = csiVec[2].getData().str();
        string toIp = csiVec[3].getData().str();

        PdpAgent::instance().insertLocalPolicy(callId, QoSData(fromIp, toIp));
        sprintf(buf, "Enabled QoS path for callId %s [%s -> %s]",
               callId.c_str(), fromIp.c_str(), toIp.c_str());
        cData = buf;
    }
    else if (csiVec[0].getData().str() == string(DISABLE_QOS))
    {
        /* First CSI == ENABLE/DISABLE
           Second CSI == callId
	 */
        if (csiVec.size() != 2)
        {
            cpLog(LOG_ALERT, "Incorrect number of CSI in QoS enable");
            return;
        }

        string callId = csiVec[1].getData().str();

        PdpAgent::instance().removeLocalPolicy(callId);
        sprintf(buf, "Disabled QoS path for callId %s",
                callId.c_str());
        cData = buf;
    }
    else
    {
        cpLog(LOG_ALERT, "Unsupported QoS request");
        return;
    }

    COPSHeader hdr(COPS_OP_RPT, C_SIP);
    COPSHandle handle ;
    COPSData id(msg.getClientHandle().getId());
    handle.setId(id);
    COPSReportType report(COPSReportType::SUCCESS);

    COPSClientSI cSi(COPSClientSI::CSI_NAMED);
    COPSData copsData(cData);
    COPSData csiData(copsData);
    cSi.setData(csiData);

    COPSReportMsg repMsg;
    repMsg << hdr ;
    repMsg << handle ;
    repMsg << report ;
    repMsg << cSi;
    repMsg.writeData(_connId);
}

void
COPSQoSAgent::processReport(const COPSReportMsg& msg)
throw (VBadDataException&, VNetworkException&)
{}


void
COPSProvisioningAgent::processRequest(const COPSReqMsg& msg)
throw (VBadDataException&, VNetworkException&)
{
    char oBuf[COPS_BUFSIZE];
    ostrstream ostr(oBuf, COPS_BUFSIZE-1);
    ostr <<  "Received req:" << msg << endl;
    ostr << ends;
    cpLog(LOG_DEBUG, "%s", ostr.str());
}

void
COPSProvisioningAgent::processReport(const COPSReportMsg& msg)
throw (VBadDataException&, VNetworkException&)
{}


void
COPSDefaultAgent::processRequest(const COPSReqMsg& msg)
throw (VBadDataException&, VNetworkException&)
{}


void
COPSDefaultAgent::processReport(const COPSReportMsg& msg)
throw (VBadDataException&, VNetworkException&)
{}
