
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const CPLOpAckLookup_cxx_Version =
    "$Id$";
#include "CPLOpAckLookup.hxx"
#include "TimerEvent.hxx"
#include "AckMsg.hxx"
#include "FsConfigFile.hxx"
#include "RoundRobinServers.hxx"

using namespace Vocal;

///Default Constructor
CPLOpAckLookup::CPLOpAckLookup():
    myClearLocations(false)
{
#ifdef MEM_TEST
    cpLog ( LOG_DEBUG, "new %s[%x]", name(), this );
#endif
}  //c'tor


///Destructor
CPLOpAckLookup::~CPLOpAckLookup()
{
#ifdef MEM_TEST
    cpLog ( LOG_DEBUG, "~%s[%x]", name(), this );
#endif
}  //d'tor


///Append an operator to this class
bool
CPLOpAckLookup::setNextOperator( Sptr < Operator > op, int prevNode )
{
    switch ( prevNode )
    {
        case CPLNodeSuccess:
        {
            mySuccessOp = op;
            return true;
        }
        break;
        case CPLNodeFailure:
        {
            myFailureOp = op;
            return true;
        }
        break;
        case CPLNodeNotfound:
        {
            myNotFoundOp = op;
            return true;
        }
        break;
        default:
        {
            cpLog( LOG_ERR, "Unknown Node '%d.'", prevNode );
            return false;
        }
    } /*switch*/

}  //setNextOperator


///Initialize the Attributes member variables
bool
CPLOpAckLookup::setAttributes( const char* attrName, const char* attrValue)
{
    if (strcmp (attrName, "clear") == 0 )
    {
        if (strcmp( attrValue, "yes" ) == 0 )
            myClearLocations = true;
        else
            myClearLocations = false;
    }  //clear
    return true;
}  //setAttributes


///
const Sptr < State >
CPLOpAckLookup::process( Sptr < SipProxyEvent > event)
{
    int lookupResults = CPLOpAckLookup::Lookup_Undefined;

    //Determine if event is a timeout or a sip message
    Sptr < TimerEvent > timeoutEvent;
    timeoutEvent.dynamicCast( event );

    if (timeoutEvent != 0)
    {
        //The timer was triggered before Redirect Server replied.
        lookupResults = CPLOpAckLookup::Lookup_Failure;
    } /* It is a timer event */
    else
    {
        // The event is a Sip message
        Sptr < SipEvent > sipEvent;
        sipEvent.dynamicCast(event);

        if ( sipEvent != 0 )
        {
            Sptr < StatusMsg > statusMsg;
            statusMsg.dynamicCast( getSipMsg( event ) );

            //Find out which Status message was received.
            if ( statusMsg != 0 )
            {
                int statusCode = statusMsg->getStatusLine().getStatusCode();
                cpLog( LOG_DEBUG, "Received %d", statusCode );

                if(statusCode > 200)
                {
                    if( statusMsg->getCSeq().getMethod() == INVITE_METHOD )
                    {
                       //Acknowledge Redirect Server
                       cpLog( LOG_DEBUG, "ACK %d response for INVITE",
                              statusCode );
                       ackAll( sipEvent, statusMsg );
                    }
                }
                

                if ( (statusCode >= 300) && (statusCode < 400) )
                {
                    cancelTimer( event );
                    lookupResults = processRedirectionMsg( sipEvent, statusMsg );
                } /* if status msg is 3xx */
                else if ( (statusCode >= 400) && (statusCode < 500 ) )
                {
                    cancelTimer( event );
                    lookupResults = CPLOpAckLookup::Lookup_NotFound;
                } /*else if */
            } /*if statusMsg*/
        } /*if sipEvent*/
    } /* else */

    if ( lookupResults != CPLOpAckLookup::Lookup_Undefined )
        return ( doOutput( event, lookupResults ) );
    else
        return 0;

}  //process


///
int
CPLOpAckLookup::processRedirectionMsg(Sptr < SipEvent > sipEvent,
                                      Sptr < StatusMsg > statusMsg)
{

    int statusCode = statusMsg->getStatusLine().getStatusCode();

    //302 Moved Temporarily
    if (statusCode == 302)
    {

        //Get Contact information
        //TODO:  Change sipEvent to sipProxyEvent
        Sptr<BaseUrl> nextHop = getContactUrl( sipEvent );

        //Save Contact information
        if ( myClearLocations == true )
            getStateMachineData(sipEvent)->getLocations()->clearAndPush( nextHop );
        else
            getStateMachineData(sipEvent)->getLocations()->push( nextHop );

        return CPLOpAckLookup::Lookup_Success;
    }  //statusCode = 302

    return CPLOpAckLookup::Lookup_Undefined;
}  //processRedirectMsg


///
void
CPLOpAckLookup::ackAll( Sptr < SipEvent > sipEvent, Sptr < StatusMsg > statusMsg )
{
    AckMsg ack( *statusMsg);

    //Get Contact information
    //TO DO:  Change sipEvent to sipProxyEvent
    Sptr<BaseUrl> redirectServerUrl = getContactUrl( sipEvent );
    RoundRobinServers& rServers = RoundRobinServers::instance( SERVER_RS );

    Sptr<SipUrl> sipRedirectServerUrl;
    sipRedirectServerUrl.dynamicCast( redirectServerUrl );
    Data lHost;
    int lPort;
    rServers.getHostNPort(statusMsg->getFrom().getUrl()->encode(), lHost, lPort);
    sipRedirectServerUrl->setHost(lHost); 
    sipRedirectServerUrl->setPort(lPort);

    SipRequestLine ackRequestLine( SIP_ACK, sipRedirectServerUrl);
    ack.setRequestLine(ackRequestLine);
    sipEvent->getSipStack()->sendAsync(ack);
}  //ackAll


/// Obtain the contact list from the contact header:
/// The next hop is the only contact item returned by the RS
Sptr<BaseUrl>
CPLOpAckLookup::getRequestUrlFromContact( Sptr < StatusMsg > statusMsg ) const
{
    Sptr <SipUrl> originalUrl;

    if ( (statusMsg->getNumContact())  > 0)
    {
        originalUrl.dynamicCast( statusMsg->getContact(-1).getUrl() ); 
    } // if statusMsg->getNumContact() > 0
    else
    {
        cpLog( LOG_ERR, "Missing Contact header.");
        originalUrl.dynamicCast( statusMsg->getTo().getUrl() ); 
    } /*else*/

    Sptr<SipUrl> returnUrl = new SipUrl;
    *returnUrl = *originalUrl;

    return returnUrl;

}    //getRequestUrlFromContact


///
Sptr < State >
CPLOpAckLookup::doOutput( Sptr < SipProxyEvent > event, int lookupResult )
{
    if ( lookupResult == CPLOpAckLookup::Lookup_Failure )
    {
        if ( myFailureOp != 0 )
            return myFailureOp->process( event );
        else
            lookupResult = CPLOpAckLookup::Lookup_NotFound;
    } /* if */

    if ( lookupResult == CPLOpAckLookup::Lookup_NotFound )
    {
        if ( myNotFoundOp != 0 )
            return myNotFoundOp->process( event );
        else
            lookupResult = CPLOpAckLookup::Lookup_Success;
    } /* if */

    if ( mySuccessOp != 0 )
        return mySuccessOp->process( event );
    else
    {
        cpLog( LOG_ERR, "Lookup has 0 operators." );
        assert( 0 );
    } /*else*/

    //Should never get to this point.
    return 0;
}  //doOutput


///Outputs the values of the member variables
void
CPLOpAckLookup::logData() const
{
    if (myClearLocations == true)
        cpLog(LOG_DEBUG, "[%x] CLEAR(yes)", name(), this );
    else
        cpLog(LOG_DEBUG, "[%x] CLEAR(no)", name(), this );

}  //logData


///
void
CPLOpAckLookup::doCpLog() const
{
    logData();

    if ( mySuccessOp != 0 )
    {
        cpLog(LOG_DEBUG, "%s[%x]'s SUCCESS follows:", name(), this );
        mySuccessOp->doCpLog();
        cpLog(LOG_DEBUG, "%s[%x]'s SUCCESS ends.", name(), this );
    } //if

    if ( myFailureOp != 0 )
    {
        cpLog(LOG_DEBUG, "%s[%x]'s FAILURE follows:", name(), this );
        myFailureOp->doCpLog();
        cpLog(LOG_DEBUG, "%s[%x]'s FAILURE ends.", name(), this );
    } //if

    if ( myNotFoundOp != 0 )
    {
        cpLog(LOG_DEBUG, "%s[%x]'s NOTFOUND follows:", name(), this );
        myNotFoundOp->doCpLog();
        cpLog(LOG_DEBUG, "%s[%x]'s NOTFOUND ends.", name(), this );
    } //if

    if ( myOperators.size() > 0 )
    {
        cpLog(LOG_DEBUG, "%s[%x]'s default follows:", name(), this );
        myOpConstIterator it;
        myOpConstIterator endOfList = myOperators.end();
        for (it = myOperators.begin(); it != endOfList; ++it )
        {
            if ( strcmp( (*it)->name(), "CPLOpProxy" ) == 0 )
                cpLog( LOG_DEBUG, (*it)->name() );
            else
                (*it)->doCpLog();
        }  /*for*/
        cpLog(LOG_DEBUG, "%s[%x]'s default ends.", name(), this );
    }  //if

}  //doCpLog


///Returns the name of the class

const char* const
CPLOpAckLookup::name() const
{
    return "CPLOpAckLookup";
}  //name


//End of file

