/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const CPLOpReject_cxx_Version =
    "$Id: CPLOpReject.cxx,v 1.22 2002/04/05 00:21:47 bko Exp $";

#include "CPLOpReject.hxx"
#include "StatusMsg.hxx"
#include "SipProxyEvent.hxx"
#include "SystemInfo.hxx"
#include "FsConfigFile.hxx"

using namespace Vocal;

///Constructor
CPLOpReject::CPLOpReject()
{
    //initializing member vars
    myStatus = CPLOpReject::Reject_rejectCodes;

    //Unnecessary to initialize myRejectReason
    //since "string" types default to "".

#ifdef MEM_TEST
    cpLog ( LOG_DEBUG, "new CPLOpReject[%x]", this );
#endif
}  //c'tor


///Destructor
CPLOpReject::~CPLOpReject()
{
#ifdef MEM_TEST
    cpLog ( LOG_DEBUG, "~CPLOpReject[%x]", this );
#endif
}  //d'tor


///Append an operator to this class
bool
CPLOpReject::setNextOperator( Sptr < Operator > op, int prevNode )
{
    //No operators need to be added.  The next thing to do is
    //go to the proxyState.
    return true;
}  //setNextOperator


///Initialize the Attributes member variables
bool
CPLOpReject::setAttributes( const char* attrName, const char* attrValue)
{
    if ( strcmp( attrName, "reason" ) == 0 )
    {
        myRejectReason = attrValue;
        return true;
    } /*if*/

    if ( strcmp( attrName, "status" ) == 0 )
    {
        if ( toUp(attrValue) == "BUSY" )
            myStatus = CPLOpReject::Busy_rejectCodes;
        else if ( toUp( attrValue ) == "REJECT" )
            myStatus = CPLOpReject::Reject_rejectCodes;
        else if ( toUp( attrValue ) == "NOTFOUND" )
            myStatus = CPLOpReject::NotFound_rejectCodes;
        else if ( toUp( attrValue ) == "ERROR" )
            myStatus = CPLOpReject::Error_rejectCodes;
        else
        {
            //See if the status code is >=400 and <700.
            unsigned int statusCode = atoi( attrValue );
            if (( statusCode >= 400 ) && ( statusCode <= 699 ))
                myStatus = statusCode;
            else return false;
        } /*else*/

        return true;
    } /*if*/

    return false;
}  //setAttributes


///
const Sptr < State >
CPLOpReject::process( Sptr < SipProxyEvent > event)
{
    if ( cpLogGetPriority() == LOG_DEBUG )
        logData();

    //Need the INVITE message to compose a 403.
    Sptr < InviteMsg > inviteMsg = getInvite( event );

    //Last Via may be us or it may be someone else.  Remove only
    //if it's us.
    removeVia( event, inviteMsg );

    Sptr < SipCommand > sipCommand = inviteMsg;
    int statusCode;
    switch (myStatus)
    {
        case Reject_rejectCodes:
        statusCode = 403;
        break;
        case Busy_rejectCodes:
        statusCode = 486;
        break;
        case NotFound_rejectCodes:
        statusCode = 404;
        break;
        case Error_rejectCodes:
        statusCode = 500;
        break;
        default:
        statusCode = myStatus;
    }
    StatusMsg statusMsg( *sipCommand, statusCode );
    sendReply( event, statusMsg );
    return myNextState;
}  //process


///
void
CPLOpReject::removeVia( Sptr < SipProxyEvent > event, Sptr < InviteMsg > inviteMsg )
{
    // Get last VIA's address
    LocalScopeAllocator lo;
    string viaHost( inviteMsg->getVia(0).getHost().getData(lo) );
    int viaPort = inviteMsg->getVia(0).getPort();
    NetworkAddress viaAddr( viaHost, viaPort );

    //Get Feature Server's host and port
    string hostString( theSystem.gethostAddress() );

    NetworkAddress FSAddress( hostString, sipPort );

    //Remove last VIA only if it belongs to FS's host
    if ( viaAddr == FSAddress )
        inviteMsg->removeVia( 0 );
    else
    {
        cpLog( LOG_DEBUG, "Not removing last VIA at %s:%d.",
               viaAddr.getIpName().c_str(),
               viaAddr.getPort() );
    } /*else*/

} //removeVia


///Outputs the values of the member variables
void
CPLOpReject::logData() const
{
    const int bufferSize = 10;

    char buffer[bufferSize];

    switch ( myStatus )
    {
        case CPLOpReject::Reject_rejectCodes:
        {
            strncpy( buffer, "reject", bufferSize );
        } 
        break;
        case CPLOpReject::Busy_rejectCodes:
        {
            strncpy( buffer, "busy", bufferSize );
        } 
        break;
        case CPLOpReject::NotFound_rejectCodes:
        {
            strncpy( buffer, "notfound", bufferSize );
        } 
        break;
        case CPLOpReject::Error_rejectCodes:
        {
            strncpy( buffer, "error", bufferSize );
        } 
        break;
        default:
        {
            sprintf( buffer, "%d", myStatus );
        } /*default*/
    } /*switch*/

    cpLog( LOG_DEBUG, "%s-MYSTATUS(%s) MYREJECTREASON(%s)",
           name(), buffer, myRejectReason.c_str() );

}  //logData


///Returns the name of the class
const char* const
CPLOpReject::name() const
{
    return "CPLOpReject";
}  //name


/// returns a case insensitive string
string CPLOpReject::toUp(const char *str)
{
    string retStr;
    for (; *str; str++)
        if (*str >= 'a' && *str <= 'z')
            retStr += (*str + 'A' -'a');
        else
            retStr += *str;
    return retStr;
}

//End of file

