/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const CPLOperator_cxx_Version =
    "$Id: CPLOperator.cxx,v 1.29 2002/04/05 02:40:40 bko Exp $";

#include "CPLOperator.hxx"
#include "CallInfo.hxx"
#include "SipEvent.hxx"
#include "TimerEvent.hxx"
#include "CallContainer.hxx"
#include "AckMsg.hxx"  
#include "FsConfigFile.hxx"

using namespace Vocal;

CPLOperator::CPLOperator()
{
}


CPLOperator::~CPLOperator()
{
    while ( !myOperators.empty() )
    {
        Sptr < CPLOperator > cplOp;
        cplOp.dynamicCast( myOperators.back() );
        cplOp->clearAllOperators();
        myOperators.pop_back();
    } /*while*/

    //Force the state to remove its operators.
    //State may have the ProxyOp which causes a
    //Sptr loop.  This will break that loop allowing
    //the objects to be deleted.
    if ( myNextState != 0 )
        myNextState->clearAllOperators();

} //Destructor


///
void
CPLOperator::clearAllOperators()
{
    if ( myNextState != 0 )
        myNextState->clearAllOperators();
    myNextState = 0;

}  //clearAllOperators


///
void
CPLOperator::setState( Sptr < CPLState > nextState )
{
    myNextState = nextState;
}  //setState


///
void
CPLOperator::addOperator( Sptr < Operator > op )
{

    myOperators.push_back( op );
}  //addOperator


///
const Sptr < State >
CPLOperator::process( const Sptr < SipProxyEvent > event )
{
    logData();

    if ( myOperators.size() == 0 )
        return this->myNextState;
    else
    {
        Sptr < State > returnState;
        myOpIterator op = myOperators.begin();
        myOpIterator opEnd = myOperators.end();

        while ( op != opEnd )
        {
            returnState = (*op)->process( event );
            if (returnState != 0) return returnState;

            ++op;

        } /* while */

        return 0;
    }  /* else */
}  //process


///
bool
CPLOperator::setTimer( Sptr < SipProxyEvent > event, unsigned int timeout ) const
{
    if ( timeout <= 0 ) return true;

    Sptr < InviteMsg > inviteMsg;
    inviteMsg = getInvite( event );
    if ( inviteMsg == 0 )
    {
        cpLog( LOG_ERR, "Expecting inviteMsg." );
        return false;
    } /*if inviteMsg */

    Sptr < TimerEvent > timerEvent = new TimerEvent(event->getFifo());

    cpLog(LOG_DEBUG, "Start TIMER for %d s.", timeout );

    // Save the timer event and the set INVITE message in the
    // Call Info as they are needed when timer goes off
    event->getCallInfo()->setTimer (timerEvent);
    event->getCallInfo()->setSipCommand ( inviteMsg );
    timerEvent->setCallInfo( event->getCallInfo(),
                             event->getCallContainer() );

    // Set the sipStack
    timerEvent->setSipStack( getStack(event) );

    //Timer runs in milliseconds.  Must convert timeout from seconds
    //to ms by multiplying by 1000.
    timerEvent->startTimer(timerEvent, timeout*1000);

    return true;
}  //setTimer


///
bool
CPLOperator::cancelTimer( Sptr < SipProxyEvent > event ) const
{
    if ( (event->getCallInfo()->getTimer()) != 0 )
    {
        cpLog( LOG_DEBUG, "Cancelling timer." );
        event->getCallInfo()->getTimer()->cancelTimer();

        //This forces CallInfo to release TimerEvent
        //allowing smart pointers to free the TimerEvent object.
        event->getCallInfo()->removeTimer( );

        return true;
    }  //getTimer


    cpLog( LOG_ERR, "No timer was found." );
    return false;

} //cancelTimer


///
void 
CPLOperator::ackStatusMsg( Sptr<SipProxyEvent> event )
{
    Sptr<StatusMsg> statusMsg;
    statusMsg.dynamicCast( getSipMsg( event ) );
    if ( statusMsg == 0)
    {
        cpLog( LOG_ERR, "No Status message found." );
        return;
    } /*if*/
    
    //Retrieve INVITE associated with sipEvent 
    Sptr < InviteMsg > savedInvite = getInvite( event ); 
    if ( savedInvite == 0 )
    {
        // error, no corresponding INVITE, return without sending ACK
        cpLog( LOG_ERR, "No INVITE associated with STATUS" );
        return ;
    }

    //TEST CODE
    SipVia v = getSipMsg( event )->getVia();
    cpLog (LOG_DEBUG, "Via[%s]-Alternate way to get sender (%s:%s)", 
                    v.encode().logData(), 
                    v.getReceivedhost().logData(),
                    v.getReceivedport().logData() );
    ///////////

    // send ACK to receiver of corresponding INVITE by setting Request URI
    // of ACK to that of the INVITE
    AckMsg ack( *statusMsg );
    SipRequestLine ackRequestLine( SIP_ACK,
    savedInvite->getRequestLine().getUrl() );
    ack.setRequestLine( ackRequestLine );

    cpLog(LOG_DEBUG, "Sending ACK[%s\n]",
                    ack.encode().logData() ); 
    sendAsync( event, ack );
} //ackStatusMsg


///
void
CPLOperator::addCallControlDiversion( Sptr < SipProxyEvent > event,
                                      Sptr < SipMsg > sipMsg,
                                      Sptr < BaseUrl> baseUrl ) const
{
    if (!addCCDiversion) return ;

    SipDiversion ccDiversion( baseUrl );

    switch ( getStateMachineData( event )->getProxyStatus() )
    {
        case FsStateMachnData::Busy:
        ccDiversion.setReason( Data("user-busy") );
        break;
        case FsStateMachnData::NoAnswer:
        ccDiversion.setReason( Data("no-answer") );
        break;
        default:
        ccDiversion.setReason( Data("unconditional") );
    } /*switch*/

#if 0  //Change to 1 if needed for debugging 
    cpLog( LOG_DEBUG, "CC-DIVERSION(%s)", ccDiversion.encode().logData() );
#endif

    sipMsg->setDiversion( ccDiversion );

} //addCallControlDiversion


///
Sptr < CallInfo >
CPLOperator::getCallInfo( Sptr < SipProxyEvent > event ) const
{
    return ( event->getCallInfo() );

} //getCallInfo


///
Sptr < InviteMsg >
CPLOperator::getInvite( Sptr < SipProxyEvent > event ) const
{
    //See if event is a SipEvent first
    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast( event );

    if (sipEvent != 0 )
    {
        Sptr < InviteMsg > inviteMessage;
        inviteMessage = sipEvent->getInvite();
        assert( inviteMessage != 0 );
        return inviteMessage;
    }  /*if sipEvent != 0 */

    //See if event is a TimerEvent
    Sptr < TimerEvent > timerEvent;
    timerEvent.dynamicCast( event );

    if ( timerEvent != 0 )
    {
        Sptr < SipCommand > sipCommand;
        sipCommand = timerEvent->getCallInfo()->getSipCommand();
        assert( sipCommand != 0 );

        Sptr < InviteMsg > inviteMessage;
        inviteMessage.dynamicCast( sipCommand );
        assert( inviteMessage != 0 );

        return inviteMessage;
    } /*if*/

    //If we reach here, then it's an unknown event
    assert( 0 );
    return 0;

} //getInvite


///
Sptr < SipCommand >
CPLOperator::getSipCommand( Sptr < SipProxyEvent > event ) const
{
    //See if event is a SipEvent first
    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast( event );

    if (sipEvent != 0 )
    {
        Sptr < SipMsg > sipMsg = sipEvent->getSipMsg();
        assert ( sipMsg != 0 );
        Sptr < SipCommand > sipCommand;
        sipCommand.dynamicCast( sipMsg );
        assert( sipCommand != 0 );
        return sipCommand;
    }

    //See if event is a TimerEvent
    Sptr < TimerEvent > timerEvent;
    timerEvent.dynamicCast( event );

    if ( timerEvent != 0 )
    {
        Sptr < SipCommand > sipCommand;
        sipCommand = event->getCallInfo()->getSipCommand();
        assert( sipCommand != 0 );
        return sipCommand;
    } /*if*/

    //If we reach here, then it's an unknown event
    assert( 0 );
    return 0;
} //getSipCommand


///
Sptr < SipMsg >
CPLOperator::getSipMsg( Sptr < SipProxyEvent > event ) const
{
    //See if event is a SipEvent first
    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast( event );

    if (sipEvent != 0 )
    {
        Sptr < SipMsg > sipMsg = sipEvent->getSipMsg();
        assert ( sipMsg != 0 );
        return sipMsg;
    } /*if*/

    //See if event is a TimerEvent
    Sptr < TimerEvent > timerEvent;
    timerEvent.dynamicCast( event );

    if ( timerEvent != 0 )
    {
        Sptr < SipCommand > sipCommand;
        sipCommand = event->getCallInfo()->getSipCommand();

        Sptr < SipMsg > sipMsg = sipCommand;
        assert ( sipMsg != 0 );
        return sipMsg;
    } /*if*/

    //If we reach here, it's an unknown event
    return 0;

}  //getSipMsg


///
Sptr < SipTransceiver >
CPLOperator::getStack( Sptr < SipProxyEvent > event ) const
{
    //See if event is a SipEvent first
    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast( event );

    if (sipEvent != 0 )
    {
        return sipEvent->getSipStack();
    } /*if*/

    //See if event is a TimerEvent
    Sptr < TimerEvent > timerEvent;
    timerEvent.dynamicCast( event );

    if ( timerEvent != 0 )
    {
        return timerEvent->getSipStack();
    } /*if*/

    //If we reach here, it's an unknown event?
    assert( 0 );

    return 0;
}  //getStack


///
Sptr < FsStateMachnData >
CPLOperator::getStateMachineData( Sptr < SipProxyEvent > event ) const
{
    Sptr < StateMachnData > callData;
    callData = getCallInfo( event )->getCallStateData();
    if ( callData == 0 )
    {
        cpLog (LOG_ERR, "Expecting StateMachnData." );
        assert( callData != 0 );
    }  /*if*/


    Sptr < FsStateMachnData > FsCallData;
    FsCallData.dynamicCast( callData );

    if (FsCallData == 0 )
    {
        cpLog (LOG_ERR, "Expecting FSStateMachnData." );
        assert( FsCallData != 0 );
    }  /*if*/

    return FsCallData;
} //getStateMachineData


///
Sptr<BaseUrl>
CPLOperator::getContactUrl( Sptr < SipProxyEvent > event ) const
{
    Sptr<StatusMsg> statusMsg;
    statusMsg.dynamicCast( getSipMsg( event ) );
    if ( statusMsg == 0)
    {
        cpLog( LOG_ERR, "No Status message found." );
        assert( 0 ); 
    } /*if*/

    Sptr<SipUrl> originalUrl;

    //Get URL from Contact-field or To-field              
    if ( (statusMsg->getNumContact())  > 0)
    {
        originalUrl.dynamicCast( statusMsg->getContact(-1).getUrl() );
    } // if statusMsg->getNumContact() > 0
    else
    {
        cpLog( LOG_ERR, "Missing Contact header.");
        originalUrl.dynamicCast( statusMsg->getTo().getUrl() );
    } /*else*/

    assert( originalUrl != 0 );

    Sptr<SipUrl> returnValue = new SipUrl;
    *returnValue = *originalUrl;

    return returnValue; 

} //getContactUrl


///
void
CPLOperator::sendReply( Sptr < SipProxyEvent > event, StatusMsg& sipMessage) const
{
    if ( event != 0 )
        getStack( event )->sendReply( sipMessage );
}  //sendReply


///
void
CPLOperator::sendAsync( Sptr < SipProxyEvent > event, SipCommand& sipCommand) const
{
    if ( event != 0 )
        getStack( event )->sendAsync( sipCommand );
}  //sendAsync


///
void
CPLOperator::logData() const
{
    myNextState->doCpLog();
    cpLog( LOG_DEBUG, "Size of myOperators=%d", myOperators.size() );

}  //logData


const char* const
CPLOperator::name() const
{
    return "CPLOperator";
}  //name


///
bool
CPLOperator::setNextOperator( Sptr < Operator > op, int prevNode )
{
    return false;
}  //setOutput


///
void
CPLOperator::doCpLog() const
{
    logData();

    if ( myNextState != 0 )
    {
        myNextState->doCpLog();
        return ;
    } /*if*/

    myOpConstIterator it = myOperators.begin();
    myOpConstIterator listEnd = myOperators.end();

    while (it != listEnd )
    {
        (*it)->doCpLog();
        ++it;
    } /*while*/

}  //doCpLog

//End of File

