/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package com.vovida.telephony;

import javax.telephony.*;
import javax.telephony.capabilities.*;
import com.vovida.telephony.events.*;
import com.vovida.telephony.capabilities.*;
import java.util.*;

/*******************************************************/
/* VCAddress is the Vocal Client Address Class which  */
/* is an implementation of the Address interface from  */
/* the JTapi Spec.                                     */
/*******************************************************/

/**
 * VCAddress is the Vocal Client Address Class which
 * is an implementation of the Address interface from the JTapi Spec.
 */
public class VCAddress implements Address
{
    // address is local if it is in the domain provided bby Provider

    final private static int LOCAl = 1;
    final private static int REMOTE = 2;
    protected String name;
    protected Provider provider;
    protected Vector terminals;
    protected Vector connections;
    protected Vector addressObservers;
    protected Vector callObservers;
    protected VCDynamicAddressCapabilities addressCapabilities;
    protected boolean isLocal;

    ///Private default constructor
    private VCAddress() {}

    // address only created from this package & subclasses
    // must know about provider

    ///Constructor
    protected VCAddress (Provider prov, boolean isLocal, String str)
    {
	this(prov, isLocal);
	name = str;
    }

    ///Private Constructor
    private VCAddress (Provider prov, boolean isLocal)
    {
	super();
	provider = prov;
	this.isLocal = isLocal;
	connections = new Vector();
	addressObservers = new Vector();
	callObservers = new Vector();
	terminals = new Vector();
	addressCapabilities = new VCDynamicAddressCapabilities(this);
    }

/****************************************************/
/* JTapi Implementation                             */
/****************************************************/

    // return number or ip for address
    /**
     * Refer To JTapi Address Interface
     */
    public String getName()
    {
	return name;
    }

    // return provider for address
    /**
     * Refer To JTapi Address Interface
     */
    public Provider getProvider()
    {
	if (provider == null)
	{
	    return null;
	}
	else
	    return provider;
    }

    // return all terminals associated with this address
    /**
     * Refer To JTapi Address Interface
     */
    public Terminal[] getTerminals()
    {
	Terminal[] terminalsArray = null;
	if (terminals.size() > 0)
	{
	    terminalsArray = new Terminal[terminals.size()];
	    terminals.copyInto(terminalsArray);
	}
	return terminalsArray;
    }

    // return all connections associated with this address
    /**
     * Refer To JTapi Address Interface
     */
    public Connection[] getConnections()
    {
	Connection[] connectionsArray = null;
	if (connections.size() > 0)
	{
	    connectionsArray = new Connection[connections.size()];
	    connections.copyInto(connectionsArray);
	}
	return connectionsArray;
    }

    // return capabilities for this address
    /**
     * Refer To JTapi Address Interface
     */
    public AddressCapabilities getCapabilities()
    {
	return addressCapabilities;
    }

    // add an Observer for this address - used when address is observed
    /**
     * Refer To JTapi Address Interface
     */
    public void addObserver(AddressObserver observer)
	throws ResourceUnavailableException,MethodNotSupportedException
    {
	addressObservers.addElement(observer);
    }

    // return all address observers
    /**
     * Refer To JTapi Address Interface
     */
    public AddressObserver[] getObservers()
    {
	AddressObserver[] observersArray = null;
	if (addressObservers.size() > 0)
	{
	    observersArray = new AddressObserver[addressObservers.size()];
	    addressObservers.copyInto(observersArray);
	}
	return observersArray;
    }

    // remove observer from the Address
    /**
     * Refer To JTapi Address Interface
     */
    public void removeObserver(AddressObserver observer)
    {
	addressObservers.removeElement(observer);
    }

    // Add a call observer to this Address - used when a call attaches
    // to this address (via connection) - added to the observers in
    // call - see addConnection() 
    /**
     * Refer To JTapi Address Interface
     */
    public void addCallObserver(CallObserver observer)
    	throws ResourceUnavailableException,MethodNotSupportedException
    {
	callObservers.addElement(observer);
    }

    // return all call observers in this Address
    /**
     * Refer To JTapi Address Interface
     */
    public CallObserver[] getCallObservers()
    {
	CallObserver[] observersArray = null;
	if (callObservers.size() > 0)
	{
	    observersArray = new CallObserver[callObservers.size()];
	    callObservers.copyInto(observersArray);
	}
	return observersArray;
    }

    // remove call observer from this address
    /**
     * Refer To JTapi Address Interface
     */
    public void removeCallObserver(CallObserver observer)
    {
	callObservers.removeElement(observer);
    }

    /// @deprecated function
    public AddressCapabilities getAddressCapabilities(Terminal terminal)
	throws InvalidArgumentException, PlatformException
    {
	return null;
    }


/**********************************************************/
/* Additional Functionality                               */
/**********************************************************/

    /**
     * Check if the Address local to the Provider domain
     */
    public boolean isLocal()
    {
	return isLocal;
    }

    /**
     * Associate a Terminal with this address
     */
    protected void addTerminal(Terminal term)
    {
	terminals.addElement(term);
    }

    /**
     * disassociate a terminal from this address
     */
    protected void removeTerminal(Terminal term)
    {
	terminals.removeElement(term);
    }

    /**
     * add a connection to this address - if call observers exist on this
     * address add them to the call associated with this connection
     */
    protected void addConnection(Connection conn)
    {
	boolean found;
	connections.addElement(conn);
	System.out.println("AddConnection " + getName());
	// Any call observers should now be added to the associated call
	try
	{
	    Call connCall = conn.getCall();
	    System.out.println("Call = " + ((VCCall)connCall).getCallId());
	    CallObserver[] callObs = getCallObservers();
	    CallObserver[] existingObs = connCall.getObservers();
	    if (callObs == null)
	    {
		return;
	    }
	    for (int i=0; i < callObs.length; i++)
	    {
		found = false;
		if (existingObs != null)
		{
		    for (int j=0; j < existingObs.length; j++)
		    {
		        if (callObs[i] == existingObs[j])
		        {
			    found = true;
		        }
		    }
		}
		if (found == false)
		{

	            connCall.addObserver(callObs[i]);
		}
	    }
	}
	catch (Exception e)
	{
	    e.printStackTrace();
	}
    }

    /**
     * Remove Connection from Address - remove address from current active
     * addresses
     */
    protected void removeConnection(Connection conn)
    {
	connections.removeElement(conn);
	((VCProvider)provider).removeAddressFromList(this);
    }

    /**
     * check that the provider is in a suitable state
     * if not, an InvalidStateException is thrown
     */
    protected void isProviderOK() throws InvalidStateException
    {
        if (this.getProvider().getState() != Provider.IN_SERVICE)
        {
            throw (new InvalidStateException(this.getProvider(),
                                        InvalidStateException.PROVIDER_OBJECT,
                                        this.getProvider().getState()));
        }
        return;
    }

}


