
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const JSProvider_cxx_Version =
    "$Id$";
#include "JSProvider.hxx"
#include <cpLog.h>
#include "VNetworkException.hxx"
#include "JSStateMachine.hxx"
#include "SipRequestLine.hxx"
#include "SipUrl.hxx"
#include "SipCommand.hxx"
#include "JSCommon.hxx"
#include "JSNotification.hxx"


///
JSProvider::JSProvider()
{
    itsStateMachine = new JSStateMachine;
}

///
JSProvider::JSProvider(const JSProvider& rhs)
{
    if (this != &rhs)
    {
        copyRhsToThis( rhs );
    }
}

///
JSProvider&
JSProvider::operator=(const JSProvider& rhs)
{
    if (this != &rhs)
    {
        copyRhsToThis( rhs );
    }
    return *this;
}

void
JSProvider::process( const Sptr < SipProxyEvent > nextEvent )
{
    cpLog ( LOG_DEBUG, "JSProvider::process: received event %s", nextEvent->name());
    assert ( nextEvent != 0 );

    assert ( itsUserCallContainer != 0 );
    assert ( itsSipCallContainer != 0 );
    Sptr < JSUserCallContainer > userMap;
    Sptr < JSSipCallContainer > sipMap;
    Sptr < JSUserRegContainer > addrMap;
    userMap.dynamicCast( itsUserCallContainer );
    sipMap.dynamicCast( itsSipCallContainer );
    addrMap.dynamicCast( itsUserRegContainer );
    assert( userMap != 0 );
    assert( sipMap != 0 );
    assert( addrMap != 0 );

    Sptr < JSCall > call;

    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast( nextEvent );

    if ( sipEvent != 0 ) 	// SIP event
    {

        Sptr < SipMsg > sipMsg = sipEvent->getSipMsg();
        if ( sipMsg == NULL )
        {
            cpLog(LOG_ERR, "sipMsg of sip event is NULL");
            assert(0);
        }

        cpLog(LOG_DEBUG, "******** SIP message received ********");
        cpLog(LOG_DEBUG, "%s", sipMsg->encode().logData() );
        cpLog(LOG_DEBUG, "******** SIP message received ********");

        // Get Sip Call ID from the message
        SipCallId callId = sipMsg->getCallId();

        call = sipMap->findCall( callId );

        // If we found the call object, don't do anything,
        // let state machine handle this event
        // If we can't find the call object, go on
        if ( call == 0 )
        {
            // Check if this is an INVITE
            if ( SIP_INVITE != sipMsg->getType() )
            {
                // Except INVITE, every message should have an existing call
                // object that the message is belong to, this is an error
                cpLog(LOG_ERR, "No call object found for the message");
                return ;
            }

            // Search the user map for call object
            // If found, must be an INVITE from caller.
            // Add <SIP Call ID, call> pair to sip map, then let the
            // state machine handle this event.
            Sptr <BaseUrl> fromBaseUrl = sipMsg->getFrom().getUrl();
            Sptr <BaseUrl> toBaseUrl = sipMsg->getTo().getUrl();
            assert( fromBaseUrl.getPtr() != 0 );
            assert( toBaseUrl.getPtr() != 0 );
            assert( fromBaseUrl->getType() == SIP_URL );
            assert( toBaseUrl->getType() == SIP_URL );

            Data fromUser = fromBaseUrl->getUserValue();
            Data toUser = toBaseUrl->getUserValue();
            call = itsUserCallContainer->findCall( fromUser );
            if ( call == 0 )
                call = itsUserCallContainer->findCall( toUser );
            else
                cpLog(LOG_WARNING, "A call object is found with from user of the new INVITE");

            if ( call != 0 )
            {
                // Add <SIP Call ID, call> pair to sip map
                //itsSipCallContainer->addCall( callId, call );
                cpLog(LOG_WARNING, "A call object is found with to user of the new INVITE");
                // This is the potential place to handle multiple calls
                // But, for now, we don't support the capability yet.
                return ;
            }
            else
            {
                Sptr < SipCommand > invite;
                invite.dynamicCast( sipMsg );
                // Check which port is this message received
                Sptr <BaseUrl> baseUrl;

                baseUrl = invite->getRequestLine().getUrl();
                assert( baseUrl.getPtr() != 0 );
                assert( baseUrl->getType() == SIP_URL ); // JTAPI doesn't support TEL_URL

                Sptr <SipUrl> sipUrl;
                assert( sipUrl.dynamicCast( baseUrl ) != 0 );

                Data port = sipUrl->getPort();

                cpLog(LOG_DEBUG, "Sip event received on port: %s",
                      port.logData());

                if ( (port == getConfig()->getJSNSipPort()) ||
                        (port == getConfig()->getJSDSipPort()) )
                {
                    // This is either the case of
                    // Caller dial from phone set directly, or
                    // Answering call from non-JS caller.
                    // Either way, we will create the call object and
                    // let state machine handle the event.
                    call = itsUserCallContainer->newCall( fromUser );
                    assert( call != 0 );
                    itsUserCallContainer->addCall( toUser, call );
                    call->setFeature( itsStateMachine );
                    call->setState( itsStateMachine->findState( "JSStateIdle" ));
                    call->setProvider( this );
                    itsSipCallContainer->addCall( callId, call );
                    call->setFromUser( fromUser );
                    call->setToUser( toUser );

                    // If received on JSN, caller calling from set
                    // It is considered a controlled call
                    // If received on JSD, answering call from outside
                    if ( port == getConfig()->getJSNSipPort() )
                        call->setControlled( true );
                    else
                        call->setControlled( false );
                }
                else
                {
                    cpLog( LOG_ERR, "Sip event recevied on an unknown port: %s",
                           port.logData() );
                    return ;
                }
            }
        }
    }
    else	// JSReqEvent
    {
        Sptr < JSReqEvent > reqEvent;
        reqEvent.dynamicCast( nextEvent);
        if ( reqEvent != 0 )
        {
            cpLog(LOG_DEBUG, "******** JS message received ********");
            cpLog(LOG_DEBUG, "%s", reqEvent->getReqMsg().c_str() );
            cpLog(LOG_DEBUG, "************************************");
            JSRequest* request = reqEvent->decode();

            if ( request == 0 )
            {
                // Rather than fail the JTAPI server, we should just fail
                // this particular request and move on because it is
                // possible to receive a ill-formatted request from
                // may be an older version of vocalpad
                // assert( 0 );
                return ;
            }

            //	    request->setSender( reqEvent->getSender() );
            Data fromUser( request->getCallerAddr() );
            Data toUser( request->getCalleeAddr() );

            switch ( request->getMsgType() )
            {
                case TYPECHKREGISTER:
                {
                    // add some code for chek reg response
                    // Applet send this request when the web is changed
                    // and then come back
                }

                break;
                case TYPEREGISTER:
                {
                    cpLog(LOG_DEBUG, " Received TRGISTER request event");

                    NetworkAddress* netAddress = addrMap->findAddr( fromUser );
                    if ( netAddress == 0 )
                    {
                        // Add new user to address map
                        netAddress = new NetworkAddress();
                        netAddress->setPort( atoi((request->getPort()).c_str()) );
                        netAddress->setHostName((reqEvent->getSender()).getHostName());
                        addrMap->newUser( fromUser, netAddress );
                    }

                    // Send REGISTER response to vocalpad
                    Sptr < UdpStack > clientUdp = getClientUdpStack();
                    JSNotification notification(clientUdp, *netAddress);

                    notification.setType(REGISTERRESP);
                    notification.setId( request->getMsgID() );
                    notification.setAddr(request->getCallerAddr());
                    notification.setResult(JS_SUCCESS);
                    notification.send();
                    cpLog(LOG_DEBUG, "Registration response sent: %s",
                          notification.getNotificationString().c_str());
                    return ;
                }
                case TYPEDISCONNECT:
                {
                    NetworkAddress* na = addrMap->findAddr( fromUser );
                    if (na != 0)
                    {
                        request->setSender( *na );
                    }
                    else
                    {
                        // Problem with user not registered, log & return
                        cpLog(LOG_ERR, "JSReqEvent from non-registered user %s",
                              fromUser.logData());
                        return ;
                    }

                    call = itsUserCallContainer->findCall( fromUser);
                    if ( call == NULL)
                    {
                        cpLog(LOG_ERR, "Receive disconnect request but no cll object associated with it!");
                        return ;
                    }
                    else
                    {
                        call->setRequest( request );
                    }
                }
                break;
                case TYPECALLCREATERESP:
                {
                    Data id( request->getMsgID() );
                    SipCallId callId;
                    callId.setData( id );

                    call = itsSipCallContainer->findCall( callId );
                    if ( call == NULL)
                    {
                        cpLog(LOG_ERR, "Receive call create response but no cll object associated with it!");
                        return ;
                    }
                    else
                    {
                        call->setRequest( request );
                    }
                }
                break;
                case TYPECONNECT:
                {
                    NetworkAddress* na = addrMap->findAddr( fromUser );
                    if (na != 0)
                    {
                        request->setSender( *na );
                    }
                    else
                    {
                        // Problem with user not registered, log & return
                        cpLog(LOG_ERR, "JSReqEvent from non-registered user %s",
                              fromUser.logData());
                        return ;
                    }

                    call = itsUserCallContainer->findCall( fromUser);
                    if ( call == NULL)
                    {
                        // Create a new call and let the state machine
                        // handle the call
                        // Change to provider->getUserCallContainer()->newCall( user) ;
                        call = userMap->newCall( fromUser );
                        assert( call != 0 );
                        userMap->addCall( toUser, call );
                        call->setRequest( request );
                        call->setFeature( itsStateMachine );
                        call->setState( itsStateMachine->findState( "JSStateIdle" ));
                        call->setProvider( this );
                        call->setControlled( true );
                        call->setFromUser( fromUser );
                        call->setToUser( toUser );
                    }
                    else
                    {
                        // If receive a connect request and an call object
                        // already, must be an retransmit, drop it
                        return ;
                    }
                }
                break;
                default:
                cpLog(LOG_WARNING,
                      "JTAPI request type %d not yet supported",
                      request->getMsgType());
            }
        }
        else	// Other types of events, i.e. not SIP, not JTAPI request
        {
            // So far, there are no other types of event.
            // But something may come later. You never know.
            cpLog(LOG_ERR, "Unknown type of event received");
            return ;
        }
    }

    nextEvent->setCallInfo ( call, sipMap );
    itsStateMachine->process( nextEvent );
}

///
void
JSProvider::copyRhsToThis (const JSProvider& rhs)
{ }

///

JSProvider::~JSProvider()
{}

///

void
JSProvider::setUserCallContainer( const Sptr < JSUserCallContainer > callContainer )
{
    itsUserCallContainer = callContainer;
}

///
Sptr < JSUserCallContainer >
JSProvider::getUserCallContainer( )
{
    return itsUserCallContainer;
}

///
void
JSProvider::setSipCallContainer( const Sptr < JSSipCallContainer > callContainer )
{
    itsSipCallContainer = callContainer;
}

///
Sptr < JSSipCallContainer >
JSProvider::getSipCallContainer( )
{
    return itsSipCallContainer;
}

///
void
JSProvider::setUserRegContainer( const Sptr < JSUserRegContainer > regContainer )
{
    itsUserRegContainer = regContainer;
}

///
Sptr < JSUserRegContainer >
JSProvider::getUserRegContainer( )
{
    return itsUserRegContainer;
}

///
void
JSProvider::setJSNSipStack( const Sptr < SipTransceiverFilter > sipStack )
{
    assert (sipStack != 0 );

    itsJSNSipStack = sipStack;
}

void
JSProvider::setClientUdpStack( const Sptr < UdpStack > udpStack )
{
    assert (udpStack != 0 );
    itsClientUdpStack = udpStack;
}

///
bool
JSProvider::isExist(string user)
{
    Data newUser(user);
    Sptr < JSCall > call;
    call = itsUserCallContainer->findCall(newUser);
    if (call != 0)
        return true;
    else
        return false;
}

///
void
JSProvider::setJSDSipStack( const Sptr < SipTransceiverFilter > sipStack )
{
    assert (sipStack != 0 );

    itsJSDSipStack = sipStack;
}

//End of File
