
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const MrshlUserCache_cxx_Version =
    "$Id: MrshlUserCache.cxx,v 1.34 2002/03/19 21:39:47 jason Exp $";

#include "MrshlUserCache.hxx"
#include "ProvisionInterface.hxx"
#include "VAccountData.hxx"
#include "limits.h"
#include "VPs_def.h"

using namespace Vocal;

const unsigned int  MS_CACHE_SIZE = 10000;

// initialization of static members
MrshlUserCache* MrshlUserCache::myInstance = 0;
unsigned int MrshlUserCache::indexCounter  = 0;
unsigned int MrshlUserCache::indexMarker   = 0;
bool MrshlUserCache::firstPass             = false;

VRWLock MrshlUserCache::cacheLock;


MrshlUserCache::MrshlUserCache( const string msGroup)
{
    try
    {
        ProvisionInterface& pif = ProvisionInterface::instance();

        myMarshalGroup = msGroup.c_str();

        pif.registerDirForUpdate( MrshlUserCache::updateUsers,
                                  VACCOUNTS );
        pif.registerDirForUpdate( MrshlUserCache::updateUsers,
                                  VALIASES );

    }
    catch ( VException& e )
    {
        cpLog( LOG_ALERT, "Failed to connect to Prov Server, reason %s",
               e.getDescription().c_str() );
    }
}


MrshlUserCache::~MrshlUserCache()
{
}


MrshlUserCache&
MrshlUserCache::instance( const char* msGroup )
{
    if ( !myInstance )
    {
        assert( msGroup );
        myInstance = new MrshlUserCache( msGroup);
    }
    return *myInstance;
}


void
MrshlUserCache::createUserData(const string& user, MrshlUserData& userData )
throw ( VMissingDataException& )
{
    // Get info from ps and create.
    ProvisionInterface& pif = ProvisionInterface::instance();


    // Need to do a look up in provisioning to find out the master user
    // using the given user name... it is possible that the given user 
    // name is an alias. 

    string masterUser = user;

    try
    {
        if ( !pif.isAccount( masterUser ) )
        {
            // reuse the masterUser to store the result.
            masterUser = pif.getMasterUser( masterUser );
            cpLog(LOG_DEBUG, "Converting %s to master %s",
                  user.c_str(), masterUser.c_str());
        }
    }
    catch( VException &e )
    {
        // Got an exception.  User must already be the master user.
        cpLog(LOG_DEBUG, "User %s is already a master", 
              masterUser.c_str() );
    }


    // Now get provisioning info for this user based on the master user.
    try
    {
        VAccountData accountData;

        cpLog(LOG_DEBUG, "Getting user %s info from prov...",
              masterUser.c_str() );

        pif.getAccountData( masterUser, accountData );

        // Create new userData and copy accountData before inserting
        // into cache.
        // userData members realm and nonce are set when user authenicates

        userData.authType = accountData.getAuthorizationType();
        userData.ip = accountData.getUserIP();
        userData.password = accountData.getAuthorizationPassword();
        userData.index = getIndex();
        //Set the default trasport UDP
        userData.transport = "UDP";

#if 1
        if ( myMarshalGroup == accountData.getMarshalGroup() )
        {
            userData.inMarshalGroup = true;
        }
        else
        {
            // Found user but not in our marshal group.
            userData.inMarshalGroup = false;

            cpLog(LOG_DEBUG, "User not is this group: %s", 
                             accountData.getMarshalGroup().c_str() );

            throw VMissingDataException ("*** User not in this group ***",
                                         __FILE__, __LINE__, 0 );
        }
#endif

    }
    catch ( VException &e )
    {
        throw VMissingDataException( "*** User not found in provisioning ***",
                                     __FILE__, __LINE__, 0 );
    }

    cpLog( LOG_DEBUG, "*** updating cache <user:%s>***", user.c_str() );

    // Check cache size for clean up.

    cpLog( LOG_DEBUG, "*** cache size (before add) = %d ***", 
           myUserCache.size() );

    cacheLock.WriteLock();

    // insert into map.
    myUserCache[ user ] = userData;

    cacheLock.Unlock();

    cpLog( LOG_DEBUG, "*** cache size (after add) = %d ***", 
           myUserCache.size() );
}


void
MrshlUserCache::getMrshlUserData( const string& user, MrshlUserData& userData)
throw ( VMissingDataException& )
{
    cpLog (LOG_DEBUG, "getMrshlUserData:  finding user <%s>", user.c_str() );

    if( myInstance->findUserData( user, userData ) )
    {
        if ( userData.inMarshalGroup )
        {
            // found match in myUserCache
            cpLog( LOG_DEBUG, "***** User Data (Cache) *****" );
            cpLog( LOG_DEBUG, "      target(key): %s", user.c_str() );
            cpLog( LOG_DEBUG, "                    -begin data-" );
            cpLog( LOG_DEBUG, "         password: %s", userData.password.c_str() );
            cpLog( LOG_DEBUG, "         authType: %s", userData.authType.c_str() );
            cpLog( LOG_DEBUG, "               ip: %s", userData.ip.c_str() );
            cpLog( LOG_DEBUG, "            realm: %s", userData.realm.c_str() );
            cpLog( LOG_DEBUG, "            nonce: %s", userData.nonce.c_str() );
            cpLog( LOG_DEBUG, "            index: %d", userData.index );
            cpLog( LOG_DEBUG, "            Transport: %s", userData.transport.c_str() );
            cpLog( LOG_DEBUG, "***** End User Data *****" );
        }
        else
        {
            throw VMissingDataException( "*** User not in Marshal Group ***",
                                         __FILE__, __LINE__, 0 );
        }
    }
    else
    {
        cpLog( LOG_INFO,
               "*** User not found in cache.  try to create now... ***" );

        try
        {
            myInstance->createUserData( user, userData );
        }
        catch ( VException &e )
        {
            throw VMissingDataException( "*** User not found in provisioning ***",
                                         __FILE__, __LINE__, 0 );
        }
    }
}



bool
MrshlUserCache::findUserData( const string& user, MrshlUserData& userData )
{
    cacheLock.ReadLock();
    MrshlUserMap::iterator p = myUserCache.find( user );

    bool found = ( p != myUserCache.end() );
    cacheLock.Unlock();

    if ( found )
    {
        userData =  p->second;
        return true;
    }
    else
    {
        return false;
    }
}


void
MrshlUserCache::addUserAuthData( const string& user,
                                 const string& userRealm,
                                 const string& userNonce )
throw ( VMissingDataException& )
{
    cpLog( LOG_DEBUG, "*** add user auth data <user:%s> <realm:%s> <nonce:%s>",
           user.c_str(),
           userRealm.c_str(), 
           userNonce.c_str() );

    cacheLock.WriteLock();

    MrshlUserMap::iterator p = myUserCache.find( user );

    if ( p!= myUserCache.end() )
    {
        (p->second).realm = userRealm;
        (p->second).nonce = userNonce;
    }
    else
    {
        cacheLock.Unlock();
        throw VMissingDataException( "*** User not found in myUserCache ***",
                                     __FILE__, __LINE__, 0 );
    }
    cacheLock.Unlock();
}


void
MrshlUserCache::updateUserTransport( const string& user,
				     const string& transport)
throw ( VMissingDataException& )
{
    cpLog( LOG_DEBUG, "*** Update user transport <user:%s> <Transport:%s> ",
           user.c_str(),
           transport.c_str() );

    cacheLock.WriteLock();

    MrshlUserMap::iterator p = myUserCache.find( user );

    if ( p!= myUserCache.end() )
    {
        (p->second).transport = transport;
    }
    else
    {
        cacheLock.Unlock();
        throw VMissingDataException( "*** User not found in myUserCache ***",
                                     __FILE__, __LINE__, 0 );
    }
    cacheLock.Unlock();
}
     

void
MrshlUserCache::updateUserData( const string& user,
                                const string& userAuthType,
                                const string& userIp,
                                const string& userPassword )
throw ( VMissingDataException& )
{
    cpLog( LOG_DEBUG, "*** update user data <user:%s> <auth:%s> <ip:%s> <passwd:%s>",
           user.c_str(),
           userAuthType.c_str(),
           userIp.c_str(),
           userPassword.c_str() );

    cacheLock.WriteLock();
    MrshlUserMap::iterator p = myUserCache.find( user );

    if ( p!= myUserCache.end() )
    {
        (p->second).authType = userAuthType;
        (p->second).ip = userIp;
        (p->second).password = userPassword;

    }
    else
    {
        cacheLock.Unlock();
        throw VMissingDataException( "*** User not found in myUserCache ***",
                                     __FILE__, __LINE__, 0 );
    }
    cacheLock.Unlock();
}


void
MrshlUserCache::updateUsers( const string& data,
                             const string& fileName,
                             const bool deletedFlag )
{
    MrshlUserCache & myUserCache = MrshlUserCache::instance();

    // delete flag is set...  user has been deleted from provisioning.
    if (deletedFlag)
    {
        myUserCache.removeUser( data );
        return ;
    }


    ProvisionInterface& pif = ProvisionInterface::instance();
    VAccountData accountData;

    // strip off xml part of data
    string userProv = data;
    string::size_type pos = userProv.find( ".xml" );
    if ( pos != string::npos )
    {
        userProv = userProv.substr(0, pos);
    }

    cpLog( LOG_INFO, "*** updateUser: %s ***", userProv.c_str() );

    // read new changes from provisioning
    try
    {
        pif.getAccountData( userProv, accountData );
    }
    catch ( VException &e )
    {
        cpLog( LOG_INFO, "*** User not exist...  Update failed. ***" );
        return ;
    }

    list<string> aliasList;
    accountData.getAliases( aliasList );

    // add master to list of keys to search for in map
    aliasList.push_back( userProv );

    // update all aliases and master the user can be stored under
    for( list<string>::const_iterator itr = aliasList.begin();
         itr != aliasList.end(); itr++ )
    {
        try
        {
            myUserCache.updateUserData( *itr,
                                        accountData.getAuthorizationType(),
                                        accountData.getUserIP(),
                                        accountData.getAuthorizationPassword() );
        }
        catch( VException &e )
        {
            cpLog( LOG_INFO, "Alias %s not found while updating %s",
                   (*itr).c_str(), userProv.c_str() );
        }
    }
}

int
MrshlUserCache::getIndex()
{
    // getIndex() keeps track of the index that it gives out for all
    // users.  cleanCache() will remove the users based on the it's index.

    if ( indexCounter == MS_CACHE_SIZE )
    {
        indexCounter = 0;
        indexMarker = 0;
        firstPass = true;
    }

    if ( firstPass )
    {
        cleanCache();
    }

    // return then increment.

    return (indexCounter ++);
}


void
MrshlUserCache::cleanCache()
{
    // cleanCache will remove users from the cache to create room for 
    // new users.

    cpLog(LOG_DEBUG, "***  cleaning cache... ***");

    MrshlUserData userData;

    cacheLock.WriteLock();

    MrshlUserMap::iterator itr;

    // go through the cache until we've deleted enough users.

    itr = myUserCache.begin();

    while( itr != myUserCache.end() )
    {
        userData = itr->second;

        if ( userData.index == indexMarker )
        {
            cpLog(LOG_DEBUG, "*** Removing user <%s :%d> from cache ***",
                      itr->first.c_str(), userData.index );

            MrshlUserMap::iterator itr2 = itr ++;
            myUserCache.erase( itr2 );
            break;
        }
        else
        {
            itr ++;
        }
    }

    indexMarker ++;

    cacheLock.Unlock();
}


void
MrshlUserCache::removeUser(const string& user)
{
    cacheLock.WriteLock();

    cpLog( LOG_DEBUG, "*** Trying to delete user <%s> ***", user.c_str() );
    MrshlUserMap::iterator itr = myUserCache.find( user );

    if ( itr != myUserCache.end() )
    {
        myUserCache.erase( itr );

        cpLog( LOG_DEBUG, "*** User <%s> deleted ***", user.c_str() );
    }
    else
    {
        cpLog( LOG_DEBUG, "*** User <%s> not in cache ***", user.c_str() );
    }

    cacheLock.Unlock();
}

