
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const GroupContainer_cxx_Version =
    "$Id: GroupContainer.cxx,v 1.32 2002/02/25 05:51:19 sprajpat Exp $";
#include "cpLog.h"
#include "ProvisionInterface.hxx"
#include "GroupContainer.hxx"
#include "LockHelper.hxx"


Sptr < GroupContainer > GroupContainer::theGroupContainer = 0;

/**
 *  Return the only Group container or create it
 *  if it doesn't already exist. 
 * @return  Reference to the singleton for this object
 */
GroupContainer&
GroupContainer::instance( )
{
    if ( theGroupContainer == 0 )
    {
        theGroupContainer = new GroupContainer;
    }
    return *(theGroupContainer);
}

/**
 * Constructor
 */
GroupContainer::GroupContainer( )
{
    // initialize
    itsValidGroupList = new vector < Group > ;

    // get valid groups from PS
    getValidGroupsFromPS();

}

/**
 * Process a notification of an update to group data in provisioning.
 */
void
GroupContainer::handleUpdateFromPS( )
{
    // delete all group entries
    itsValidGroupList->clear();

    // get valid groups from PS
    getValidGroupsFromPS();
}


/**
 * Obtain the valid group information from provisioning
 */
void
GroupContainer::getValidGroupsFromPS( )
{
    //get all valid Jtapi groups
    getValidGroupsOfType(SERVER_JS);
    //get all valid Feature groups
    getValidGroupsOfType(SERVER_FS);
    //get all valid Marshal groups
    getValidGroupsOfType(SERVER_MS);

    if (cpLogGetPriority() >= LOG_DEBUG)
    {
        //display all valid groups
        cpLog( LOG_DEBUG, "Valid feature/marshal groups:");
        for (vector < Group > ::iterator i = itsValidGroupList->begin();
                i != itsValidGroupList->end(); ++i)
        {
            cpLog( LOG_DEBUG, "Group: %s Type: %d",
                   i->getName().logData(), i->getType() );
        }
    }
}

/**
 * Obtain valid groups of a specific type from provisioning
 * @param  type  Type of server
 */
void
GroupContainer::getValidGroupsOfType(const ServerType type)
{
    list < string > groups;

    try
    {
        ProvisionInterface& pif = ProvisionInterface::instance();
        switch (type)
        {
            case SERVER_JS:
            {
                pif.getJtapiGroups( groups );
                break;
            }
            case SERVER_FS:
            {
                pif.getFeatureGroups( groups );
                break;
            }
            case SERVER_MS:
            {
                pif.getMarshalGroups( groups );
                break;
            }
            default:
            {
                cpLog(LOG_DEBUG, "Unknown server type");
                break;
            }
        }

    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
        return;
    }

    for (list < string > ::iterator j = groups.begin();
            j != groups.end(); ++j)
    {
        Group groupPtr( type, *j );
        itsValidGroupList->push_back( groupPtr );
    }
}

/**
 * Verify if a specific group is valid
 * @param  group  Group identifier
 * @param  type  Type of servers belonging to the given group
 * @return  True if the group is valid, false otherwise.
 */
bool
GroupContainer::isValid( const Data& group, ServerType& type)
{
    for (vector < Group > ::iterator i = itsValidGroupList->begin();
            i != itsValidGroupList->end(); ++i)
    {
        if ( i->getName() == group)
        {
            type = i->getType();
            return true;
        }
    }
    type = SERVER_Unknown;
    return false;
}

/**
 * Destructor
 */
GroupContainer::~GroupContainer( )
{}

/**
 * Add a group to the container
 * @param  key  Group identifier
 * @param  listofServers  Servers belonging to the group to be added
 */
void
GroupContainer::addGroup( const Data& key,
                          const Sptr < vector < Data > > listofServers)
{
    cpLog( LOG_DEBUG, "entering GroupContainer::addGroup");

    /// insert the Group in the map
    groupList[key] = listofServers;

}

/**
 * Add a server to a group already in the container
 * @param  key  Group identifier
 * @param  server  Server to be added to the group's list of servers
 */
void
GroupContainer::addServerToGroup( const Data& key,
                                  const Data& server)
{
    cpLog( LOG_DEBUG, "entering GroupContainer::addServerToGroup");

    /// insert the server into its Group
    TableIter i;

    { /// turn on lock
        LockHelper helper(lock);

        Sptr < vector < Data > > serverList;

        /// find the group in the database using the key
        i = groupList.find(key);

        /// verify that a group was found
        if (i != groupList.end())
        {
            // insert server into group
            serverList = i->second;
            serverList->push_back(server);
        }
        else
        {
            // create new group
            serverList = new vector < Data > ;
            serverList->push_back(server);
            addGroup(key, serverList);
        }
    } /// turn off lock

    // display group after a new server is added
    display(key);
}


/**
 * Remove the specified group from the map
 * @param  key  Group identifier
 */
void
GroupContainer::removeGroup( const Data& key )
{
    cpLog( LOG_DEBUG, "GroupContainer::removeGroup");
    TableIter i;

    /// Find the Group in the database using the key
    i = groupList.find(key);

    /// Verify that a Group was found
    if (i != groupList.end())
    {
        /// remove the key/Group from the map
        groupList.erase(i);
    }
}


/**
 * Remove server from group  
 * @param  key  Group identifier
 * @param  server  Server to be removed
 */
void
GroupContainer::removeServerFromGroup( const Data& key,
                                       const Data& server )
{
    cpLog( LOG_DEBUG, "GroupContainer::removeServerFromGroup");
    TableIter i;
    {
        LockHelper helper(lock);

        /// Find the Group in the database using the key
        i = groupList.find(key);

        /// Verify that a Group was found
        if (i != groupList.end())
        {
            Sptr < vector < Data > > & serverList = i->second;
            if ( serverList->empty() )
            {
                // this is an error condition. listofServers should
                // never be empty since the entire group is deleted
                // from the container once the last item in the group
                // is removed.
                cpLog( LOG_DEBUG, "no servers to remove");
            }
            else
            {
                for (vector < Data > ::iterator j = serverList->begin();
                        j != serverList->end(); ++j)
                {
                    if ( *j == server)
                    {
                        if ( serverList->size() == 1 )
                        {
                            //remove group since j is the last element
                            removeGroup(key);
                            break;
                        }
                        else
                        {
                            serverList->erase(j);
                            break;
                        }
                    }
                }

            }

        }
        else
        {
            cpLog( LOG_DEBUG, "group not found");
        }


    }
    // display group to make sure server/group is deleted.
    display(key);
}

/**
 * Obtain a server from the specified group
 * @param  key  Group identifier
 * @param  server  The resulting server found within the given group
 * @return  True if a server was found, false if the group is not in the container
 */
bool
GroupContainer::getServer( const Data& key, Data& server )
{

    Sptr < vector < Data > > listofServers;

    /// Find the Group in the container and
    /// retrieve the list of servers in group
    if (search(key, listofServers))
    {
        LockHelper helper(lock);

        /// returning the last server in the list
        server = *listofServers->begin();
        return true;
    }
    return false;
}


/**
 * Search for a given group and obtain its list of servers
 * @param  key  Group
 * @param  listofServers  The servers belonging to the specified group
 * @return 
 */
bool
GroupContainer::search( const Data& key, Sptr < vector < Data > > & listofServers )
{
    TableIter i;

    LockHelper helper(lock);

    /// Search the container for the key
    i = groupList.find(key);

    if (i == groupList.end())
    {
        return false;
    }
    else
    {
        /// Retrieve the list of Servers for this group
        listofServers = i->second;
        return true;
    }
}


/**
 * Display a specific group
 * @param  key  Name of group to be displayed
 */
void
GroupContainer::display( const Data& key)
{
    if (cpLogGetPriority() >= LOG_DEBUG)
    {
        Sptr < vector < Data > > serverList;

        cpLog( LOG_DEBUG, "key: %s", key.logData());

        /// Find the Group in the container
        if (search(key, serverList))
        {
            //cpLog(LOG_DEBUG, "Group found");
            displayGroup( *serverList );
        }
        else
        {
            cpLog(LOG_DEBUG, "Group not found");
        }
    }
}


/**
 * print list of servers belonging to this Group 
 * @param  listofServers  List of servers to print
 */
void
GroupContainer::displayGroup( const vector < Data > & listofServers)
{
    /// print list of servers belonging to this Group
    for (vector < Data > ::const_iterator i = listofServers.begin();
            i != listofServers.end(); ++i)
    {
        cpLog( LOG_DEBUG, " Group: %s", i->logData() );
    }
}
