#ifndef IPPlanContainer_HXX
#define IPPlanContainer_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



#include <map>
#include "Plan.hxx"
#include "VMissingDataException.hxx"


static const char* const IPPlanContainerVersion = "$Id: IPPlanContainer.hxx,v 1.12 2001/02/13 18:37:05 marcl Exp $";

/**
 * The Provisioning Server is notifying us of a change in an IP Plan.
 * Obtain the plan from the PS and store it.
 * @param  data  Contents of xml file
 * @param  fileName  Name of plan file
 * @param  deleteFlag  Indicator of whether file was deleted or not.
 */
void updateIPPlan(const string& data, const string& fileName, bool deleteFlag );

/*
 *  This class is a singleton containing a map of Ip plan 
 *  objects and their names.
 *  Currently only one IP Plan will be used, but in the
 *  future we may want to handle more than one plan.
 */

class IPPlanContainer
{
    public:

        /**
         * Return the singleton IP plan container or create it if
         * it does not already exist. 
         * @return IP plan container singleton
         */
        static IPPlanContainer& instance();

        /// constructor
        IPPlanContainer();

        /// destructor
        virtual ~IPPlanContainer();

        /**
         * Add IP plan elements one at a time
         * A new plan is created if the plan is not already in the container.
         * @param  planName  Unique name of plan to which the element is to be added.
         * @param  key  Regular expression
         * @param  contacts  List of contacts
         */
        void add( const Data& planName, const Data& key, const vector < Data > & contacts);

        /**
         * Remove the specified IP plan
         * @param  planName  Identifier of plan to be removed.
         */
        void remove( const Data& planName );

        /**
         * Search plan using regular expression matching to find key.
         * Retrieve and return contact list.
         * Throw an exception if plan is not found.
         * @param  planName  Unique identifier of plan to be searched.
         * @param  key  String to be matched against regular expressions in plan.
         * @return  Contact list 
         */
        bool search( const Data& planName, const Data& key, vector < Data > & planFound);

        /**
         * Display the specified plan.
         * @param  planName  Unique identifier of plan to be displayed.
         */
        void display( const Data& planName) ;

    protected:

        /**
         * Look for and obtain a specific IP plan from the container
         * @param  planName  Identifier of plan to search for.
         * @param  workingPlan  The obtained plan.
         * @return true if found, false if not
         */
        bool getPlan( const Data& planName, Sptr<Plan> workingPlan);

        /**
         * create a new plan and store it against the given name
         * @param  planName  Unique identifier of new plan.
         * @param  workingPlan  The plan object created.
         */
        void create( const Data& planName , Sptr<Plan> workingPlan);

        /// structure for storing the IP plan
        static Sptr < IPPlanContainer > theIPContainer;

    private:

        /// Definition of a map containing names and associated plans
        typedef map < Data, Sptr < Plan > > Table;

        /// Iterator for the map
        typedef Table::iterator PlanIter;

        /// Class's instance of the map
        mutable Table IPContainer;
}
;

#endif
