
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const OpInvite_cxx_Version =
    "$Id: OpInvite.cxx,v 1.44 2002/03/19 21:39:49 jason Exp $";


#include "SipContact.hxx"
#include "InviteMsg.hxx"
#include "SipEvent.hxx"
#include "StatusMsg.hxx"
#include "OpInvite.hxx"
#include "RSStateWait.hxx"
#include "SipContactIf.hxx"
#include "RSCommon.h"
#include "RSProxy.hxx"

using namespace Vocal;

/**
 * Constructor
 */
OpInvite::OpInvite()
{ }

/**
 * Destructor
 */
OpInvite::~OpInvite()
{}

/**
 * Return an identifying name for this object
 * @return  String of characters to identify this object.
 */
const char* const
OpInvite::name() const
{
    return "OpInvite";
}

/**
 * Decode the event, trigger the process of handling the invite, and return a
 * status response containing a contact list.
 * @param  event  Sip event containing a sip message which may be an Invite.
 * @return  The next state
 */
const Sptr < State >
OpInvite::process( const Sptr < SipProxyEvent > event)
{
    Sptr < SipEvent > sipEvent;
    sipEvent.dynamicCast(event);

    if (sipEvent == 0)
    {
        cpLog(LOG_ERR, "Event is not a sipEvent");
        return PROXY_CONTINUE;
    }

    Sptr < SipMsg > sipMsg = sipEvent->getSipMsg();

    if (sipMsg == 0)
    {
        cpLog( LOG_DEBUG, 
	       "In OpInvite::process, sipEvent->getSipMsg() returned 0" );
        return PROXY_CONTINUE;
    }

    Sptr < InviteMsg > sipInvite;
    sipInvite.dynamicCast(sipMsg);

    if ( sipInvite == 0)
    {
        return PROXY_CONTINUE;
    }

    doCpLog();

    // If the rs is not active, return here
    RSSTATE rsState = RSProxy::instance().getRsState();

    Data uriUserName, uriUserParm, uri, passwd;
    Sptr <BaseUrl> baseReqUrl( sipInvite->getRequestLine().getUrl() );
    if (baseReqUrl.getPtr() != 0) 
    { 
        if (baseReqUrl->getType() == SIP_URL) 
        { 
            Sptr <SipUrl> sipUrl; 
            uriUserName = (sipUrl.dynamicCast(baseReqUrl))->getUserValue(); 
            uriUserParm = (sipUrl.dynamicCast(baseReqUrl))->getUserParam(); 
            uri = (sipUrl.dynamicCast(baseReqUrl))->getNameAddr();
            passwd =  (sipUrl.dynamicCast(baseReqUrl))->getPasswd();

            // check if there's a password in the url as it should
            // also be added to the generated contact
            if (passwd.length() > 0)
            {
                uriUserName += COLON;
                uriUserName += passwd;
            }
        } 
        else
        {
            cpLog(LOG_DEBUG, 
		  "Request Uri field is not defined as type SIP_URL");
        }
    } 
    else
    {
        cpLog(LOG_DEBUG, "Unable to obtain Req Uri field's base pointer");
    }


    Data from, fromUserName;
    Sptr <BaseUrl> baseFromUrl( sipInvite->getFrom().getUrl() );
    if (baseFromUrl.getPtr() != 0) 
    { 
        if (baseFromUrl->getType() == SIP_URL) 
        { 
            Sptr <SipUrl> sipUrl; 
            from = (sipUrl.dynamicCast(baseFromUrl))->getNameAddr();
            fromUserName = (sipUrl.dynamicCast(baseFromUrl))->getUserValue(); 
        }
        else
        {
            cpLog(LOG_DEBUG, "From field is not defined as type SIP_URL");
        }
    }
    else
    {
        cpLog(LOG_DEBUG, "Unable to obtain From field's base pointer");
    }


    /// Obtain the last two Via entries
    int viaSize = sipInvite->getNumVia();
    Data viaHostPort = sipInvite->getVia(0).getHost();
    viaHostPort += ":";
    viaHostPort += sipInvite->getVia(0).getPort();
    cpLog(LOG_DEBUG, " 1st via's host and port = %s", viaHostPort.logData());

    vector < Data > lastTwoVias;
    lastTwoVias.push_back(viaHostPort);
    if (viaSize > 1)
    {
        viaHostPort = sipInvite->getVia(1).getHost();
        viaHostPort += ":";
        viaHostPort += sipInvite->getVia(1).getPort();
        cpLog(LOG_DEBUG, " 2nd via's host and port = %s", viaHostPort.logData());
        lastTwoVias.push_back(viaHostPort);
    }

    /// Determine if To URL is a phone number or IP address
    cpLog(LOG_DEBUG, " uri user parm [%s]", uriUserParm.logData());
    bool isPhoneNumber = (uriUserParm == "phone") ? true : false;

    cpLog(LOG_DEBUG, " uri %s", uri.logData());

    int responseCode = 302;
    SipContactIf sipContactIf;
    vector < Data >  workingContacts;
    vector < Data >  finalContacts;

    if (rsState == RSINACTIVE)
    {
	cpLog(LOG_NOTICE, "returning 503: the RS is in the inactive state");
        responseCode = 503;
    }
    else
    {
        /**
         *  Preparing the CONTACT field for the Status Response
         *  message is comprised of three main parts:
         *     1) Composing the contact list 
         *     2) Locating the Via in the contact list
         *     3) Determining the next contact to return
         *
         *  1) The contact list is composed of the 
         *     From's calling + Uri's called + Uri's Term. 
         *  Look for the From in the Subscribers only.
         *  Next, search for the Uri in the Subscribers first, 
         *  then the Phone or IP lists depending upon the 
         *  flag passed in.
         *
         *  2) To determine the next contact, use the top Via 
         *  entry to find our place in the generated contact list.
         *  
         *  3) The next contact in the list needs to be validated
         *  and returned in the correct status response message.
         *
         */
        bool uriRegistered = true;

        responseCode = sipContactIf.generateContactList(fromUserName, uri, isPhoneNumber, uriRegistered, workingContacts);

        if (302 == responseCode)
        {
            vector < Data > ::iterator contactIter = workingContacts.begin();
            Data validContact = "";

            /// Loop through the recently composed contact list in search of
            /// a match on the top VIA.
            if (sipContactIf.locateVia( workingContacts, lastTwoVias, contactIter))
            {
                responseCode = sipContactIf.obtainNextContact(workingContacts, contactIter, isPhoneNumber, uriRegistered, validContact);
                if (validContact != "")
                {
                    finalContacts.push_back(validContact);
                }
            } 
            else
            {
                // Unable to determine location in contact list. Fail call.
                responseCode = 404;
            }
        }
    }

    cpLog(LOG_DEBUG, " Status Response code: [%d]", responseCode);
    /// Generate the Status Response message
    StatusMsg statusMsg( *sipInvite, responseCode );

    // The dial plans may have a generic user name which needs to
    // be replaced with the Req Uri user name.
    sipContactIf.replaceWildcard( uriUserName, finalContacts);

    string url;
    LocalScopeAllocator lo;
    for ( vector < Data > ::iterator i = finalContacts.begin(); 
	  i != finalContacts.end(); ++i )
    {
        cpLog(LOG_DEBUG, " ------> Validated CONTACT: [%s]", i->logData());
        url = i->getData(lo);
        Sptr<SipUrl> sipUrl = new SipUrl(url);
        SipContact sipContact;
        sipContact.setUrl(sipUrl);
        statusMsg.setContact( sipContact );
    }

    sipEvent->getSipStack()-> sendReply( statusMsg );

    return PROXY_CONTINUE;
}

// End of File
