#ifndef Plan_HXX
#define Plan_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const Plan_hxx_Version =
    "$Id: Plan.hxx,v 1.7 2001/02/13 18:37:08 marcl Exp $";


#include "PlanElement.hxx"
#include "VMissingDataException.hxx"
#include "VRWLock.hxx"

/**
 *  This class defines a plan of elements to be used to
 *  resolve a phone number or IP address. 
 *  The regular expression form of an element's key will be 
 *  compared against a SIP address. If the key matches a 
 *  portion of it, then the element's contacts will be retrieved.
 *
 * Assumptions:
 *  o The PS will deliver the plan in a particular order.
 *    That order must be maintained for searching purposes.
 *  o When the plan is changed, the whole plan is sent.
 */
class Plan
{
    public:

        /// constructor
        Plan();

        /// destructor
        ~Plan();

        /**
         * Create a new entry by adding it to the bottom of the list
         * Order is important when it comes time to search through the list.
         * @param  key  Regular expression
         * @param  contacts  Contact list 
         */
        void addEntry( const Data& key, const  vector < Data > & contacts );

        /**
         * Delete the specified key and its related contacts from the container.
         * @param  key  Regular expression to be deleted.
         */
        void deleteEntry( const Data& key );

        /**
         * Remove the whole plan.
         * This is done when a change is made in Provisioning.
         */
        void clear();

        /**
         * Search table using regexp matching. Retrieve and return a contact list
         * @param  key  String to match against the regular expressions in the table.
         * @param  elementFound List of contacts corresponding to expression matched.
         * @return True if found, false if not.
         */
        bool search( const Data& key, vector < Data > & elementFound);

        /**
         * Display all the stored expressions and their related contacts.
         */
        void display();

    protected:

        /**
         *  Structure for storing the dial plan
         *  Defined in this way due to the search needing
         *  to be done in a particular order defined by the PS.
         */
        vector < PlanElement > itsPlan;

    private:

        VRWLock PlanMutex;

};

#endif
