
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const RSServerContainer_cxx_Version =
    "$Id: RSServerContainer.cxx,v 1.49 2002/04/11 23:47:41 bko Exp $";
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "ProvisionInterface.hxx"
#include "VRedirectServerData.hxx"
#include "RSServerContainer.hxx"
#include "GroupContainer.hxx"
#include "ReadLockHelper.hxx"
#include "WriteLockHelper.hxx"
#include "ServerWithGroup.hxx"
#include "HeartbeatUtil.h"
#include "RSProxy.hxx"
#include "VIoException.hxx"
#include "CommandLine.hxx"
#include "SystemInfo.hxx"

using namespace Vocal;

static const string RSfilename("ListOfRedirectServers");
static const string MSfilename("ListOfMarshalServers");
static const string FSfilename("ListOfFeatureServers");
static const string JSfilename("ListOfJtapiServers");
static const string GlobConfigfilename("GlobalConfigData");

Sptr < RSServerContainer > RSServerContainer::theRSSrvContainer = 0;

/**
 * Return the one RS Server container or create it if it doesn't already exist. 
 * @return  Reference to the singleton
 */
RSServerContainer&
RSServerContainer::instance()
{
    if ( theRSSrvContainer == 0 )
    {
        theRSSrvContainer = new RSServerContainer;
    }
    return *theRSSrvContainer;
}


/**
 * Constructor
 */
RSServerContainer::RSServerContainer()
        : BaseServerContainer()
{
    /// get config data from PS
    configRedirectServers();
    configMarshalServers();
    configFeatureServers();
    configJtapiServers();

    /// create instance of Group Container so that we can get all the
    /// feature and marshal groups from PS.
    GroupContainer::instance();

    displayLists();

    /// register interest in the following files

    try
    {
        cpLog(LOG_DEBUG, "registering for update with PS");
        ProvisionInterface::instance().registerForUpdate( updateRSRedirectServers, RSfilename,
                               ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to register for update with Provisioning Server for %s, reason %s", RSfilename.c_str(), e.getDescription().c_str());
    }

    try
    {
        ProvisionInterface::instance().registerForUpdate( updateMarshalServers, MSfilename,
                               ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to register for update with Provisioning Server for %s, reason %s", MSfilename.c_str(), e.getDescription().c_str());
    }

    try
    {
        ProvisionInterface::instance().registerForUpdate( updateFeatureServers, FSfilename,
                               ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to register for update with Provisioning Server for %s, reason %s", FSfilename.c_str(), e.getDescription().c_str());
    }

    try
    {
        ProvisionInterface::instance().registerForUpdate( updateJtapiServers, JSfilename,
                               ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to register for update with Provisioning Server for %s, reason %s", JSfilename.c_str(), e.getDescription().c_str());
    }

    try
    {
        ProvisionInterface::instance().registerForUpdate( updateRSHeartbeatParms, GlobConfigfilename,
                               ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to register for update with Provisioning Server for %s, reason %s", GlobConfigfilename.c_str(), e.getDescription().c_str());
    }

}


/**
 * Get Servers of the specified type from Provisioning   
 * @param  type  Type of servers requested
 * @param  retList  List of servers of the requested type
 */
void
RSServerContainer::getServersFromPS( const ServerType type, vector < Data > & retList)
{
    list < string > tmpList;

    try
    {
        /// get instance of Provisioning Server
        ProvisionInterface& pif = ProvisionInterface::instance();


        switch (type)
        {
            case SERVER_RS:
            {
                pif.getRedirectServers( tmpList );
            }
            break;
            case SERVER_JS:
            {
                pif.getJtapiServers( tmpList );
            }
            break;
            case SERVER_FS:
            {
                pif.getFeatureServers( tmpList );
            }
            break;
            case SERVER_MS:
            {
                pif.getMarshalServers( tmpList );
            }
            break;
            default:
            {
                cpLog(LOG_ALERT, "Unknown server type");
            }
            break;
        }
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
        return;
    }

    string serverInList;
    bool duplicateEntryFound = false;
    /// add server to vector
    for (list < string > ::iterator i = tmpList.begin(); i != tmpList.end(); ++i)
    {
        /// check for duplicate entries in list
        duplicateEntryFound = false;
        LocalScopeAllocator lo;
        for (vector < Data > ::iterator j = retList.begin(); j != retList.end(); ++j)
        {
            serverInList = string(j->getData(lo) );
            if ( *i == serverInList )
            {
                duplicateEntryFound = true;
                break;
            }
        }

        if (duplicateEntryFound)
        {
            cpLog( LOG_ALERT, "*** Duplicate Entry Found In LIST: %s ***", i->c_str() );
            cpLog( LOG_ALERT, "*** DO NOT add to Container ***");
        }
        else
        {
            Data addressPort(*i);
            cpLog( LOG_DEBUG, "ip:port=%s", addressPort.logData());
            retList.push_back( addressPort );
        }
    }
}


/**
 * Remove Duplicates entries in the List of Servers passed in.   
 * @param  list  list of servers to be added to server container.  
 */
void
RSServerContainer::removeDuplicatesFromList( Sptr < vector < Data > > list )
{
    ReadLockHelper tableHelper(tableLock);

    bool done = false;
    bool duplicateEntryFound = false;
    while (!done)
    {
        TableIter iter;

        duplicateEntryFound = false;
        for (vector < Data > ::iterator i = list->begin();
                i != list->end(); ++i)
        {
            iter = serverList.find(*i);
            if (iter != serverList.end())
            {
                cpLog( LOG_ALERT, "*** Duplicate Entry Found in CONTAINER: %s ***", i->logData() );
                cpLog( LOG_ALERT, "*** DO NOT add to Container ***");
                duplicateEntryFound = true;
                list->erase(i);
                break;
            }
        }
        if (!duplicateEntryFound) 
        {
            done = true;
        }
    }
}


/**
 * Configure Redirect Servers    
 */
void
RSServerContainer::configRedirectServers()
{
    WriteLockHelper rsHelper(rsLock);

    /// get RSlist
    itsRSList = new vector < Data > ;
    getServersFromPS( SERVER_RS, *itsRSList );

    removeDuplicatesFromList(itsRSList);

    /// create Redirect Servers
    for (vector < Data > ::iterator i = itsRSList->begin();
            i != itsRSList->end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new RS: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        addRSRedirectServer( smartServer );

        /// display updated data
        display( smartServer );
    }

}


/**
 * Configure Marshal Servers    
 */
void
RSServerContainer::configMarshalServers()
{
    WriteLockHelper msHelper(msLock);

    /// get MSlist
    itsMSList = new vector < Data > ;
    getServersFromPS( SERVER_MS, *itsMSList );

    removeDuplicatesFromList(itsMSList);

    /// create Marshal Servers
    LocalScopeAllocator lo;
    for (vector < Data > ::iterator i = itsMSList->begin();
            i != itsMSList->end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new MS: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        /// get server group list
        Sptr < vector < Data > > smartGroup = new vector < Data > ;

        HostToGroupMap msGroupMap;

        try
        {
            ProvisionInterface::instance().getMarshalHostToGroupMap(msGroupMap);
        }
        catch (VException& e)
        {
            cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
            cpLog( LOG_DEBUG, "Skipping MS: %s", i->logData() );
            continue;
        }

        Data groupData( msGroupMap[smartServer.getData(lo)].c_str() );
        smartGroup->push_back(groupData);

        /// create new Marshal
        addServerWithGroup( smartServer, smartGroup, SERVER_MS );

        /// display updated data
        display( smartServer );
    }
}


/**
 * Configure Feature Servers    
 */
void
RSServerContainer::configFeatureServers()
{
    WriteLockHelper fsHelper(fsLock);

    /// get FSlist
    itsFSList = new vector < Data > ;
    getServersFromPS( SERVER_FS, *itsFSList );

    removeDuplicatesFromList(itsFSList);

    /// create Feature Servers
    LocalScopeAllocator lo;
    for (vector < Data > ::iterator i = itsFSList->begin();
            i != itsFSList->end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new FS: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        /// get server group list
        Sptr < vector < Data > > smartGroup = new vector < Data > ;

        HostToGroupMap fsGroupMap;

        try
        {
            ProvisionInterface::instance().getFeatureHostToGroupMap(fsGroupMap);
        }
        catch (VException& e)
        {
            cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
            cpLog( LOG_DEBUG, "Skipping FS: %s", i->logData() );
            continue;
        }

        Data groupData( fsGroupMap[smartServer.getData(lo)].c_str() );
        smartGroup->push_back(groupData);

        /// create new Feature
        addServerWithGroup( smartServer, smartGroup, SERVER_FS );

        /// display updated data
        display( smartServer );
    }
}


/**
 * Configure Jtapi Servers    
 */
void
RSServerContainer::configJtapiServers()
{
    WriteLockHelper jsHelper(jsLock);

    /// get JSlist
    itsJSList = new vector < Data > ;
    getServersFromPS( SERVER_JS, *itsJSList );

    removeDuplicatesFromList(itsJSList);

    /// create Jtapi Servers
    LocalScopeAllocator lo;
    for (vector < Data > ::iterator i = itsJSList->begin();
            i != itsJSList->end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new JS: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        /// get server group list
        Sptr < vector < Data > > smartGroup = new vector < Data > ;

        HostToGroupMap jsGroupMap;

        try
        {
            ProvisionInterface::instance().getJtapiHostToGroupMap(jsGroupMap);
        }
        catch (VException& e)
        {
            cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
            cpLog( LOG_DEBUG, "Skipping JS: %s", i->logData() );
            continue;
        }

        Data groupData( jsGroupMap[smartServer.getData(lo)].c_str() );
        smartGroup->push_back(groupData);

        /// create new Jtapi
        addServerWithGroup( smartServer, smartGroup, SERVER_JS );

        /// display updated data
        display( smartServer );
    }
}


/**
 * Destructor    
 */
RSServerContainer::~RSServerContainer( )
{}


/**
 * Get List of Redirect Servers   
 * @param  listofRS  list of all redirect servers in RSServerontainer.  
 */
void
RSServerContainer::getListofRS( Sptr < vector < Data > >  listofRS )
{
    cpLog( LOG_DEBUG_HB, "RSServerContainer::getListofRS");

    ReadLockHelper rsHelper(rsLock);

    for (vector < Data > ::iterator i = itsRSList->begin();
            i != itsRSList->end(); ++i)
    {
        listofRS->push_back(*i);
    }
}


/**
 * Get List of Active Servers of a certain type  
 * @param  type  the type of server to search for   
 * @param  listofServers  list of all servers in container that matches the type.  
 */
void
RSServerContainer::getListofActive( const ServerType type, Sptr < vector < Data > > &  listofServers )
{
    // function not used in the RS at this time.
}


/**
 * Determine if specified server is a valid server configured in the system 
 * @param  key  Server's address   
 * @param  type  type of server.  
 * @param  listofGroups  list of groups this server belongs to.  
 */
bool
RSServerContainer::isValid( const Data& key, ServerType& type, vector < Data > & listofGroups)
{
    Sptr < Server > serverPtr;

    type = SERVER_Unknown;

    if (search( key, serverPtr) )
    {
        Sptr < ServerWithGroup > serverWithGroupPtr;

        ReadLockHelper tableHelper(tableLock);

        type = serverPtr->getType();
        /// get list of Groups from server object
        switch (type)
        {
            case SERVER_JS:
            case SERVER_FS:
            case SERVER_MS:
            {
                serverWithGroupPtr.dynamicCast(serverPtr);
                serverWithGroupPtr->getGroups(listofGroups);
                break;
            }
            default:
            {
                listofGroups.clear();
            }
            break;
        };

        return true;
    }
    else
    {
        cpLog( LOG_DEBUG_HB, "Server not in system");
        listofGroups.clear();
        return false;
    }
}


/**
 *  get the sync port for a particular RS
 *  @param  key  Identifier of server
 *  @return  Port number for specified server
 */
unsigned short
RSServerContainer::getRSSyncPort( const Data& key )
{
    TableIter i;

    ReadLockHelper tableHelper(tableLock);

    /// Search the container for the key
    i = serverList.find(key);

    if (i == serverList.end())
    {
        return 0;
    }
    else
    {
        /// Retrieve the Server
        if (SERVER_RS == i->second->getType())
        {
            Sptr < RSRedirectServer > server;
            server.dynamicCast(i->second);
            return server->getSyncPort();
        }
        else
        {
            return 0;
        }
    }
}


/**
 * Add a server object that has groups associated with the server.   
 * @param  key  Server's address   
 * @param  groups  all the groups this servers belongs to.  
 * @param  serverType  type of server.  
 */
void
RSServerContainer::addServerWithGroup( const Data& key,
                                       const Sptr < vector < Data > > groups,
                                       const ServerType serverType )
{
    cpLog( LOG_DEBUG, "RSServerContainer::addServerWithGroup %s",
           convertToString(serverType) );
    /// create a Server with 'new'
    Sptr < ServerWithGroup > server = new ServerWithGroup(serverType);

    /// fill in the fields
    server->setGroups(groups);

    WriteLockHelper tableHelper(tableLock);

    /// insert the Server in the map
    //serverList.insert(pair< Sptr<Data>, Sptr<Server> >(key, server));
    serverList[key] = server;
    if(!CommandLine::instance()->getInt("HEARTBEAT"))
    {
        cpLog( LOG_INFO, "Heartbeat mechanism is disabled, setting server (%s) to be active", key.logData());
	server->processHeartbeat(key); // need to add to server groups
        server->setStatus(Active);
        server->displayStatus(key);
    }
}


/**
 * Add a redirect server object.   
 * @param  key  Server's address   
 */
void
RSServerContainer::addRSRedirectServer( const Data& key)
{
    cpLog( LOG_DEBUG, "entering RSServerContainer::addRSRedirectServer");

    /// Obtain the Sync Port from provisioning
    VRedirectServerData myRS;
    LocalScopeAllocator lo;
    try
    {
        ProvisionInterface::instance().getRedirectData(key.getData(lo), myRS);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
        return;
    }

    unsigned short syncPort = myRS.getSyncPort();

    /// create a Server with 'new'
    Sptr < RSRedirectServer > server = new RSRedirectServer(syncPort);

    WriteLockHelper tableHelper(tableLock);

    /// insert the Server in the map
    serverList[key] = server;
    if(!CommandLine::instance()->getInt("HEARTBEAT"))
    {
        cpLog( LOG_INFO, "Heartbeat mechanism is disabled, setting server (%s) to be active", key.logData());
        server->setStatus(Active);
    }
}


/**
 * Deal with a heartbeat event from the given server
 * @param  key  Server who sent the heartbeat
 * @return  True if server is recognized, false if not
 */
bool
RSServerContainer::processHeartbeat( const Data& key)
{
    Sptr < Server > serverPtr;

    if (search( key, serverPtr) )
    {

        ReadLockHelper tableHelper(tableLock);

        Sptr < ServerWithGroup > serverWithGroupPtr;

        /// print data contained in the server object
        switch (serverPtr->getType() )
        {
            case SERVER_JS:
            case SERVER_FS:
            case SERVER_MS:
            {
                serverWithGroupPtr.dynamicCast(serverPtr);
                serverWithGroupPtr->processHeartbeat(key);
            }
            break;
            case SERVER_RS:
            {
                serverPtr->processHeartbeat(key);
            }
            break;
            default:
            break;
        };

        if (heartbeatDebug)
        {
            displayServer(*serverPtr);
        }
        return true;
    }
    else
    {
        return false;
    }
}


/**
 * Check all servers to see if they are still actively heartbeating
 */
void
RSServerContainer::doHouseKeeping()
{
    Data serverId;
    Sptr < Server > serverPtr;

    cpLog( LOG_DEBUG_HB, "");
    cpLog( LOG_DEBUG_HB, "****** do HouseKeeping *****");

    ReadLockHelper tableHelper(tableLock);

    if ( !serverList.empty() )
    {
        Sptr < ServerWithGroup > serverWithGroupPtr;

        for ( TableIter i = serverList.begin(); i != serverList.end(); ++i)
        {
            serverId = i->first;
            serverPtr = i->second;
            //cpLog( LOG_DEBUG, "%s", serverId->logData() );
            if ( serverPtr->getStatus() != Inactive )
            {
                cpLog( LOG_DEBUG_HB, "%s", serverId.logData() );
            }
            switch (serverPtr->getType() )
            {
                case SERVER_JS:
                case SERVER_FS:
                case SERVER_MS:
                {
                    serverWithGroupPtr.dynamicCast(serverPtr);
                    serverWithGroupPtr->doHouseKeeping(serverId);
                }
                break;
                case SERVER_RS:
                {
                    serverPtr->doHouseKeeping(serverId);
                }
                break;
                default:
                break;
            }
        }
    }
    else
    {
        cpLog( LOG_DEBUG_HB, "nothing in server container");
    }

    cpLog( LOG_DEBUG_HB, "****** finished HouseKeeping *****");
    cpLog( LOG_DEBUG_HB, "");
}



/**
 * A change was made to the RS's in the provisioning server. 
 * Delete the stored RS server list and recreate it with the new data.
 */
void
RSServerContainer::updateRSsFromPS()
{

    {
        WriteLockHelper rsHelper(rsLock);

        /// delete all RSs in Server Container
        for (vector < Data > ::iterator i = itsRSList->begin();
                i < itsRSList->end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete RS: %s", i->logData() );
            Data server(*i);
            /// delete from Server Container
            remove(server);
        }

        /// clear RSList
        itsRSList->clear();
    }

    /// reconfigure RSs from scratch
    configRedirectServers();

    displayLists();
}


/**
 * A change was made to the MS's in the provisioning server. 
 * Delete the stored MS server list and recreate it with the new data.
 */
void
RSServerContainer::updateMSsFromPS()
{

    {
        WriteLockHelper msHelper(msLock);

        /// delete all MSs in Server Container
        for (vector < Data > ::iterator i = itsMSList->begin();
                i < itsMSList->end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete MS: %s", i->logData() );
            Data key(*i);

            /// Search the container for the key
            Sptr < Server > serverPtr;
            if ( search( key, serverPtr ) )
            {
                /// if active, remove from activeList in GroupContainer
                if ( serverPtr->getStatus() == Active )
                {
                    Sptr < ServerWithGroup > serverWithGroupPtr;
                    serverWithGroupPtr.dynamicCast(serverPtr);
                    serverWithGroupPtr->processServerDown(key);
                }

                /// delete from Server Container
                remove(key);
            }
        }

        /// clear MSList
        itsMSList->clear();
    }

    /// reconfigure MSs from scratch
    configMarshalServers();

    /// update Valid Group List
    GroupContainer& grpContainer = GroupContainer::instance();
    grpContainer.handleUpdateFromPS();

    displayLists();
}


/**
 * A change was made to the FS's in the provisioning server. 
 * Delete the stored FS server list and recreate it with the new data.
 */
void
RSServerContainer::updateFSsFromPS()
{

    {
        WriteLockHelper fsHelper(fsLock);

        /// delete all FSs in Server Container
        for (vector < Data > ::iterator i = itsFSList->begin();
                i < itsFSList->end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete FS: %s", i->logData() );
            Data key(*i);

            /// Search the container for the key
            Sptr < Server > serverPtr;
            if ( search( key, serverPtr ) )
            {
                /// if active, remove from activeList in GroupContainer
                if ( serverPtr->getStatus() == Active )
                {
                    Sptr < ServerWithGroup > serverWithGroupPtr;
                    serverWithGroupPtr.dynamicCast(serverPtr);
                    serverWithGroupPtr->processServerDown(key);
                }

                /// delete from Server Container
                remove(key);
            }
        }

        /// clear FSList
        itsFSList->clear();
    }

    /// reconfigure FSs from scratch
    configFeatureServers();

    /// update Valid Group List
    GroupContainer& grpContainer = GroupContainer::instance();
    grpContainer.handleUpdateFromPS();

    displayLists();
}


/**
 * A change was made to the JS's in the provisioning server. 
 * Delete the stored JS server list and recreate it with the new data.
 */
void
RSServerContainer::updateJSsFromPS()
{

    {
        WriteLockHelper jsHelper(jsLock);

        /// delete all JSs in Server Container
        for (vector < Data > ::iterator i = itsJSList->begin();
                i < itsJSList->end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete JS: %s", i->logData() );
            Data key(*i);

            /// Search the container for the key
            Sptr < Server > serverPtr;
            if ( search( key, serverPtr ) )
            {
                /// if active, remove from activeList in GroupContainer
                if ( serverPtr->getStatus() == Active )
                {
                    Sptr < ServerWithGroup > serverWithGroupPtr;
                    serverWithGroupPtr.dynamicCast(serverPtr);
                    serverWithGroupPtr->processServerDown(key);
                }

                /// delete from Server Container
                remove(key);
            }
        }

        /// clear JSList
        itsJSList->clear();
    }

    /// reconfigure JSs from scratch
    configJtapiServers();

    /// update Valid Group List
    GroupContainer& grpContainer = GroupContainer::instance();
    grpContainer.handleUpdateFromPS();

    displayLists();
}


/**
 * A change was made to the Global Config Parms Data on the provisioning server. 
 * Delete the stored parameter(s) and replace it with new values.
 */
void
RSServerContainer::updateHBParmsFromPS()
{
    cpLog( LOG_DEBUG, "updating Heartbeat Parms in RSServerContainer");

    WriteLockHelper tableHelper(tableLock);

    /// get pointer to global config object
    VGlobalConfigData globalConfigData;

    try
    {
        ProvisionInterface::instance().getConfigData(globalConfigData);
    }
    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed to get data from Provisioning Server, reason %s", e.getDescription().c_str());
        return;
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
        return;
    }

    int newValue = globalConfigData.getMaxMissedHB();
    int oldValue = serverList.begin()->second->getMaxMissedHB();

    cpLog( LOG_DEBUG, "maxMissedHB old: %d new: %d", oldValue, newValue );

    if (( !serverList.empty() ) && ( oldValue != newValue ))
    {
        Data serverId;
        Sptr < Server > serverPtr;

        for ( TableIter i = serverList.begin(); i != serverList.end(); ++i)
        {
            serverId = i->first;
            i->second->setMaxMissedHB( newValue );
        }
    }
}


/**
 * Display a server given the pointer to a server object  
 * @param  serverPtr  pointer to server in server container   
 */
void
RSServerContainer::displayServer( const Server& serverPtr)
{
    if (cpLogGetPriority() >= LOG_DEBUG_HB)
    {

        Sptr < ServerWithGroup > serverWithGroupPtr;
        vector < Data > groupPtr;

        /// print data contained in the server object
        cpLog( LOG_DEBUG_HB, "------------------------");
        ServerType type = serverPtr.getType();
        cpLog( LOG_DEBUG_HB, "ServerType: %s", convertToString(type) );

        switch ( type )
        {
            case SERVER_RS:
            {
                const RSRedirectServer* sptr = static_cast < const RSRedirectServer* > (&serverPtr);
                assert(sptr);
                cpLog( LOG_DEBUG_HB, "Sync Port: %d", sptr->getSyncPort() );
            }
            break;
            case SERVER_JS:
            case SERVER_FS:
            case SERVER_MS:
            {
                const ServerWithGroup* sptr = static_cast < const ServerWithGroup* > (&serverPtr);
                assert(sptr);
                sptr->getGroups(groupPtr);

                /// print all feature groups this server belongs to
                for (vector < Data > ::iterator i = groupPtr.begin();
                        i != groupPtr.end(); ++i)
                {
                    cpLog( LOG_DEBUG_HB, "Group: %s", i->logData() );
                }
            }
            break;
            default:
            break;
        };

        cpLog( LOG_DEBUG_HB, "Status: %s", convertToString( serverPtr.getStatus() ) );
        cpLog( LOG_DEBUG_HB, "HB Count: %d  #Missed: %d", serverPtr.getHeartbeatCount(),
               serverPtr.getNumOfHeartbeatMissed() );
        cpLog( LOG_DEBUG_HB, "------------------------");
    }

}


/**
 * Display list of servers by type 
 */
void
RSServerContainer::displayLists()
{
    if (cpLogGetPriority() >= LOG_DEBUG)
    {
        ReadLockHelper jsHelper(jsLock);
        ReadLockHelper fsHelper(fsLock);
        ReadLockHelper msHelper(msLock);
        ReadLockHelper rsHelper(rsLock);

        cpLog( LOG_DEBUG, "----------- Jtapi Servers ---------");
        for (vector < Data > ::iterator i = itsJSList->begin();
                i < itsJSList->end(); ++i)
        {
            cpLog( LOG_DEBUG, "JS: %s", i->logData() );
        }
    
        cpLog( LOG_DEBUG, "----------- Feature Servers ---------");
        for (vector < Data > ::iterator i = itsFSList->begin();
                i < itsFSList->end(); ++i)
        {
            cpLog( LOG_DEBUG, "FS: %s", i->logData() );
        }

        cpLog( LOG_DEBUG, "----------- Marshal Servers ---------");
        for (vector < Data > ::iterator i = itsMSList->begin();
                i < itsMSList->end(); ++i)
        {
            cpLog( LOG_DEBUG, "MS: %s", i->logData() );
        }

        cpLog( LOG_DEBUG, "----------- Redirect Servers --------");
        for (vector < Data > ::iterator i = itsRSList->begin();
                i < itsRSList->end(); ++i)
        {
            cpLog( LOG_DEBUG, "RS: %s  Sync Port: %d", i->logData(), getRSSyncPort(*i));
        }
    }
}


/**
 * The Provisioning Server is notifying us of a change in RS's.
 */
void
updateRSRedirectServers(const string& data, const string& fileName, bool)
{
    cpLog(LOG_DEBUG, "updateRSRedirectServers - Received update from PS");
    /// Update the sync port for this RS
    Data localHostPort(theSystem.gethostAddress());
    localHostPort += ":";
    localHostPort += RSProxy::instance().getLocalSipPort();

    unsigned short oldSyncPort(RSServerContainer::instance().getRSSyncPort(localHostPort));

    /// update RS list
    RSServerContainer::instance().updateRSsFromPS();

    unsigned short newSyncPort(RSServerContainer::instance().getRSSyncPort(localHostPort));

    if (oldSyncPort != newSyncPort)
    {
        cpLog(LOG_DEBUG, "Updating Sync port from [%d] to [%d] for this RS [%s]", oldSyncPort, newSyncPort, localHostPort.logData());
        RSProxy::instance().updateRedundancyPort(oldSyncPort, newSyncPort);
    }
}


/**
 * The Provisioning Server is notifying us of a change in MS's.
 */
void
updateMarshalServers(const string& data, const string& fileName, bool)
{
    /// update MS list
    RSServerContainer::instance().updateMSsFromPS();
}


/**
 * The Provisioning Server is notifying us of a change in FS's.
 */
void
updateFeatureServers(const string& data, const string& fileName, bool)
{
    /// update FS list
    RSServerContainer::instance().updateFSsFromPS();
}


/**
 * The Provisioning Server is notifying us of a change in JS's.
 */
void
updateJtapiServers(const string& data, const string& fileName, bool)
{
    /// update JS list
    RSServerContainer::instance().updateJSsFromPS();
}


/**
 * The Provisioning Server is notifying us of a change in Heartbeat Parms.
 */
void
updateRSHeartbeatParms(const string& data, const string& fileName, bool)
{
    /// update Heartbeat Parms
    RSServerContainer::instance().updateHBParmsFromPS();
}

// End of File
