#ifndef REDUNDANCY_WORKER_THREAD_HXX
#define REDUNDANCY_WORKER_THREAD_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


#include "Fifo.h"
#include "Sptr.hxx"
#include "ThreadIf.hxx"
#include "SipProxyEvent.hxx"
#include "RSTimerEvent.hxx"
#include "RSServerContainer.hxx"
#include "RedundancyEvent.hxx"
#include "RSTimerEvent.hxx"


static const char* const RedundancyWorkerThreadVersion = "$Id$";

/** RedundancyWorkerThread is a ThreadIf.  
 * It has an input queue on which it is blocked.
 * The input queue contains events.
 */

class RedundancyWorkerThread: public ThreadIf
{
    public:

        /**
         * Constructor  
         * @param  inputFifo 
         * @param  udpStack 
         * @param  sipPort  The port on which this server receives SIP messages.
         * @param  redundancyPort  The port on which this server receives redundancy messages.
         */
        RedundancyWorkerThread( const Sptr < Vocal::Fifo < Sptr < Vocal::SipProxyEvent > > > inputFifo, 
				Sptr < UdpStack > udpStack, 
				unsigned short sipPort,
				unsigned short redundancyPort );

        /// Destructor
        virtual ~RedundancyWorkerThread();

        /**
         * Obtain priority of this thread.
         * @return  Thread's priority
         */
        int getThreadPriority()
        {
            return ThreadIf::itsThread.getPriority();
        }

        /// Definitions of the available redundancy states.
        enum RedundancyStates { Unsynchronized ,
                                WaitingForSyncReqAck,
                                SyncReqAckReceived,
                                Synchronized };

        /**
         * Update the Redundancy port in response to a change in provisioning.
         * @param  oldSyncPort  Value of previously stored sync port
         * @param  newSyncPort  Value of sync port to be stored.
         */
        void setRedundancyPort(unsigned short oldSyncPort, unsigned short newSyncPort);

    protected:
        /**
         * Process Redundancy related events
         * The two main types are messages and timers.
         */
        virtual void thread();

        /**
         * Compose and send a redundancy message via UDP to the specified server.
         * @param  server  Address of destination server  
         * @param  msgType  Type of message to be sent.
         * @param  eventId  Unique identifier of event.
         * @return  True if message was sent, false otherwise
         */
        bool sendRedundancyMsg( const Data & server, 
                                const InfoType msgType, 
                                const Vocal::FifoEventId eventId);

        /**
         * Begin the process of synchronizing subscriber registration data
         * by finding an active RS and sending a Sync Request to it.
         */
        void beginSync();

        /**
         * Obtain an active RS. Allow a number of retries on the same RS before 
         * looking for an alternate active RS.
         * @param  activeRS  Address of a currently active Redirect Server.
         * @return  True if an active RS was found, false otherwise.
         */
        bool getActiveRS( Data & activeRS );

        /**
         * Activate RS by changing its state and have it begin heartbeating.
         */
        void activateRS();

    private:
        /// Local fifo
        Sptr < Vocal::Fifo < Sptr < Vocal::SipProxyEvent > > > itsFifo;
        /// Local stack for Udp messages
        Sptr < UdpStack > itsUdpStack;

        /// The Redundancy state of this RS.
        RedundancyStates itsRedundancyState;

        /// The number of times the chosen RS has been attempted to be synched with.
        unsigned int itsRetryCounter;

        /// Local timers
        Vocal::FifoEventId itsSyncRequestTimerId;
        Vocal::FifoEventId itsSynchronizingTimerId;

        /// List of RS's attempted to be synched with.
        Sptr < vector < Data > > itsListOfRSTried;

        /// Local port to receive SIP messages.
        unsigned short itsSipPort;
        /// Local port to recieve redundancy messages.
        unsigned short itsRedundancyPort;

        /// process the redundancy event
        void processRedundancyEvent(Sptr <RedundancyEvent> redundancyEvent);

        /// process the timer event
        void processTimerEvent(Sptr <RSTimerEvent> rsTimerEvent);

};


#endif //REDUNDANCY_WORKER_THREAD_HXX
