
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const ServerWithGroup_cxx_Version =
    "$Id: ServerWithGroup.cxx,v 1.16 2002/02/25 05:51:20 sprajpat Exp $";
#include "cpLog.h"
#include "ServerWithGroup.hxx"
#include "GroupContainer.hxx"
#include "HeartbeatUtil.h"

/**
 * Constructor
 * @param  serverType  Type of server
 */
ServerWithGroup::ServerWithGroup( const ServerType serverType) : Server(serverType)
{
    itsGroups = new vector < Data > ;
}


/**
 * Copy Constructor
 * @param  rhs  Object from which to copy
 */
ServerWithGroup::ServerWithGroup( const ServerWithGroup& rhs ) : Server(rhs)
{
    if (this != &rhs)
    {
        itsGroups = new vector < Data > ;
        copyRhsToThis( rhs );
    }
}

/**
 * Operator overload for "="
 * @param  rhs  Object from which to copy
 * @return  Reference to this object with new data
 */
ServerWithGroup&
ServerWithGroup::operator=( const ServerWithGroup& rhs )
{
    if ( this != &rhs )
    {
        Server::operator= (rhs);
        copyRhsToThis( rhs );
    }
    return *this;
}


/**
 * Copy member data from given object into this one
 * @param  rhs  Object from which to copy
 */
void
ServerWithGroup::copyRhsToThis( const ServerWithGroup& rhs )
{
    *itsGroups = *(rhs.itsGroups);
}


/**
 * Destructor
 */
ServerWithGroup::~ServerWithGroup()
{ }


/**
 * Obtain the group data
 * @param  groups  Structure to hold the groups to be returned
 */
void
ServerWithGroup::getGroups(vector < Data > & groups) const
{
    groups = *itsGroups;
}

/**
 * Store the group data
 * @param  groups  Groups to be stored
 */
void
ServerWithGroup::setGroups(const Sptr < vector < Data > > & groups )
{
    itsGroups = groups;
}

/**
 * Process a heartbeat according to the current status
 * @param  key  Server who sent the heartbeat
 */
void
ServerWithGroup::processHeartbeat( const Data& key )
{
    incrHeartbeatCountByOne();

    if (itsStatus == Inactive)
        gotFirstHeartbeat(key);
    else
        if (itsStatus == ActivePending )
            gotHiccup();

}

/**
 * Process the initial heartbeat from a server by adding it to the active list
 * @param  key  Server who sent the heartbeat
 */
void
ServerWithGroup::gotFirstHeartbeat( const Data& key )
{
    Server::gotFirstHeartbeat(key);

    /// get instance of group container.
    GroupContainer& groupContainer = GroupContainer::instance();

    /// add server to active list
    for (vector < Data > ::iterator i = itsGroups->begin();
            i != itsGroups->end(); ++i)
    {
        /// add server to active list
        Data group(*i);
        groupContainer.addServerToGroup( group, key );
    }
}

/**
 * Check to see if server is still active. Mark a server as no longer
 * active if it has reached it's max missed heartbeats.
 * @param  serverId  Server on which to perform housekeeping
 */
void
ServerWithGroup::doHouseKeeping( const Data& serverId)
{
    if ( itsStatus != Inactive )
        if ( heartbeatCount == 0 )
        {
            incrNumOfHeartbeatMissed();

            /// *** keep this code around in case we decide
            /// *** to reinstate the ActivePending state.
            ///if ( itsStatus == Active)
            ///   setStatus(ActivePending);

            ///cpLog( LOG_DEBUG,"missed %d HBs", numOfHeartbeatMissed);
            if ( serverIsDead() )
            {
                cpLog( LOG_ALERT, "\n");
                cpLog( LOG_ALERT, "** %s Down: %s **", convertToString(itsType),
                       serverId.logData() );
                processServerDown(serverId);
            }
        }
        else
        {
            /// reset heartbeat counters
            clearHeartbeatCount();
            clearNumOfHeartbeatMissed();
        }
}

/**
 * Perform actions to mark a server as no longer active.
 * @param  serverId  Server to be marked inactive
 */
void
ServerWithGroup::processServerDown( const Data& serverId)
{
    setStatus(Inactive);

    /// reset heartbeat counters
    clearHeartbeatCount();
    clearNumOfHeartbeatMissed();

    /// get instance of group container.
    GroupContainer& groupContainer = GroupContainer::instance();

    /// remove server from active list
    for (vector < Data > ::iterator i = itsGroups->begin();
            i != itsGroups->end(); ++i)
    {
        Data group(*i);
        groupContainer.removeServerFromGroup( group, serverId );
    }
}
