
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const SubscriberContainer_cxx_Version =
    "$Id: SubscriberContainer.cxx,v 1.64 2002/03/19 21:39:49 jason Exp $";

#include "cpLog.h"
#include "ProvisionInterface.hxx"
#include "RegisterMsg.hxx"
#include "SipExpires.hxx"
#include "SipVia.hxx"
#include "SubscriberContainer.hxx"
#include "VIoException.hxx"
#include "SystemInfo.hxx"

using namespace Vocal;


Sptr < SubscriberContainer > SubscriberContainer::theSubContainer = 0;

/**
 * Return the singleton subscriber container or create it
 * if it doesn't already exist. 
 * @return Reference to the subscriber container singleton object.
 */
SubscriberContainer&
SubscriberContainer::instance( )
{
    if ( theSubContainer == 0 )
    {
        theSubContainer = new SubscriberContainer;
    }

    return *(theSubContainer);
}

/**
 * Constructor
 */
SubscriberContainer::SubscriberContainer( )
{
    localSipPort = 0;
    MaxExpireTime = 3600;
    localRSProxy = 0;
}

/**
 * Destructor
 */
SubscriberContainer::~SubscriberContainer( )
{}


/**
 * Add a master user to the Subscriber Container
 * @param  key  Unique identifier for this subscriber
 * @param  calling  Contact list used for calls made from this subscriber.
 * @param  called  Contact list used for calls terminating to this subscriber.
 * @param  term  Contact list used for calls terminating to this subscriber.
 * @param  expireReq  Time in seconds originator of Register message would
 *                    like before having to register again.
 * @param  staticFlag  Flag indicating if subscriber is statically registered.
 * @return Maximum time in seconds before another register message
 *         is required.
 */
unsigned long
SubscriberContainer::add( const Data& key,
                          const Sptr < vector < Data > > calling,
                          const Sptr < vector < Data > > called,
                          const Sptr < vector < Data > > term,
                          const unsigned long expireReq,
                          const bool staticFlag)
{

    /// Obtain the global Max expiry time
    unsigned long allowedExpiry = getMaxExpireTime();

    /// create a Subscriber with 'new'
    Sptr < Subscriber > sub = new Subscriber(key,
                                             calling,
                                             called,
                                             staticFlag);

    /// fill in the fields
    sub->setTerm(term);
    sub->setRegTime();

    /// Determine which is shorter, the Expire length of time requested
    /// in the Register message or the time provided by the Provisioning Server.
    /// Store the result in the Subscriber and pass it back.
    if (expireReq < allowedExpiry )
    {
        sub->setExpiry(expireReq);
        allowedExpiry = expireReq;
    }
    else
    {
        sub->setExpiry(allowedExpiry);
    }

    SubContainerMutex.WriteLock();

    /// insert the Subscriber in the map
    subList.insert(pair < Data, Sptr < Subscriber > > (key, sub ));

    SubContainerMutex.Unlock();

    return allowedExpiry;
}
/**
 * Upon startup, add each user with the same empty subscriber object
 * Aliases at this time are not linked to their master users. That is
 * done later when their provisioning data is obtained.
 * @param  users  List of all valid user names to be added.
 * @param  aliasFlag  Flag indicating whether users in list are aliases.
 */
void
SubscriberContainer::add( vector < Data > & users, const bool aliasFlag)
{
    /// Generate an empty subscriber
    Sptr < Subscriber > blankUser = new Subscriber;
    blankUser->setAlias(aliasFlag);

    SubContainerMutex.WriteLock();

    /// add every subscriber into the subscriber container with the same empty subscriber.
    for (vector < Data > ::iterator listIter = users.begin(); listIter != users.end(); ++listIter)
    {

        /// strip off .xml
        string thisUser = (*listIter).convertString();
        string::size_type pos = thisUser.find(".xml");
        if (pos != string::npos)
        {
            thisUser = thisUser.substr(0, pos);
        }

        if ("" != thisUser)
        {
            Data key(thisUser);


            /// insert the Subscriber in the map
            subList.insert(pair < Data, Sptr < Subscriber > > (key, blankUser ));
        }
    }

    SubContainerMutex.Unlock();
}

/**
 * Delete the subscriber key from the container and make the
 * smart pointer point to a brand new subscriber in case this master key
 * has aliases pointing to the same subscriber object.
 * The aliases will be removed via a separate callback which should
 * immediately follow this one.
 * @param  key  Unique identifier for subscriber.
 */
void
SubscriberContainer::removeMaster( const Data& key )
{
    TableIter i;

    SubContainerMutex.WriteLock();

    /// Find the Subscriber in the database using the key
    i = subList.find(key);

    /// Verify that a subscriber was found
    if (i != subList.end())
    {
        if ( i->second->isProvisioned())
        {
            /// Have the smart pointer point to an empty subscriber so the
            /// aliases will be deleted if they are accessed again.
            i->second->clearCalling();
            i->second->clearCalled();
            i->second->clearTerm();
            i->second->setAlias(true);
        }

        cpLog(LOG_DEBUG, "Removing master Subscriber [%s]", key.logData());
        /// remove the key/Subscriber from the map
        subList.erase(i);
    }

    SubContainerMutex.Unlock();
}

/**
 * Using the key, find the Subscriber object in the map and delete it
 * @param  key  Unique identifier for subscriber.
 */
void
SubscriberContainer::remove( const Data& key )
{
    TableIter i;

    SubContainerMutex.WriteLock();

    /// Find the Subscriber in the database using the key
    i = subList.find(key);

    /// Verify that a subscriber was found
    if (i != subList.end())
    {
        /// remove the key/Subscriber from the map
        subList.erase(i);
    }

    SubContainerMutex.Unlock();
}

/**
 * Store the given contacts for the specified subscriber.
 * @param  user  Unique identifier for subscriber.
 * @param  contacts  Contact list used for terminating a call 
 *                   on this subscriber.
 */
void
SubscriberContainer::updateTerm( Subscriber& user, const Sptr < vector < Data > > contacts)
{
    user.setTerm(contacts);
}

/**
 * Store the given data for the specified subscriber, usually due to a provisioning change.
 * @param  user  Unique identifier for subscriber.
 * @param  calling  Contact list used for outgoing calls from subscriber.
 * @param  called  Contact list used for terminating calls to subscriber.
 * @param  expire  Time in seconds before registration expires.
 */
void
SubscriberContainer::updateConfig( Subscriber& user, const Sptr < vector < Data > > calling, const Sptr < vector < Data > > called, const unsigned long expire )
{
    user.setCalling(calling);
    user.setCalled(called);
    user.setExpiry(expire);

}

/**
 * Store registration specific data for subscriber.
 * @param  to  Unique identifier for registering subscriber.
 * @param  from  Unique identifier for subscriber which triggered the registration.
 * @param  newContacts  Contact list used for terminating calls to this subscriber.
 * @param  reqExpire  Requested time in seconds before next registration must be sent.
 * @param  expire  Time in seconds granted before next registration must be sent.
 * @param  existingContacts  Previously stored terminating contact list.
 * @return  Status response value to be sent back to originator of Register message.
 */
int
SubscriberContainer::updateOnRegister(const Data& to,
                                      const Data& from,
                                      const Sptr < vector < Data > > & newContacts,
                                      const unsigned long &reqExpire,
                                      unsigned long &expire,
                                      vector < Data > & existingContacts)
{
    int statusResponse = 200;  /// ok

    /// Obtain data from provisioning only if the system is Active
    /// Need to avoid flooding Provisioning upon startup during sync.
    RSSTATE rsState = RSProxy::instance().getRsState();

    Sptr < Subscriber > user;
    if (false == search(to, user))
    {
        cpLog( LOG_DEBUG, "Subscriber is not provisioned, failing Registration for [%s]", to.logData());

        statusResponse = 404;  /// not found
        return statusResponse;
    }

    /// RS must be either Active or the Register is for an unprovisioned Alias
    if ((RSACTIVE == rsState) || (true == user->isAlias(to)))
    {
        /// Obtain user info from Provisioning if necessary.
        checkProvisioned(to);

        /// Reassign subscriber since it may have changed.
        if ( false == search(to, user))
        {
            cpLog( LOG_DEBUG, "Subscriber lost, failing Registration for [%s]", to.logData());
            statusResponse = 404;  /// not found
            return statusResponse;
        }

    }
    else
    {
        Data blankName("");
        /// Set is master user, check if it has not yet been registered or provisioned
        if (blankName == user->getMasterName())
        {
            /// create a Subscriber with 'new'
            /// get rid of the blank subscriber for this user
            remove(to);

            /// First time this user is updated, create a new subscriber.
            Sptr < Subscriber > newUser = new Subscriber;
            newUser->setMasterName(to);

            SubContainerMutex.WriteLock();

            /// insert the Subscriber in the map
            subList.insert(pair < Data, Sptr < Subscriber > > (to, newUser ));

            SubContainerMutex.Unlock();

            user = newUser;
        }
    }

    if (user->isStaticReg())
    {
        cpLog( LOG_DEBUG, "User [%s] is configured for Static Registration. Disregarding Register Message.", to.logData());
        statusResponse = 409;   /// Conflict (or should it be 403 forbidden?)
    }
    else
    {
        /// Note: if this user is not yet provisioned and happens to have static registration,
        ///       the correct contacts will be retrieved when either the 1st Invite comes in,
        ///       the user file is modified in Provisioning, or the user Registers again
        ///       when the RS becomes active. Therefore, it is ok to store the term contacts now.

        cpLog( LOG_DEBUG, "Setting Term Contacts and expiry time");
        user->appendTerm(existingContacts);
        user->setRegTime();

        /// Determine which is shorter, the Expire length of time requested
        /// in the Register message or the time provided by the Provisioning Server.
        /// Store the result in the Subscriber and pass it back in the parameter.
        if (reqExpire < expire)
        {
            user->setExpiry(reqExpire);
            expire = reqExpire;
        }
        else
        {
            user->setExpiry(expire);
        }

        /// Update term contacts. Should the Register message be for deregistration
        /// clear out the term contacts and the expiry timer.
        user->setTerm(newContacts);
    }

    /// Display the subscriber's contact information
    display(to);

    return statusResponse;
} /// end of updateOnRegister

/**
 * Store the data against the specified subscriber. Obtain the provisioning
 * data if it has not already been done. Aliases are not updated here.
 * @param  identifier  Unique identifier for subscriber.
 * @param  newCalled  Contact list for terminating a call to this subscriber.
 * @param  newCalling  Contact list for calls originating from this subscriber.
 * @param  newTerm  Contact list for terminating a call to this subscriber.
 * @param  staticRegFlag  Flag to indicate if subscriber is statically registered.
 */
void
SubscriberContainer::updateMasterUser(const Data& identifier,
                                      const Sptr < vector < Data > > & newCalled,
                                      const Sptr < vector < Data > > & newCalling,
                                      const Sptr < vector < Data > > & newTerm,
                                      const bool staticRegFlag)
{
    cpLog( LOG_DEBUG, "SubscriberContainer::updateMasterUser");

    Sptr < Subscriber > user;

    if (false == search(identifier, user))
    {
        return;
    }

    vector < Data > existingTerm;
    user->appendTerm(existingTerm);
    Data blankName("");

    /// User needs a new subscriber object if it has not been provisioned
    /// and it has not already registered.
    if ((false == user->isProvisioned()) && (blankName == user->getMasterName()) )
    {
        /// get rid of the blank subscriber for this user
        remove(identifier);

        /// First time this user is updated, create a new subscriber.
        Sptr < Subscriber > newUser = new Subscriber(identifier,
                                                     newCalling,
                                                     newCalled,
                                                     staticRegFlag);

        /// Only overwrite the term contacts if it is statically registered.
        if (staticRegFlag)
        {
            newUser->setTerm(newTerm);
        }

        SubContainerMutex.WriteLock();
        subList.insert(pair < Data, Sptr < Subscriber > > (identifier, newUser));
        SubContainerMutex.Unlock();
    }
    else
    {
        cpLog( LOG_DEBUG, "Updating Subscriber Data");
        /// update the subscriber data
        user->setCalled(newCalled);
        user->setCalling(newCalling);
        /// Be sure to clear out term contacts if changing from static
        /// to non-static registration.
        cpLog( LOG_DEBUG, "Static Reg flag: stored [%d], new [%d]", user->isStaticReg(), staticRegFlag);
        if ((true == user->isStaticReg()) && (false == staticRegFlag))
        {
            cpLog( LOG_DEBUG, "Clearing Term Contacts");
            user->clearTerm();
        }

        user->setProvisioned(true);
        user->setStaticFlag(staticRegFlag);
        /// Only overwrite the term contacts if it is statically registered.
        if (staticRegFlag)
        {
            user->setTerm(newTerm);
        }
    }
} /// end of updateMasterUser


/**
 * Obtain subscriber object for specified subscriber.
 * @param  key  Unique identifier for subscriber.
 * @param  subFound  Pointer to requested subscriber object.
 * @return True if found, false if not.
 */
bool
SubscriberContainer::search( const Data& key, Sptr < Subscriber > & subFound)
{
    TableIter i;

    SubContainerMutex.ReadLock();

    i = subList.find(key);

    SubContainerMutex.Unlock();

    if (i == subList.end())
    {
       cpLog(LOG_DEBUG,"Failed to get the subscriber for key:%s", key.logData());
       return false;
    }
    else
    {
        subFound = (i->second);
        return true;
    }
}

/**
 * Using the key, find the subscriber and combine its Called and
 * Term contacts to be passed back.
 * @param  key  Unique identifier for subscriber.
 * @param  contacts  List of contacts obtained from specified subscriber.
 * @param  isExpired  Flag indicating if registration has expired.
 * @return  true if subscriber was found, false if not
 */
bool
SubscriberContainer::searchTo( const Data& key, vector < Data > & contacts, bool& isExpired, bool termOnly)
{
    Sptr < Subscriber >  user;

    if (false == search(key, user))
    {
        return false;
    }

    long secondsLeft = 0;

    /// Check to see if the Subscriber has expired
    if (user->expired( secondsLeft ) )
    {
        isExpired = true;
    }

    /// obtain the Called contacts
    if (!termOnly) user->appendCalled(contacts);

    /// obtain the Term contacts
    user->appendTerm(contacts);

    return true;
}

/**
 * Using the key, find the subscriber and obtain it's Calling contacts
 * @param  key  Unique identifier for subscriber.
 * @param  contacts  List of contacts obtained from specified subscriber.
 * @return  true if subscriber was found, false if not
 */
bool
SubscriberContainer::searchFrom( const Data& key, vector < Data > & contacts )
{
    Sptr < Subscriber >  user;

    if ( false == search(key, user))
    {
        return false;
    }

    /// obtain the Calling contact list
    user->appendCalling(contacts);

    return true;

}

/**
 * Diagnostic - Display all information stored against a subscriber
 * @param  key  Unique identifier for subscriber.
 */
void
SubscriberContainer::display( const Data& key)
{
    if (cpLogGetPriority() >= LOG_DEBUG)
    {
        cpLog(LOG_DEBUG, "-------------------------------");
        cpLog( LOG_DEBUG, "Subscriber: %s", key.logData() );

        vector < Data > contactVector;
        /// Find the subscriber in the container
        if (searchFrom(key, contactVector))
        {
            /// print the data contained in the vector
            for (vector < Data > ::iterator i = contactVector.begin(); i != contactVector.end(); ++i)
            {
                cpLog( LOG_DEBUG, "Calling: %s", i->logData() );
            }
        }
        else
        {
            cpLog(LOG_DEBUG, "From not found");
        }

        // clear out the vector
        contactVector.clear();

        bool toExpire = false;
    
        if (searchTo(key, contactVector, toExpire))
        {
            /// print the data contained in the vector
            for (vector < Data > ::iterator i = contactVector.begin(); i != contactVector.end(); ++i)
            {
                cpLog( LOG_DEBUG, "Called/Term: %s", i->logData() );
            }
        }
        else
        {
            cpLog(LOG_DEBUG, "To not found");
        }

        Sptr < Subscriber >  user;
        if (true == search(key, user))
        {
            long expiryTimer = 0;
            user->expired( expiryTimer );
            cpLog( LOG_DEBUG, "Expire Time left: %d", expiryTimer);
        }

        cpLog(LOG_DEBUG, "-------------------------------");
    }
}

/**
 * Display all subscribers.
 */
void
SubscriberContainer::displayAll()
{
    if (cpLogGetPriority() >= LOG_DEBUG)
    {
        cpLog( LOG_DEBUG, "SubscriberContainer::displayAll");

        vector < Data > tempList;
    
        SubContainerMutex.ReadLock();
        for ( TableIter i = subList.begin(); i != subList.end(); ++i)
        {
            tempList.push_back(i->first);
        }
        SubContainerMutex.Unlock();

        for ( vector < Data > ::iterator j = tempList.begin(); j != tempList.end(); ++j)
        {
            /// Display each element of the map
            display(*j);
        }
    }
}

/**
 * Loop through each subscriber and generate a Register 
 * message for each one that is currently registered to be
 * sent to all other RS's configured in the system
 * @param  remoteRSHost  Host of RS in need of synchronization.
 * @param  remoteRSPort  Port of RS on which to send Register messages.
 * @param  count  Number of registered users currently in the system.
 * @return  true if synchronization occurred correctly.
 */
bool
SubscriberContainer::synchronize(const Data& remoteRSHost, const Data& remoteRSPort, int& count)
{
    // get the host name from theSystem.
    // theSystem is a global SystemInfo
    Data localRSHost(theSystem.gethostAddress());
    cpLog (LOG_DEBUG, "localRSHost = %s", localRSHost.logData());

    Data localRSPort(getPort());

    SubContainerMutex.ReadLock();
    /// Generate a Register message for each subscriber
    for ( TableIter currentSubscriber = subList.begin(); currentSubscriber != subList.end(); ++currentSubscriber)
    {
        if (generateRegisterMsg(*(currentSubscriber->second), currentSubscriber->first, remoteRSHost,
                                remoteRSPort, localRSHost, localRSPort))
        {
            ++count;
        }
    }
    SubContainerMutex.Unlock();
    return true;

} /// end of Synchronize

/**
 * Generate and send a Register message for specified subscriber.
 * @param  subscriber  Subscriber object containing info to be sent.
 * @param  userName  Unique identifier for subscriber.
 * @param  remoteRSHost  Host of RS in need of synchronization.
 * @param  remoteRSPort  Port of RS on which to send Register messages.
 * @param  localRSHost  Host of this RS which is sending the Register message.
 * @param  localRSPort  Port of this RS on which responses should be sent.
 * @return  true if message sent, false if not.
 */
bool
SubscriberContainer::generateRegisterMsg( Subscriber& subscriber,
        const Data& userName,
        const Data& remoteRSHost,
        const Data& remoteRSPort,
        const Data& localRSHost,
        const Data& localRSPort)
{


    long secTilExpiry = 0;

    /// Obtain Subscriber's Term contacts
    vector < Data > contacts;
    subscriber.appendTerm(contacts);


    /// Only create a Register message for master subscribers who have Term Contacts and
    /// are not set for static registration.
    if ( (false == subscriber.isAlias(userName)) && (0 < contacts.size()) && (false == subscriber.isStaticReg()))
    {
        cpLog( LOG_DEBUG, "Sending sync registration for [%s]", userName.logData());
        subscriber.expired(secTilExpiry);

        RegisterMsg registerMsg;

        SipExpires sipExpires;
        sipExpires.setDelta( secTilExpiry );
        registerMsg.setExpires( sipExpires );

        /// Host:port of destination RS
        Sptr <SipUrl> sipUrl = new SipUrl();
        sipUrl->setUserValue(userName);
        sipUrl->setHost( remoteRSHost );
        sipUrl->setPort( remoteRSPort );

        SipRequestLine reqLine = registerMsg.getRequestLine();
        reqLine.setUrl( sipUrl );
        registerMsg.setRequestLine( reqLine );

        /// Subscriber as the user, host:port of this RS
        SipFrom sipfrom = registerMsg.getFrom();
        //TODO  sipfrom.setDisplayName( config->getDisplayName() ); //TODO
        sipfrom.setUser( userName );
        sipfrom.setHost( localRSHost );
        sipfrom.setPort( localRSPort );
        registerMsg.setFrom( sipfrom );

        /// Subscriber as the user, host:port of destination RS
        SipTo sipto = registerMsg.getTo();
        sipto.setUser( userName );
        sipto.setHost( remoteRSHost );
        sipto.setPortData( remoteRSPort );
        registerMsg.setTo( sipto );

        /// Host:port of this RS
        SipVia sipvia = registerMsg.getVia();
        sipvia.setHost( localRSHost ) ;
        sipvia.setPortData( localRSPort ) ;
        registerMsg.flushViaList();
        registerMsg.setVia( sipvia );

        /// Add contacts to the message
        for (vector < Data > ::iterator i = contacts.begin(); i != contacts.end(); ++i)
        {
            /// Rely upon the constructor to decode the contact
            SipContact myContact( *i );
            registerMsg.setContact(myContact);
        }
        registerMsg.setNumContact( contacts.size() );

        /// Send the newly created Register message
        localRSProxy->getSipStack()->sendAsync( registerMsg );
        cpLog( LOG_DEBUG, "Sending Register Message for [%s]", userName.logData());
        return true;

    }
    else
    {
        //cpLog(LOG_DEBUG, "Subscriber is static or has yet to get PS data [%s]", userName.logData());
    }

    return false;
} /// end of generateRegisterMsg

/**
 * Obtain the list of all Subscribers from PS and
 * store them in the Subscriber Container.   
 * Begin with the master users and then get the aliases.
 */
void
SubscriberContainer::updateAllSubscribers()
{
    cpLog( LOG_DEBUG, "updateAllSubscribers triggered");

    /// Retrieve the subscriber list from the PS
    try
    {
        cpLog( LOG_DEBUG, "Calling getSubscribers ");
        vector < Data > subscriberList;
        ProvisionInterface::instance().getSubscribers(subscriberList);
        add(subscriberList, false);
    }
    catch (VException& e)
    {
        /// failed to connect to the Provisioning Server
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
    }

    try
    {
        cpLog( LOG_DEBUG, "Calling getAliases ");
        vector < Data > aliasList;
        ProvisionInterface::instance().getAllAliases(aliasList);
        add(aliasList, true);

    }
    catch (VException& e)
    {
        /// failed to connect to the Provisioning Server
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
    }

    /// TEST code to display updated data
    //SubscriberContainer::instance().displayAll();
    cpLog( LOG_ALERT, "Finished loading all users. [%d] users loaded into memory.", getSize());

} /// end of updateAllSubscribers


/**
 * Step through the Alias list checking to see if each one
 * already exists in the Subscriber Container. If it does not
 * it will be added and it will register an interest in it
 * with provisioning.
 * @param  masterKey  Unique identifer for this master subscriber.
 * @param  aliasList  List of alias identifiers for this master subscriber.
 */
void
SubscriberContainer::updateAliasList(const Data& masterKey, list < string > & aliasList)
{
    cpLog( LOG_DEBUG, "updateAliasList triggered");

    Sptr < Subscriber > sub;

    /// Obtain the subscriber object for the master key
    if (false == search(masterKey, sub))
    {
        return;
    }

    Sptr < Subscriber > aliasSub;
    /// Loop through aliases
    for (list < string > ::iterator i = aliasList.begin(); i != aliasList.end(); ++i)
    {
        string alias = *i;
        cpLog( LOG_DEBUG, "Alias list entry [%s]", alias.c_str());

        Data key(alias);
        cpLog(LOG_DEBUG, "Updating Alias [%s]", key.logData());

        if ("" == alias)
        {
            cpLog(LOG_DEBUG, "skipping Alias [%s]", key.logData());
            continue;
        }

        /// If the alias does not yet exist in the Subscriber Container, add it.
        if (false == search(key, aliasSub))
        {
            cpLog(LOG_DEBUG, "Inserting Alias into Subscriber Container [%s]", key.logData());

            SubContainerMutex.WriteLock();

            // add to Subscriber Container with smart pointer to same Subscriber object
            subList.insert(pair < Data, Sptr < Subscriber > > (key, sub ));

            SubContainerMutex.Unlock();
            display(key);
            continue;
        }

        if (false == aliasSub->isProvisioned())
        {
            /// Link this Alias to the master subscriber in the container.

            remove(key);

            SubContainerMutex.WriteLock();

            // add to Subscriber Container with smart pointer to same Subscriber object
            subList.insert(pair < Data, Sptr < Subscriber > > (key, sub ));

            SubContainerMutex.Unlock();
        }
    }
} /// end of updateAliasList

/**
 * Obtain data from provisioning for specified subscriber if it has not been done already.
 * If key is an alias it will be linked to its master user.
 * @param  key  Unique identifier for subscriber.
 */
void
SubscriberContainer::checkProvisioned(const Data& key)
{

    Sptr < Subscriber > user;

    if (false == search(key, user))
    {
        return ;
    }

    if (false == user->isProvisioned())
    {
        cpLog(LOG_DEBUG, "Subscriber[%s] is valid, retrieving data from PS", key.logData());

        /// This is the first time the subscriber has been accessed.
        /// Need to get the info from provisioning.
        LocalScopeAllocator lo;
        if (true == user->isAlias(key))
        {
            string masterUser;
            /// Key is an unprovisioned alias. Get master first.
            try
            {
                masterUser = ProvisionInterface::instance().getMasterUser(key.getData(lo));
            }
            catch (VMissingDataException& e)
            {
                cpLog( LOG_DEBUG, "Deleting Alias [%s]", key.logData());
                /// If unable to access the alias it must have been deleted. Remove it.
                remove(key);
                return ;
            }
            catch (VException& e)
            {
                cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
                return ;
            }
            /// updating the master user will update its aliases
            updateUser(masterUser, false);
        }
        else /// Key is an unprovisioned master user
        {
            updateUser(key.getData(lo), false);
        }
    }
} /// end of checkProvisioned

/// End of file


/**
 * The Provisioning Server is notifying us of a change in
 * a specific subscriber. A user could have been Changed or Deleted.
 * @param  data  Contents of xml file. 
 * @param  fileName  Subscriber's identifier
 * @param  deleteFlag  Indicator of whether or not subscriber was deleted.
 */
void
updateSubscriber(const string& data, const string& fileName, bool deleteFlag)
{
    updateUser(fileName, deleteFlag);
 
    if (cpLogGetPriority() >= LOG_DEBUG)
    {
        if (false == deleteFlag)
        {
            /// Strip off the '.xml' extention
            string withOutXml = fileName;
            string::size_type pos = withOutXml.find(".xml");
            if (pos != string::npos)
            {
                withOutXml = withOutXml.substr(0, pos);
            }
            SubscriberContainer::instance().display(withOutXml);
        }
    }
}

/**
 * The Provisioning Server is notifying us of a change in the 
 * Accounts Directory. A user could have been Added, Changed or Deleted.
 * @param  data  Subscriber's identifier
 * @param  fileName  Name of file as it appears in provisioning.
 * @param  deleteFlag  Indicator of whether file was deleted or not.
 */
void
updateAccounts(const string& data, const string& fileName, bool deleteFlag)
{
    updateUser(data, deleteFlag);

    if (cpLogGetPriority() >= LOG_DEBUG)
    {
        if (false == deleteFlag)
        {
            /// Strip off the '.xml' extention
            string withOutXml = data;
            string::size_type pos = withOutXml.find(".xml");
            if (pos != string::npos)
            {
                withOutXml = withOutXml.substr(0, pos);
            }
            SubscriberContainer::instance().display(withOutXml);
        }
    }
}

/**
 *  Obtain the subscriber information from provisioning
 *  This should only be called for master users, not aliases.
 * @param  userName  Master user identifier 
 * @param  deleteFlag  Indicator of whether subscriber was deleted or not.
 */
void
updateUser(const string& userName, bool deleteFlag)
{
    /// obtain the subscriber container
    SubscriberContainer& theContainer = SubscriberContainer::instance();

    /// Strip off the '.xml' extention
    string withOutXml = userName;
    string::size_type pos = withOutXml.find(".xml");
    if (pos != string::npos)
    {
        withOutXml = withOutXml.substr(0, pos);
    }
    //cpLog( LOG_DEBUG, "Stripped off .xml [%s]", withOutXml.c_str());

    Data subscriber (withOutXml);

    if (true == deleteFlag)
    {
        /// Delete the subscriber key from the container and make the
        /// Sptr point to a brand new subscriber in case this master key
        /// has aliases pointing to the same subscriber object.
        theContainer.removeMaster(subscriber);
        return ;
    }

    list < string > aliasList;

    try
    {
        /// Retrieve the alias list from the Provisioning Server
        ProvisionInterface::instance().getAliases( withOutXml, aliasList);
    }
    catch (VMissingDataException& e)
    {
        /// Unable to obtain the Subscriber file means it has been deleted.
        cpLog(LOG_ALERT, "Failed to obtain file from Provisioning Server, reason %s", e.getDescription().c_str());
    }
    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed to get data from Provisioning Server, reason %s", e.getDescription().c_str());
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
    }

    /// Obtain the Subscriber's data from the PS and
    /// add it to the container.
    Sptr < vector < Data > > calling = new vector < Data > ;
    Sptr < vector < Data > > called = new vector < Data > ;
    Sptr < vector < Data > > term = new vector < Data > ;
    list < string > termList;

    /// Obtain the local port number
    int localPort = theContainer.getPort();
    //cpLog( LOG_DEBUG, "Local Port [%d]", localPort);

    /// get the host name from theSystem
    Data hostName = theSystem.gethostAddress();
    //cpLog( LOG_DEBUG, "Local Host [%s]", hostName.logData());

    /// Retrieve the subscriber specific info from the PS
    LocalScopeAllocator lo;
    try
    {
        ProvisionInterface::instance().getCallingList( hostName.getData(lo), localPort, withOutXml, *calling);

    }
    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed to get data from Provisioning Server, reason %s", e.getDescription().c_str());
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
    }

    try
    {
        ProvisionInterface::instance().getCalledList( hostName.getData(lo), localPort, withOutXml, *called);

    }
    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed to get data from Provisioning Server, reason %s", e.getDescription().c_str());
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
    }

    bool staticRegFlag = false;
    try
    {
        /// Get Static Registration flag
        staticRegFlag = ProvisionInterface::instance().getTermContactList(withOutXml, termList);

        if (staticRegFlag)
        {
            cpLog( LOG_DEBUG, "Static Registration activated");
            string current;
            for (list < string > ::iterator stringIter = termList.begin(); stringIter != termList.end(); ++stringIter)
            {
                current = *stringIter;
                string::size_type pos = current.find(":");
                if (pos != string::npos)
                    current = "sip:$USER@" + current;
                term->push_back(current);
            }
        }
    }
    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed to get data from Provisioning Server, reason %s", e.getDescription().c_str());
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
    }

    vector < Data >  junkVector;
    if (theContainer.searchFrom( subscriber, junkVector ))
    {
        cpLog( LOG_DEBUG, "Updating an existing subscriber [%s]", subscriber.logData() );
        theContainer.updateMasterUser(subscriber, called, calling, term, staticRegFlag);
    }
    else
    {
        cpLog( LOG_DEBUG, "config new subscriber [%s]", subscriber.logData() );

        /// Create the Subscriber and store it in the Subscriber database
        theContainer.add(subscriber, calling, called, term, 0, staticRegFlag);
    }

    /// Update the aliases
    theContainer.updateAliasList(subscriber, aliasList);

} /// end of updateUser

/**
 * The Provisioning Server is notifying us of a change in the 
 * Aliases Directory. There will be no changes to an alias.
 * They can only be added and deleted. If it is deleted, remove
 * it. If it was added, rely upon the master user to add it.
 * @param  data  Alias's identifier 
 * @param  fileName  Name of file modified in provisioning.
 * @param  deleteFlag  Indicator of whether file was deleted or not.
 */
void
updateAliases(const string& data, const string& fileName, bool deleteFlag)
{

    cpLog( LOG_DEBUG, "updateAliases triggered");

    /// Strip off the '.xml' extention
    string withOutXml = data;
    string::size_type pos = withOutXml.find(".xml");
    if (pos != string::npos)
    {
        withOutXml = withOutXml.substr(0, pos);
    }
    cpLog( LOG_DEBUG, "Stripped off .xml [%s]", withOutXml.c_str());

    Data identifier(withOutXml);

    if (true == deleteFlag)
    {
        cpLog( LOG_DEBUG, "Deleting Alias [%s]", identifier.logData());
        SubscriberContainer::instance().remove(identifier);
    }

} /// end of updateAliases

/**
 * The maximum expire time was updated in provisioning. Store the new maximum.
 * @param  data  Contents of xml file.
 * @param  fileName  Name of file modified in provisioning.
 * @param  deleteFlag  Indicator of whether file was deleted or not.
 */
void
updateExpireTime(const string& data, const string& fileName, bool deleteFlag)
{
    cpLog( LOG_DEBUG, "updating Max Expiry Time");

    /// get pointer to global config object
    VGlobalConfigData globalConfigData;

    try
    {
        ProvisionInterface::instance().getConfigData(globalConfigData);
    }
    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed to get data from Provisioning Server, reason %s", e.getDescription().c_str());
        return;
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
        return;
    }

    SubscriberContainer::instance().setMaxExpireTime(globalConfigData.getExpiryTimer());
}
// End of File

