/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const MSManager_hxx_Version = 
"$Id: MSManager.cxx,v 1.2 2001/05/26 01:32:37 vzubarev Exp $";

#include <iterator>

#include "VLog.hxx"
#include "XlatorException.hxx"
#include "MSManager.hxx"
#include "../Capability.hxx"
#include "../H323_CapTable.hxx"

using namespace Translator;

MSManager::MSManager(translator::CallTranslatorEM *em) : evtMgr(em) {}

MSManager::~MSManager()
{
    MediaSesContainer::const_iterator p = mediaSesContainer.begin();
    
    while(p != mediaSesContainer.end())
    {
      delete p->second;
      p++;
    }
}

SM_MSManager* MSManager::getSM(SessionID i) const throw()
{
    VLog	    	log(fn);

    if(mediaSesContainer.find(i) == mediaSesContainer.end())
    {
	 VERR(log) << fn << "Media Session not found with SesID : " << i << VERR_END(log);
	 throw XlatorException("Media Session not found",__FILE__, __LINE__);
    }
    
    return mediaSesContainer[i];
}

void MSManager::addSM(SessionID i, SM_MSManager *sm)
{
    mediaSesContainer[i] = sm;
}


bool MSManager::handleOLC(const OLC &msgOLC)
{
  const string	fn("MSManager::handleOLC");
  VLog	    	log(fn);
  
  int pTypeCap = 0;
  unsigned int lChannelNumber = 0;
  int sID = 0, altCapN = 0, index;
  
  VDEBUG(log) << fn << ": BEGIN ----------------- " << VDEBUG_END(log);
  
  lChannelNumber = (int)msgOLC.m_forwardLogicalChannelNumber;
  VDEBUG(log) << fn << ": LC # : " <<  lChannelNumber << VDEBUG_END(log);
  
  const H245_DataType &dataType = msgOLC.m_forwardLogicalChannelParameters.m_dataType;

  // I'm afraid we should increment maxLogChN only when we form OLC
  // maxLogChN = lChannelNumber;
  
  if(msgOLC.m_forwardLogicalChannelParameters.m_multiplexParameters.GetTag() !=
     H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters
     ::e_h2250LogicalChannelParameters)
  {
    VERR(log) << fn << ": No H2250LogicalChannelParameters field in OLC..." << VERR_END(log);
    return false;
  }

  try
  {	
    pTypeCap = globalCapTable.findCapTableEntry(dataType);
  }
  catch (XlatorException)
  {
    VERR(log) << fn << ": Capability received in OLC not supported - weird..." << VERR_END(log);
    return false;
  }
      
  Capability *capability = capabilityMap[pTypeCap];
       
  Capability *copy = capability->clone();

  if(copy->setParamRecvMsg(dataType))
  {
    copy->assignCapabilityNumber(pTypeCap);
    
    H323_CapTable::CapTable_I pCap = capSetSip.h323CapTable.getTableIter_pt(pTypeCap);
    if(pCap == NULL)
    {
      VERR(log) << fn << ": Capability received in OLC - not in Sip CapTable" << VERR_END(log);
      return false;
    }
    
    VDEBUG(log) << fn << ": packetSize in OLC : " << copy->getPacketSizeMilSec() << VDEBUG_END(log);
    VDEBUG(log) << fn << ": packetSize in SDP : " << (*pCap)->getPacketSizeMilSec() << VDEBUG_END(log);
    
    if(copy->getPacketSizeSamples() > (*pCap)->getPacketSizeSamples())
    {
      if(strictModePackSize)
      {
        VERR(log) << fn << ": Packet size of Cap received in OLC > than in SDP" << VERR_END(log);
        return false;
      }
      else
      {
        VWARN(log) << fn << ": Packet size of Cap received in OLC > than in SDP" << VWARN_END(log);
      }
    }
      // add the capability to a table which then will be used to form
      // SDP message
      fromOLCtoSDPcapTable.addCapTableEntry(copy);
	    
  }
  else
  {
    delete copy;
    VERR(log) << fn << ": Could not set parameters for a cap received in OLC" << VERR_END(log);
    // should never happen
    return false;
  }

  const H245_H2250LogicalChannelParameters &params = 
        msgOLC.m_forwardLogicalChannelParameters.m_multiplexParameters;

   sID = params.m_sessionID;
	  
   if(sID == 3)
   {
	VERR(log) << fn << ": Do not support data for now..." << VERR_END(log);
	return false;
   }

   if((master)&&(sID > 3))
   {
      sA_I = sessionAltCapMap.find(sID);
      if(sA_I == sessionAltCapMap.end())
      {
	 // we did not assign this sessionID
	 VERR(log) << fn << ": Only master can assign the sessionID number > 3" << VERR_END(log);
	 return false;
      }
      else
      {
	  // this OLC message is the last in a given session
	  // we have already assigned the sessionID
	  // and sent the first OLC message with this sessionID
	      
	  index = sessionAltCapMap[sID];
	  try
	  {
	    sDescriptor.transportIDH323.setDone(curDescrN, altCapN);
	  }
	  catch (XlatorException &mS)
	  {
	    VERR(log) << fn << ": setDone error : " << mS.getDescription() << VERR_END(log);
	    VDEBUG(log) << fn << ": We are master, sID : " << sID << ", pTypeCap : " <<  pTypeCap << VDEBUG_END(log);
	    VDEBUG(log) << fn << ": curDescrN : " << curDescrN << ", altCapN : " << altCapN << VDEBUG_END(log);
            *RejCause = H245_OpenLogicalChannelReject_cause::e_unspecified;
	    return false;
	  }
      }
   }
   else if((master))
   {
        // here we should find out whether we already have
	// sent the OLC message for this particular pTypeCap
	
	// if altCapN for this pTypeCap is in mediaDescrTable
	// already but this particular session is not done yet
	// then obviously we have already sent an OLC for this
	// capability from this particular AltCapSet with this
	// altCapN, so this OLC we receive is the last in the
	// session
	
	// if we find altCapN for this particular capability,
	// for which the session is not done yet, then we
	// interpret the message we have just received
	// as the first OLC in a new session
        // going from slave endpoint, so we should assign
        // a new sessionID for this session as well as find
	// altCapN of the particular pTypeCap we received
	    
	VDEBUG(log) << fn << ": We are master and sID : " << sID << ", pTypeCap : " <<  pTypeCap << VDEBUG_END(log);
	
	altCapN = findAltCapSetN(pTypeCap);
	
	if((sID < 0)||(sID > 3))
	{
	  VERR(log) << fn << ": Invalid sessionID received from slave endpoint : " << sID << VERR_END(log);
          *RejCause = H245_OpenLogicalChannelReject_cause::e_dataTypeNotSupported;
	  return false;
	}
	
	
	if(altCapN == -1)
	{
	   VERR(log) << fn << ": Did not find the altCapSet for a cap received in OLC" << VERR_END(log);
           *RejCause = H245_OpenLogicalChannelReject_cause::e_unspecified;
	   return false;
	}

	VDEBUG(log) << fn << ": handle OLC: master now sID : " << sID << VDEBUG_END(log);
	
	MediaDescrTable_I mI = std::find(mediaDescrTable.begin(), mediaDescrTable.end(), altCapN);
	
	if(mI == mediaDescrTable.end())
	{
	  // this is the first OLC message in a session, so we should do
	  // all appropriate stuff
	  
	  generateSID(sID, altCapN);

	  mediaDescrTable.push_back(altCapN);
	   
	  // we should assign the capability in the appropriate
	  // sessionDescriptor
	  
	  try
	  {
	      sDescriptor.setCapability(pTypeCap, curDescrN, altCapN);
	  }
	  catch (XlatorException)
	  {
	      VERR(log) << fn << ": Could not set cap for a given sessionDescriptor" << VERR_END(log);
              *RejCause = H245_OpenLogicalChannelReject_cause::e_unspecified;
	      return false;
	  }
	}
	else
	{
	   // this is the last message in a session, just find out the 
	   // sessionID for this particular altCapN
	   
	   std::map<int, int>::const_iterator s_I = sessionAltCapMap.begin();
	   
	   while(s_I != sessionAltCapMap.end())
	   {
	     if(s_I->second == altCapN)
	     {
	       sID = s_I->first;
	       break;
	     }
	   }
	   
	   if(sID == 0)
	   {
	     VERR(log) << fn << ": Could not find appropriate sID, weird..." << VERR_END(log);
	     // basically should never happen
             *RejCause = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
	     return false;
	   }
	}
   }
   else if((!master)&&(sID == 0))
   {
        VERR(log) << fn << ": Being a slave and receiving sessionID - 0, weird..." << VERR_END(log);
        *RejCause = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
        return false;
   }
   else if((!master)&&(sID!=0))
   {
	// if the given sID is already associated with particular
	// altCapN, than this message
	// is the last one in a given session
	// otherwise (if this sID is not associated yet) it is
	// the first OLC message for a new session
	// so we should then send an OLC from
	// our side
	      
	if(sID == 3)
	{
	  VERR(log) << fn << ": Do not support data for now..." << VERR_END(log);
          *RejCause = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
	  return false;
	}
	
	sA_I = sessionAltCapMap.find(sID);

	if(sA_I != sessionAltCapMap.end())
	{
	   // this message is the last OLC for this session
	   
	   altCapN = sessionAltCapMap[sID];

	   MediaDescrTable_I mI = std::find(mediaDescrTable.begin(), mediaDescrTable.end(), altCapN);
	
	   if(mI == mediaDescrTable.end())
	   {
	     // this is the first OLC message in a session, so we should do
	     // all appropriate stuff
	  
	     generateSID(sID, altCapN);

	     mediaDescrTable.push_back(altCapN);
	   
	     // we should assign the capability in the appropriate
	     // sessionDescriptor
	  
	     try
	     {
	        sDescriptor.setCapability(pTypeCap, curDescrN, altCapN);
	     }
	      catch (XlatorException)
	     {
	        VERR(log) << fn << ": Could not set cap for a given sessionDescriptor" << VERR_END(log);
                *RejCause = H245_OpenLogicalChannelReject_cause::e_unspecified;
	        return false;
	     }
	}
	else
	{
	   // this is the last message in a session, just find out the 
	   // sessionID for this particular altCapN
	   
	   std::map<int, int>::const_iterator s_I = sessionAltCapMap.begin();
	   
	   while(s_I != sessionAltCapMap.end())
	   {
	     if(s_I->second == altCapN)
	     {
	       sID = s_I->first;
	       break;
	     }
	   }
	   
	   if(sID == 0)
	   {
	     VERR(log) << fn << ": Could not find appropriate sID, weird..." << VERR_END(log);
	     // basically should never happen
             *RejCause = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
	     return false;
	   }
	}

	   try
	   {
	     sDescriptor.transportIDH323.setDone(curDescrN, altCapN);
	   }
	   catch (XlatorException &mS)
	   {
	     VDEBUG(log) << fn << ": Last OLC message" << VDEBUG_END(log);
	     VDEBUG(log) << fn << ": slave sID : " << sID << ", altCapN : " << altCapN << VDEBUG_END(log);
	     VERR(log) << fn << ": setDone error" << VERR_END(log);
             *RejCause = H245_OpenLogicalChannelReject_cause::e_unspecified;
	     return false;
	   }
	}
	else
	{
	  // this message is the first one for a new session
	  // should find the altCapN for this particular pTypeCap
	  
	  VDEBUG(log) << fn << ": First OLC message" << VDEBUG_END(log);
	  VDEBUG(log) << fn << ": We are slave; sID : " << sID << ", pTypeCap : " << pTypeCap << VDEBUG_END(log);
	  
	  altCapN = findAltCapSetN(pTypeCap);
	  if(altCapN == -1)
	  {
	     VERR(log) << fn << ": Did not find the altCapSet for a cap received in OLC" << VERR_END(log);
	     VERR(log) << fn << ": Probably LC with invalid capability in it" << VERR_END(log);
             *RejCause = H245_OpenLogicalChannelReject_cause::e_unspecified;
	     return false;
	  }
	  
	  generateSID(sID, altCapN);
	  mediaDescrTable.push_back(altCapN);
	  
	  VDEBUG(log) << fn << ": We are slave; pTypeCap : " << pTypeCap
	    	      << ", atlCapN : " << altCapN
		      << ", sID : " << sID << VDEBUG_END(log);
	  
	  // this message is the first OLC in a new session so
	  // we should assign the capability in the appropriate
	  // sessionDescriptor
	  try
	  {
	     sDescriptor.setCapability(pTypeCap, curDescrN, altCapN);
	  }
	  catch (XlatorException)
	  {
	     VERR(log) << fn << ": Could not set cap for a given sessionDescriptor" << VERR_END(log);
	     // should never happen
             *RejCause = H245_OpenLogicalChannelReject_cause::e_unspecified;
	     return false;
	  }
          
	  VDEBUG(log) << fn << ": We are slave, after setCapability" << VDEBUG_END(log);
	}
   }

  if(params.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel))
  {
    const H245_TransportAddress &trAddress = params.m_mediaControlChannel;
    if(trAddress.GetTag() == H245_TransportAddress::e_unicastAddress)
    {
      const H245_UnicastAddress &uAddress = trAddress;
      if(uAddress.GetTag() == H245_UnicastAddress::e_iPAddress)
      {
        const H245_UnicastAddress_iPAddress &ipAddress = uAddress;
        
        Address address(ipAddress);   

	  try
          {
            // Setting RTCP host/port, currently use only first (# 0) capability descriptor
            sDescriptor.transportIDH323.setRTCPAddr(address, curDescrN, altCapN);
	    
	    // Setting Capability for a given RTP Session
	    sDescriptor.setCapability(pTypeCap, curDescrN, altCapN);
          }
          catch (XlatorException)
          {
            VERR(log) << fn << ": Something wrong with setting RTCP address..." << VERR_END(log);
          }
      }
      else VWARN(log) << fn << ": H245_UnicastAddress is not ipAddress" << VWARN_END(log);
    }
    else VWARN(log) << fn << ": H245_TransportAddress is not Unicast" << VWARN_END(log);
  }
  else VWARN(log) << fn << ": msgOLC does not have mediaControl channel" << VWARN_END(log);
  
  receivedOLC++;
  
  VDEBUG(log) << fn << ": Logical channel # : " << lChannelNumber << VDEBUG_END(log);
  
}

bool MSManager::makeOLCAck(OLCAck &msgOLCAck)
{

  msgOLCAck.m_forwardLogicalChannelNumber = lChannelNumber;
  msgOLCAck.IncludeOptionalField(H245_OpenLogicalChannelAck::e_forwardMultiplexAckParameters);
  msgOLCAck.m_forwardMultiplexAckParameters.SetTag(
    H245_OpenLogicalChannelAck_forwardMultiplexAckParameters::e_h2250LogicalChannelAckParameters);
  
  H245_H2250LogicalChannelAckParameters &paramAck = msgOLCAck.m_forwardMultiplexAckParameters;
  
  paramAck.IncludeOptionalField(
    H245_H2250LogicalChannelAckParameters::e_sessionID);  

  paramAck.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel);  

  paramAck.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaControlChannel);  

  paramAck.m_sessionID = sID;

  H245_UnicastAddress_iPAddress mediaChannel;

  try
  {
    mediaChannel = sDescriptor.transportIDIntSip.getRTPAddr(curDescrN, altCapN);
  }
  catch (XlatorException)
  {
    VERR(log) << fn << ": Error with setting mediaChannel" << VERR_END(log);
    return false;
  }

  H245_TransportAddress &trAddress = paramAck.m_mediaChannel;
  trAddress.SetTag(H245_TransportAddress::e_unicastAddress);
  
  H245_UnicastAddress &uAddress = trAddress;
  uAddress.SetTag(H245_UnicastAddress::e_iPAddress);
  H245_UnicastAddress_iPAddress &ipAddress = uAddress;
        
  ipAddress = mediaChannel;

  H245_UnicastAddress_iPAddress mediaControlChannel;
  try
  {
    mediaControlChannel = sDescriptor.transportIDIntSip.getRTCPAddr(curDescrN, altCapN);
  }
  catch (XlatorException)
  {
    VERR(log) << fn << ": Error with setting mediaControlChannel" << VERR_END(log);
    return false;
  }

  H245_TransportAddress &trcAddress = paramAck.m_mediaControlChannel;
  trcAddress.SetTag(H245_TransportAddress::e_unicastAddress);
  
  H245_UnicastAddress &ucAddress = trcAddress;
  ucAddress.SetTag(H245_UnicastAddress::e_iPAddress);
  H245_UnicastAddress_iPAddress &ipcAddress = ucAddress;
        
  ipcAddress = mediaControlChannel;

  VDEBUG(log) << fn << ": END ----------------- " << VDEBUG_END(log);
  
  return true;
}

void MSManager::makeOLC(OLC &msgOLC) throw ()
{
   const string     fn("MSManager::makeOLC");
   VLog     	    log(fn);
   
   int logChNumber = ++maxLogChN;
   int pTypeCap = 0, sID = maxSesID;
   int altCapN = 0;
   bool iDone = false;
   
   VDEBUG(log) << fn << ": BEGIN --------------" << VDEBUG_END(log);
   
   sA_I = sessionAltCapMap.find(sID);
   if(sA_I == sessionAltCapMap.end())
   {
     VDEBUG(log) << fn << ": Did not find an altCapNumber assosiated with current sID" << VDEBUG_END(log);
   }
   else
   {
     altCapN = sessionAltCapMap[sID];
   }

   try
   {
     iDone = sDescriptor.transportIDH323.isDone(curDescrN, altCapN);
   }
   catch (XlatorException &mS)
   {
     VDEBUG(log) << fn << ": isDone error: " << mS.getDescription() << VDEBUG_END(log);
   }
   
   if(iDone || (sID == 0))
   {
     // for the maxSesID we are done, so this is the first
     // message in a new session
     // so we start with new sessionID if we are master
     // or just assign 0 if we are slave
     
     if(master)
     {
       // if we are master, we create new sessionID
       sID = ++maxSesID;
       
       VDEBUG(log) << fn << ": We master, so sID : " << sID << VDEBUG_END(log);

       if(maxSesID == 1)
       {
         try
	 {
	   pTypeCap = findFirst('A', altCapN);
	   VDEBUG(log) << fn << ": Here is the pTypeCap we found first : " << pTypeCap << VDEBUG_END(log);
	 }
	 catch (XlatorException)
	 {
	   VCRIT(log) << fn << ": Cannot find the capability from altCapSet in GlobalCapTable" << VCRIT_END(log);
	   throw XlatorException("Cannot find the capability from altCapSet in GlobalCapTable",
	                         __FILE__, __LINE__);
	 }
       }
       else if(maxSesID == 2)
       {
         try
	 {
	   pTypeCap = findFirst('V', altCapN);
	 }
	 catch (XlatorException)
	 {
	   VCRIT(log) << fn << ": Cannot find the capability from altCapSet in GlobalCapTable" << VCRIT_END(log);
	   throw XlatorException("Cannot find the capability from altCapSet in GlobalCapTable",
	                         __FILE__, __LINE__);
	 }
       }
       // currently do not support data
       else
       {
         // here we should find the next altcapset for which we
	 // have not opened the logical channel yet
	 
	 try
	 {
	   pTypeCap = findNext(altCapN);
	 }
	 catch (XlatorException)
	 {
	   VCRIT(log) << fn << ": Trying to open logical channel, whereas we ran out of caps" << VCRIT_END(log);
	   throw XlatorException("Trying to open logical channel, whereas we ran out of caps",
	                         __FILE__, __LINE__);
	 }
       }
       
       // here we should add the association between
       // new assigned sessionID and altcapNumber for
       // which we issue OLC
       
       mediaDescrTable.push_back(altCapN);
       sessionAltCapMap[sID]=altCapN;
       VDEBUG(log) << fn << ": Orig iDone, sID 0; however now - sID : "
            	   << sID
		   << ", altCapN : "
		   << altCapN << VDEBUG_END(log);
     }
     else
     {
       // being a slave we just indicate 0
       sID = 0;
       
       // however, if primary sessions with sessionIDs 1 through 3 are not opened yet
       // we should assign appropriate sessionID
      
       if(maxSesID == 0)
       {
         // have not opened any session yet, so we are going to make OLC
	 // for to open first audio session
	 try
	 {
	   pTypeCap = findFirst('A', altCapN);
	   sID = 1;
	   VDEBUG(log) << fn << ": Here is the pTypeCap we found first : " << pTypeCap << VDEBUG_END(log);
	 }
	 catch (XlatorException)
	 {
	   VCRIT(log) << fn << ": Cannot find the capability from altCapSet in GlobalCapTable" << VCRIT_END(log);
	   throw XlatorException("Cannot find the capability from altCapSet in GlobalCapTable",
	                          __FILE__, __LINE__);
	 }
       }
       else if(maxSesID == 1)
       {
         // we have opened only audio session with sessionID = 1, so we are
	 // going to open video session
	 try
	 {
	   pTypeCap = findFirst('V', altCapN);
	   sID = 2;
	 }
	 catch (XlatorException)
	 {
	   VCRIT(log) << fn << ": Cannot find the capability from altCapSet in GlobalCapTable" << VCRIT_END(log);
	   throw XlatorException("Cannot find the capability from altCapSet in GlobalCapTable",
	                          __FILE__, __LINE__);
	 }
       }
       // currently do not support data
       else
       {
         // here we should find the next altcapset for which we
	 // have not opened the logical channel yet
	 
	 try
	 {
	   pTypeCap = findNext(altCapN);
	 }
	 catch (XlatorException)
	 {
	   VCRIT(log) << fn << ": Trying to open logical channel, whereas we ran out of caps" << VCRIT_END(log);
	   throw XlatorException("Trying to open logical channel, whereas we ran out of caps",
	                          __FILE__, __LINE__);
	 }
       }
       // add altCapN for which we open new OLC to mediaDescrTable
       // we will make the association between this altCapNumber
       // and sID which we will receive from master side in OLCAck
       // message
       
       VDEBUG(log) << fn << ": sID, altCapN, logChNumber : " << sID << ", " << altCapN << ", " << logChNumber << VDEBUG_END(log);
       
       mediaDescrTable.push_back(altCapN);
     }
       // plus we have to make an association between logChNumber
       // we are opening and altCapN
       
       altCapLogChMap[logChNumber] = altCapN;
   }
   else
   {
      // this OLC is for existing session, the last one
      // so we should use current sessionID
      
      
   // NB> should we undertake different actions in the case
   // if we are slave and master
   
      VDEBUG(log) << fn << ": sID : " << sID << VDEBUG_END(log);
      sID = maxSesID;
      
      altCapN = sessionAltCapMap[sID];
      
      try
      {
        // get the cap from a particular sessionDescriptor
        pTypeCap = sDescriptor.getCapability(curDescrN, altCapN);
      }
      catch (XlatorException)
      {
        VERR(log) << fn << ": Did not get the capability from sDescriptor right" << VERR_END(log);
	throw XlatorException("Did not get the capability from sDescriptor right",
	                       __FILE__, __LINE__);
      }

      // should set isDone flag for a given altCapSet
      sDescriptor.transportIDH323.setDone(curDescrN, altCapN);

   // NB> should we associate logChNumber with current altCapN
   // create the map between given logical channel number
   // and the altcapNumber for which we issue OLC
       
   // altCapLogChMap[logChNumber] = altCapN;
   }
     
   msgOLC.m_forwardLogicalChannelNumber = logChNumber;
   
   VDEBUG(log) << fn << ": sID, logChNumber, pTypeCap : " 
    	       << sID << ", " << logChNumber << ", " << pTypeCap << VDEBUG_END(log);
   H323_CapTable::CapTable_I cI = capSetIntersect.h323CapTable.getTableIter_pt(pTypeCap);
   
   if(cI == NULL)
   {
     VERR(log) << fn << ": Did not find the cap with pTypeCap in h323CapTable" << VERR_END(log);
     throw XlatorException("Did not find the cap with pTypeCap in h323CapTable",
                            __FILE__, __LINE__);
   }

   H245_DataType &dataType = msgOLC.m_forwardLogicalChannelParameters.m_dataType;

   Capability *capability = *cI;
   
   if(!capability->getParamSendMsg(dataType))
   {
     VERR(log) << fn << ": Did not get the Parameters correctly to form OLC msg" << VERR_END(log);
     throw XlatorException("Did not get the Parameters correctly to form OLC msg",
                            __FILE__, __LINE__);
   }

   msgOLC.m_forwardLogicalChannelParameters.m_multiplexParameters.
    SetTag(H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters::e_h2250LogicalChannelParameters);
  
   H245_H2250LogicalChannelParameters &params =
     msgOLC.m_forwardLogicalChannelParameters.m_multiplexParameters;
   
   params.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel);

   params.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaGuaranteedDelivery);
   params.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlGuaranteedDelivery);
   params.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_silenceSuppression);
   
   params.m_mediaGuaranteedDelivery = FALSE;
   params.m_mediaControlGuaranteedDelivery = FALSE;
   params.m_silenceSuppression = FALSE;
   params.m_sessionID = sID;
   
   VDEBUG(log) << fn << ": Here is the sessionID we assigned : " << (int)params.m_sessionID << VDEBUG_END(log);

   H245_UnicastAddress_iPAddress mediaControlChannel;

   try
   {
     VDEBUG(log) << fn << ": altCapN, curDescrN for mediaControlChannel : " << altCapN << ' ' << curDescrN << VDEBUG_END(log);
     mediaControlChannel = sDescriptor.transportIDIntSip.getRTCPAddr(curDescrN, altCapN);
   }
   catch (XlatorException)
   {
     VERR(log) << fn << ": Error with setting mediaControlChannel" << VERR_END(log);
     // should never happen
     throw XlatorException("Error with setting mediaControlChannel", __FILE__, __LINE__);
   }

   H245_TransportAddress &trcAddress = params.m_mediaControlChannel;
   trcAddress.SetTag(H245_TransportAddress::e_unicastAddress);
  
   H245_UnicastAddress &ucAddress = trcAddress;
   ucAddress.SetTag(H245_UnicastAddress::e_iPAddress);
   H245_UnicastAddress_iPAddress &ipcAddress = ucAddress;
        
   ipcAddress = mediaControlChannel;

   sentOLC++;
   
   VDEBUG(log) << fn << ": END --------------" << VDEBUG_END(log);

}


bool MSManager::handleOLCAck(const OLCAck &msgOLCAck)
{
   const string     fn("MSManager::handleOLCAck");
   VLog     	    log(fn);
   
   int sID = 0;
   int logChNumber = 0;
   int altCapN = 0;
   std::map<int, int>::const_iterator aI, sI;
   
   VDEBUG(log) << fn << ": BEGIN ------------- " << VDEBUG_END(log);
   
   logChNumber = msgOLCAck.m_forwardLogicalChannelNumber;
   
   if(!msgOLCAck.HasOptionalField(
      H245_OpenLogicalChannelAck::e_forwardMultiplexAckParameters))
   {
     VERR(log) << fn << ": OLCAck without forwardMultiplexParameters" << VERR_END(log);
     return false;
   }
   
   if(msgOLCAck.m_forwardMultiplexAckParameters.GetTag() !=
     H245_OpenLogicalChannelAck_forwardMultiplexAckParameters::e_h2250LogicalChannelAckParameters)
   {
     VERR(log) << fn << ": no h2250LogicalChannelAckParameters" << VERR_END(log);
     return false;
   }
   
   const H245_H2250LogicalChannelAckParameters &params =
     msgOLCAck.m_forwardMultiplexAckParameters;
   
   if(!params.HasOptionalField(
     H245_H2250LogicalChannelAckParameters::e_sessionID))sID = 0;
   else
   {
     sID = params.m_sessionID;
   }
   

   if(master)
   {
     aI = altCapLogChMap.find(logChNumber);
     
     if(aI == altCapLogChMap.end())
     {
       VERR(log) << fn << ": Did not open LC with this logChNumber - invalid OLCAck" << VERR_END(log);
       return false;
     }
     
     // find association between logChNumber and altCapN
     altCapN = altCapLogChMap[logChNumber];
   }
   else
   { // we are slave - hate when people manipulate me :.(
     if(sID == 0)
     {
       // did not receive sessionID in OLCAck, therefore this OLCAck
       // is the last in a session
       
       aI = altCapLogChMap.find(logChNumber);
       if(aI == altCapLogChMap.end())
       {
         VERR(log) << fn << ": Did not open LC with this logChNumber - invalid OLCAck" << VERR_END(log);
	 return false;
       }
       
       // find association between logChNumber and altCapN
       altCapN = altCapLogChMap[logChNumber];
     }
     else
     {
       // did receive sessionID in OLCAck
       sI = sessionAltCapMap.find(sID);
     
       if(sI == sessionAltCapMap.end())
       {
         // given sID is not associated with any particular altCapN
	 // therefore we tend to interprete this situation as
	 // we sent OLC first with sID set to 0 (we are slave)
	 // therefore this OLCAck is the first in a session
	 // with valid sID
	 
	 aI = altCapLogChMap.find(logChNumber);
	 if(aI == altCapLogChMap.end())
	 {
	   VERR(log) << fn << ": Did not open LC with this logChNumber" << VERR_END(log);
	   return false;
	 }
	 
	 // find association between logChNumber and altCapN
	 altCapN = altCapLogChMap[logChNumber];
	 
	 // associate just received valid sessionID with a given altCapN
	 sessionAltCapMap[altCapN] = sID;
	 
	 // we need to adjust maxSesID since we received valid sessionID
	 // from master
	 if(maxSesID < sID)maxSesID = sID;
       }
       else
       {
         // sID is already associated with some particular altCapN
	 // this is the last OLCAck message in a current session
	 
	 altCapN = altCapLogChMap[logChNumber];
       }
       
       if(find(mediaDescrTable.begin(), mediaDescrTable.end(), altCapN) ==
          mediaDescrTable.end())
       {
         VERR(log) << fn << ": Hey, we forgot to add altCapN to mediaDescrTable. " 
	           << "Probably, on the stage of makeOLC" << VERR_END(log);
       }
     }
   }

   if(!params.HasOptionalField(
     H245_H2250LogicalChannelAckParameters::e_mediaChannel))
   {
     VERR(log) << fn << ": no mediaChannel - no RTP address" << VERR_END(log);
     return false;
   }

    const H245_TransportAddress &trAddress = 
      params.m_mediaChannel;

    if(trAddress.GetTag() == H245_TransportAddress::e_unicastAddress)
    {
      const H245_UnicastAddress &uAddress = trAddress;
      if(uAddress.GetTag() == H245_UnicastAddress::e_iPAddress)
      {
        const H245_UnicastAddress_iPAddress &ipAddress = uAddress;
        
        Address address(ipAddress);   
   
 	  try
          {
            // Setting RTP host/port, currently use only first (# 0) capability descriptor
            sDescriptor.transportIDH323.setRTPAddr(address, curDescrN, altCapN);
          }
          catch (XlatorException)
          {
            VERR(log) << fn << ": Something wrong with setting RTP address..." << VERR_END(log);
	    // basically should never happen
	    return false;
          }
      }
    }
    else
    {
      VERR(log) << fn << ": Address is not supported" << VERR_END(log);
      return false;
    }
    

   if(!params.HasOptionalField(
     H245_H2250LogicalChannelAckParameters::e_mediaControlChannel))
   {
    const H245_TransportAddress &trAddress = 
      params.m_mediaControlChannel;

    if(trAddress.GetTag() == H245_TransportAddress::e_unicastAddress)
    {
      const H245_UnicastAddress &uAddress = trAddress;
      if(uAddress.GetTag() == H245_UnicastAddress::e_iPAddress)
      {
        const H245_UnicastAddress_iPAddress &ipAddress = uAddress;
        
        Address address(ipAddress);   
   
 	try
        {
            // Setting RTCP host/port, currently use only first (# 0) capability descriptor
            sDescriptor.transportIDH323.setRTCPAddr(address, curDescrN, altCapN);
        }
        catch (XlatorException)
        {
            VERR(log) << fn << ": Something wrong with setting RTCP address..." << VERR_END(log);
	    // basically should never happen
	    return false;
        }
      }
    }
   }

   receivedOLCAck++;
   VDEBUG(log) << fn << ": END ------------- " << VDEBUG_END(log);
   return true;
}

