/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const H323Gatekeeper_cxx_Version = 
    "$Id: H323Gatekeeper.cxx,v 1.15 2001/11/07 08:57:15 icahoon Exp $";


#include "H323Gatekeeper.hxx"
#include "GatekeeperInfo.hxx"
#include "EndpointInfoMap.hxx"
#include "RASPolicyBuilder.hxx"
#include "RASMCastGKProtocol.hxx"
#include "RASGKProtocol.hxx"
#include "DiscoveryPolicy.hxx"
#include "RegistrationPolicy.hxx"
#include "AdmissionPolicy.hxx"
#include "LocationPolicy.hxx"
#include "ResourcesAvailablePolicy.hxx"
#include "VLog.hxx"
#include <string>


using namespace Vocal;
using Vocal::H323GK::RASPolicyBuilder;
using Vocal::H323GK::RASMCastGKProtocol;
using Vocal::H323GK::RASGKProtocol;
using Vocal::H323GK::DiscoveryPolicy;
using Vocal::H323GK::RegistrationPolicy;
using Vocal::H323GK::AdmissionPolicy;
using Vocal::H323GK::LocationPolicy;
using Vocal::H323GK::ResourcesAvailablePolicy;
using Vocal::SipH323Info::GatekeeperInfo;
using Vocal::SipH323Info::EndpointInfoMap;
using Vocal::SipH323Info::EndpointIdMap;
using Vocal::SipH323Info::EndpointInfo;
using Vocal::Services::ControlEvent;
using Vocal::Services::Event;
using Vocal::Logging::VLog;


Vocal::H323GK::H323Gatekeeper::H323Gatekeeper(
    RASPolicyBuilder    	&   rasPolicyBuilder,
    GatekeeperInfo	    	&   gatekeeperInfo
)
    :	PollFifoService("H323Gatekeeper"),
    	myRasPolicyBuilder(rasPolicyBuilder),
	myGatekeeperInfo(gatekeeperInfo),
    	myRasMcastGkProtocol(0),
    	myRasGkProtocol(0),
    	myDiscoveryPolicy(0),
    	myRegistrationPolicy(0),
	myAdmissionPolicy(0),
        myLocationPolicy(0),
        myResourcesAvailablePolicy(0)
{
    const string    	fn("H323Gatekeeper::H323Gatekeeper");
    VLog    	    	log(fn);

    // Add the statically registered endpoints to the endpoint map.
    //
    list< Sptr<EndpointInfo> >  gws = myGatekeeperInfo.staticGateways();
    
    for ( list< Sptr<EndpointInfo> >::const_iterator i = gws.begin();
          i != gws.end();
          ++i
        )
    {
        myEndpointInfoMap.insert(*i);
        
        VDEBUG(log) << fn << ": inserting static gateway = { "
                    << *(*i) << " }" << VDEBUG_END(log);
    }
    
    VDEBUG(log) << fn << ": created." << VDEBUG_END(log);
}


Vocal::H323GK::H323Gatekeeper::~H323Gatekeeper()
{
    const string    	fn("H323Gatekeeper::~H323Gatekeeper");
    VLog    	    	log(fn);
    
    VDEBUG(log) << fn << ": destroyed." << VDEBUG_END(log);

    clear();
}


ReturnCode
Vocal::H323GK::H323Gatekeeper::onStart()
{
    const string    	fn("H323Gatekeeper::onStart");
    VLog    	    	log(fn);
    
    VDEBUG(log) << fn << ": starting." << VDEBUG_END(log);

    ReturnCode	rc = SUCCESS;
    
    try
    {
	if ( !myRasMcastGkProtocol )
	{
    	    myRasMcastGkProtocol = new RASMCastGKProtocol(myGatekeeperInfo.rasMulticastAddress());
	}
	if ( !myRasGkProtocol )
	{
    	    myRasGkProtocol = new RASGKProtocol(*this, myGatekeeperInfo.rasAddress());
	}

	if ( myGatekeeperInfo.discoveryEnabled() && !myDiscoveryPolicy )
	{
    	    myDiscoveryPolicy 
	    	= myRasPolicyBuilder.buildDiscoveryPolicy(
		    	    	    	*this,
		    	    	    	*myRasGkProtocol,
					myGatekeeperInfo,
					myEndpointInfoMap);


	    myRasMcastGkProtocol->setDiscoveryPolicy(*myDiscoveryPolicy);
	    myRasGkProtocol->setDiscoveryPolicy(*myDiscoveryPolicy);
	}

	if ( myGatekeeperInfo.registrationEnabled() && !myRegistrationPolicy )
	{
    	    myRegistrationPolicy 
	    	= myRasPolicyBuilder.buildRegistrationPolicy(
		    	    	    	*this,
    	    	    	    	    	*myRasGkProtocol,
					myGatekeeperInfo,
					myEndpointInfoMap);		

	    myRasGkProtocol->setRegistrationPolicy(*myRegistrationPolicy);
	}

	if ( myGatekeeperInfo.admissionEnabled() && !myAdmissionPolicy )
	{
    	    myAdmissionPolicy 
	    	= myRasPolicyBuilder.buildAdmissionPolicy(
		    	    	    	*this,
    	    	    	    	    	*myRasGkProtocol,
					myGatekeeperInfo,
					myEndpointInfoMap);		
	
            myRasGkProtocol->setAdmissionPolicy(*myAdmissionPolicy);
	}

	if ( myGatekeeperInfo.locationEnabled() && !myLocationPolicy )
	{
    	    myLocationPolicy 
	    	= myRasPolicyBuilder.buildLocationPolicy(
		    	    	    	*this,
    	    	    	    	    	*myRasGkProtocol,
					myGatekeeperInfo,
					myEndpointInfoMap);		
	
            myRasMcastGkProtocol->setLocationPolicy(*myLocationPolicy);
            myRasGkProtocol->setLocationPolicy(*myLocationPolicy);
	}

	if  (   myGatekeeperInfo.resourcesAvailableEnabled() 
            &&  !myResourcesAvailablePolicy 
            )
	{
    	    myResourcesAvailablePolicy 
	    	= myRasPolicyBuilder.buildResourcesAvailablePolicy(
		    	    	    	*this,
    	    	    	    	    	*myRasGkProtocol,
					myGatekeeperInfo,
					myEndpointInfoMap);		
	
            myRasGkProtocol->setResourcesAvailablePolicy(*myResourcesAvailablePolicy);
	}

	registerProtocol(*myRasMcastGkProtocol);
	registerProtocol(*myRasGkProtocol);
    }
    catch ( ... )
    {
    	rc = !SUCCESS;
    }

    return ( rc );    
}


ReturnCode
Vocal::H323GK::H323Gatekeeper::onStop()
{
    const string    	fn("H323Gatekeeper::onStop");
    VLog    	    	log(fn);

    VDEBUG(log) << fn << ": stopping." << VDEBUG_END(log);

    return ( SUCCESS );
}


ReturnCode	    
Vocal::H323GK::H323Gatekeeper::onShutdown()
{
    const string    	fn("H323Gatekeeper::onShutdown");
    VLog    	    	log(fn);
    
    VDEBUG(log) << fn << ": shutting down." << VDEBUG_END(log);

    EndpointIdMap  endpointIdMap = myEndpointInfoMap.endpointIdMap();
    H225_UnregRequestReason reason(H225_UnregRequestReason::e_undefinedReason);
    
    for (   EndpointIdMap::iterator it = endpointIdMap.begin();
    	    it != endpointIdMap.end();
	    it++
	)
    {
	Sptr<EndpointInfo> endpointInfo = it->second;
	Sptr<IPAddress> remoteAddr = new IPAddress(endpointInfo->rasAddress());
	
        if ( !endpointInfo->staticRegistration() )
        {
    	    myRegistrationPolicy->unregister(reason, endpointInfo, remoteAddr);
        }
    }

    try
    {
        clear();
    }
    catch ( ... )
    {
    	VWARN(log)  << fn << ": exception on shutdown ." << VWARN_END(log);
    }

    myEndpointInfoMap.clear();

    return ( SUCCESS );
}


void	    
Vocal::H323GK::H323Gatekeeper::clear()
{
    if ( myRasMcastGkProtocol )
    {
        unregisterProtocol(*myRasMcastGkProtocol);
    }
    if ( myRasGkProtocol )
    {
        unregisterProtocol(*myRasGkProtocol);
    }

    delete myRasMcastGkProtocol;
    myRasMcastGkProtocol = 0;
    
    delete myRasGkProtocol;
    myRasGkProtocol = 0;

    delete myResourcesAvailablePolicy;    
    myResourcesAvailablePolicy = 0;
    
    delete myLocationPolicy;
    myLocationPolicy = 0;
    
    delete myAdmissionPolicy;
    myAdmissionPolicy = 0;

    delete myRegistrationPolicy;
    myRegistrationPolicy = 0;

    delete myDiscoveryPolicy;
    myDiscoveryPolicy = 0;
}
