#if !defined(VOCAL_RAS_GK_PROTOCOL_HXX)
#define VOCAL_RAS_GK_PROTOCOL_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const RASGKProtocol_hxx_Version = 
    "$Id: RASGKProtocol.hxx,v 1.11 2001/10/13 01:15:53 icahoon Exp $";


#include "ptlib.h"
#include "h323pdu.h"
#include "Protocol.hxx"
#include "UDPSocket.hxx"
#include "IPAddress.hxx"
#include "Sptr.hxx"
#include <queue>
#include <strstream>


// Forward declarations
//
namespace Vocal {
    namespace Services {
    	class PollFifoService;
    }
}


/** Vovida Open Communication Application Library<br><br>
 */
namespace Vocal
{


/** H.323 Gatekeeper<br><br>
 */
namespace H323GK
{


using Vocal::Services::PollFifoService;
using Vocal::Transport::UDPSocket;
using Vocal::Transport::IPAddress;
using Vocal::IO::FileDescriptor;


class DiscoveryPolicy;
class RegistrationPolicy;
class AdmissionPolicy;
class LocationPolicy;
class ResourcesAvailablePolicy;

/** The protocol is used to receive, decode, encode and send RAS 
 *  messages.<br><br>
 *
 *  The protocol takes the incoming message from it's file descriptor
 *  and decodes the message, forwarding in on to the appropriate policy.
 *
 *  The protocol also takes the outgoing messages, encodes it, and sends
 *  it to the appropriate remote address.
 *
 *  @see    Vocal::Transport::Protocol
 *  @see    Vocal::Services::PollFifoService
 *  @see    Vocal::H323GK::DiscoveryPolicy
 *  @see    Vocal::H323GK::RegistrationPolicy
 *  @see    Vocal::H323GK::AdmissionPolicy
 *  @see    Vocal::H323GK::LocationPolicy
 *  @see    Vocal::H323GK::ResourcesAvailablePolicy
 *  @see    Vocal::IO::FileDescriptor
 *  @see    Vocal::Transport::IPAddress
 *  @see    Vocal::Transport::UDPSocket
 */
class RASGKProtocol : public Vocal::Transport::Protocol
{

    public:


    	/** Create given the poll fifo service to register with and
	 *  the ras address.
    	 */
	RASGKProtocol(
	    PollFifoService    	    	&   pollFifoService,
	    const IPAddress     	&   rasAddress
	);


    	/** Virtual destructor
    	 */
    	virtual ~RASGKProtocol();


    	/** Set the discovery policy. When incoming discovery messages arrive,
	 *  they will be forwarded to the discovery policy.
    	 */
    	void    setDiscoveryPolicy(DiscoveryPolicy &);


    	/** Set the registration policy. When incoming registration messages 
	 *  arrive, they will be forwarded to the registration policy.
    	 */
    	void    setRegistrationPolicy(RegistrationPolicy &);


    	/** Set the admission policy. When incoming admission messages arrive, 
	 *  they will be forwarded to the admission policy.
    	 */
    	void    setAdmissionPolicy(AdmissionPolicy &);
	

    	/** Set the location policy. When incoming location messages arrive, 
	 *  they will be forwarded to the location policy.
    	 */
    	void    setLocationPolicy(LocationPolicy &);
	

    	/** Set the resources available policy. When incoming admission 
         *  messages arrive, they will be forwarded to the resources available 
         *  policy.
    	 */
    	void    setResourcesAvailablePolicy(ResourcesAvailablePolicy &);
	

    	/** Access the FileDescriptor associated with this protocol.
    	 */
    	const FileDescriptor &	    getFileDescriptor() const;


    	/** Protocol callback. When an incoming message arrives, this protocol
	 *  will decode the RAS message and forward it to the appropriate
	 *  policy.
    	 */
    	void	    	    	    onIncomingAvailable();


    	/** Protocol callback. When the file descriptor is ready to be
	 *  written to, the queue of outgoing messages will be sent.
    	 */
	void	    	    	    onOutgoingAvailable();


    	/** Send the given GCF to the remote address.
    	 */
    	void	    	    	    discoveryConfirm(
	    	    	    	    	H323RasPDU  	&   gcf,
					Sptr<IPAddress>     remoteAddr,
					ostrstream  	&   confirmLog);
					

    	/** Send the given GRJ to the remote address.
    	 */
	void	    	    	    discoveryReject(
	    	    	    	    	H323RasPDU  	&   grj,
					Sptr<IPAddress>     remoteAddr);
					

    	/** Send the given RCF to the remote address.
    	 */
    	void	    	    	    registrationConfirm(
	    	    	    	    	H323RasPDU  	&   rcf,
					Sptr<IPAddress>     remoteAddr,
					ostrstream  	&   confirmLog);


    	/** Send the given RRJ to the remote address.
    	 */
	void	    	    	    registrationReject(
	    	    	    	    	H323RasPDU  	&   rrj,
					Sptr<IPAddress>     remoteAddr);


    	/** Send the given URQ to the remote address.
    	 */
    	void	    	    	    unregistrationRequest(	
	    	    	    	    	H323RasPDU  	&   urq,
					Sptr<IPAddress>     remoteAddr);


    	/** Send the given UCF to the remote address.
    	 */
    	void	    	    	    unregistrationConfirm(	
	    	    	    	    	H323RasPDU  	&   ucf,
					Sptr<IPAddress>     remoteAddr,
					ostrstream  	&   confirmLog);


    	/** Send the given URJ to the remote address.
    	 */
    	void	    	    	    unregistrationReject(	
	    	    	    	    	H323RasPDU  	&   urj,
					Sptr<IPAddress>     remoteAddr);


    	/** Send the given ACF to the remote address.
    	 */
	void	    	    	    admissionConfirm(
	    	    	    	    	H323RasPDU  	&   acf,
					Sptr<IPAddress>     remoteAddr,
					ostrstream  	&   confirmLog);
					

    	/** Send the given ARJ to the remote address.
    	 */
	void	    	    	    admissionReject(
	    	    	    	    	H323RasPDU  	&   arj,
					Sptr<IPAddress>     remoteAddr);


    	/** Send the given DRQ to the remote address.
    	 */
	void	    	    	    disengageRequest(
	    	    	    	    	H323RasPDU  	&   drq,
					Sptr<IPAddress>     remoteAddr);


    	/** Send the given DCF to the remote address.
    	 */
	void	    	    	    disengageConfirm(
	    	    	    	    	H323RasPDU  	&   dcf,
					Sptr<IPAddress>     remoteAddr,
					ostrstream  	&   confirmLog);


    	/** Send the given DRJ to the remote address.
    	 */
	void	    	    	    disengageReject(
	    	    	    	    	H323RasPDU  	&   drj,
					Sptr<IPAddress>     remoteAddr);
                                        

        /** Send the given LRQ to the remote address.
         */
        void                        locationRequest(
                                        H323RasPDU      &   lrq,
                                        Sptr<IPAddress>     remoteAddr);
                                        

        /** Send the given LCF to the remote address.
         */
        void                       locationConfirm(
                                        H323RasPDU      &   lcf,
                                        Sptr<IPAddress>     remoteAddr,
					ostrstream  	&   confirmLog);


        /** Send the given LRJ to the remote address.
         */
        void                       locationReject(
                                        H323RasPDU      &   lrj,
                                        Sptr<IPAddress>     remoteAddr);


        /** Send the given RAC to the remote address.
         */
        void                       resourcesAvailableConfirm(
                                        H323RasPDU      &   lrj,
                                        Sptr<IPAddress>     remoteAddr);
            

    private:

	struct	OutgoingMessage
	{
	    Sptr<PPER_Stream>	outgoingStream;
	    Sptr<IPAddress> 	remoteAddr;
	};

	void	    	    	    sendMessage(
    	    	    	    		H323RasPDU      &   gcf,
					Sptr<IPAddress>     remote);

    	PollFifoService	    	    & 	myPollFifoService;
	
   	DiscoveryPolicy     	    *  	myDiscoveryPolicy;
	RegistrationPolicy  	    *  	myRegistrationPolicy;
	AdmissionPolicy     	    *  	myAdmissionPolicy;
        LocationPolicy              *   myLocationPolicy;
        ResourcesAvailablePolicy    *   myResourcesAvailablePolicy;

	IPAddress        	    	myRasAddress;
    	Sptr<UDPSocket>     	    	myRasSocket;
	Sptr<PPER_Stream> 	    	myInPerStream;
	size_t	    	    	    	myInPerStreamSize;

    	queue< Sptr<OutgoingMessage> >	myOutgoingMessages;
};


} // namespace H323GK
} // namespace Vocal


#endif // !defined(VOCAL_RAS_GK_PROTOCOL_HXX)
