/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const EndpointInfo_cxx_Version = 
    "$Id: EndpointInfo.cxx,v 1.18 2001/11/08 02:47:26 icahoon Exp $";


#include "EndpointInfo.hxx"
#include "VLog.hxx"
#include <values.h>


using Vocal::SipH323Info::EndpointInfo;
using Vocal::Transport::IPAddress;
using Vocal::Logging::VLog;
using Vocal::Threads::Lock;
using Vocal::Services::Event;
using Vocal::FifoBase;


EndpointInfo::EndpointInfo()
    :	myEndpointIdentifier(""),
        myEndpointType(UNDEFINED_TYPE),
        myStripPrefix(false),
        myStaticRegistration(false),
    	myTimeToLive(0),
	myTimeToLiveTimerEventId(0),
        myAdmissions(0),
        myOutOfResources(false)
{
}


EndpointInfo::EndpointInfo(
    H225_RegistrationRequest    &	rrq
)
    :	myEndpointIdentifier(""),
        myEndpointType(UNDEFINED_TYPE),
        myStripPrefix(false),
        myStaticRegistration(false),
    	myTimeToLive(0),
	myTimeToLiveTimerEventId(0),
        myAdmissions(0),
        myOutOfResources(false)
{
    NTS_update(rrq);
}


EndpointInfo::EndpointInfo(const EndpointInfo & src)
{
    Lock    lock(src.myMutex); (void)lock;
    
    NTS_endpointIdentifier(src.NTS_endpointIdentifier());
    NTS_endpointType(src.NTS_endpointType());
    NTS_alias(src.NTS_alias());
    NTS_prefix(src.NTS_prefix());
    NTS_stripPrefix(src.NTS_stripPrefix());
    NTS_staticRegistration(src.NTS_staticRegistration());
    NTS_rasAddress(src.NTS_rasAddress());
    NTS_callSignalAddress(src.NTS_callSignalAddress());
    NTS_timeToLive(src.NTS_timeToLive());
    NTS_timeToLiveTimerEventId(0);

    myLastRegistration = src.myLastRegistration;
    myAdmissions = src.myAdmissions;
    myOutOfResources = src.myOutOfResources;
}


EndpointInfo::~EndpointInfo()
{
}


EndpointInfo &
EndpointInfo::operator=(const EndpointInfo & src)
{
    Lock    lock(myMutex); (void)lock;

    if ( this != &src )
    {
    	Lock    lock(src.myMutex); (void)lock;
	
	NTS_endpointIdentifier(src.NTS_endpointIdentifier());
        NTS_endpointType(src.NTS_endpointType());
	NTS_alias(src.NTS_alias());
        
        // For now, static registration only means that the prefix
        // doesn't get changed.
        //
        if ( !myStaticRegistration )
        {
            NTS_staticRegistration(src.NTS_staticRegistration());
            NTS_prefix(src.NTS_prefix());
            NTS_stripPrefix(src.NTS_stripPrefix());
        }

	NTS_rasAddress(src.NTS_rasAddress());
	NTS_callSignalAddress(src.NTS_callSignalAddress());
	NTS_timeToLive(src.NTS_timeToLive());
	myLastRegistration = src.myLastRegistration;
        myAdmissions = src.myAdmissions;
        myOutOfResources = src.myOutOfResources;
    }
    return ( *this );
}


bool	    	
EndpointInfo::update(
    H225_RegistrationRequest        &	rrq
)
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_update(rrq) );
}


bool	    	
EndpointInfo::NTS_update(
    H225_RegistrationRequest        &	rrq
)
{
    // Get the alias.
    //
    NTS_alias(rrq.m_terminalAlias);

    // Get the ras address. Look for the first IPAddress.
    //
    IPAddress rasAddr;
    
    for ( size_t i = 0; i < rrq.m_rasAddress.GetSize(); i++ )
    {
        if ( rrq.m_rasAddress[i].GetTag() == H225_TransportAddress::e_ipAddress )
        {
    	    H225_TransportAddress_ipAddress & rasIP = rrq.m_rasAddress[i];
            rasAddr.setIP(rasIP.m_ip.GetPointer());
	    rasAddr.setPort(rasIP.m_port);
	    NTS_rasAddress(rasAddr);
	    break;
        }
    }
    
    // Pull the endpointIdentifier out of the rrq.
    //
    if ( rrq.HasOptionalField(H225_RegistrationRequest::e_endpointIdentifier) )
    {
    	string eptId;
	eptId = rrq.m_endpointIdentifier.GetValue();
    	NTS_endpointIdentifier(eptId);
    }

    // Pull the endpointType out of the rrq.
    //
    NTS_endpointType(UNDEFINED_TYPE);
    
    if ( rrq.m_terminalType.HasOptionalField(H225_EndpointType::e_terminal) )
    {
        NTS_endpointType(TERMINAL_TYPE);
    }
    if ( rrq.m_terminalType.HasOptionalField(H225_EndpointType::e_gateway) )
    {
        NTS_endpointType(GATEWAY_TYPE);
    }
    
    // If this is a gateway endpoint and this endpoint type wasn't created
    // statically, pull out the supported prefix.
    //
    if  (   NTS_endpointType() == GATEWAY_TYPE 
        &&  !NTS_staticRegistration()
        )
    {
        NTS_prefix(rrq.m_terminalType.m_gateway);
    }
    
    // Get the callSignalAddress.
    //
    IPAddress csAddr;
    
    for ( size_t i = 0; i < rrq.m_callSignalAddress.GetSize(); i++ )
    {
        if ( rrq.m_callSignalAddress[i].GetTag() == H225_TransportAddress::e_ipAddress )
        {
    	    H225_TransportAddress_ipAddress & csIP = rrq.m_callSignalAddress[i];
            csAddr.setIP(csIP.m_ip.GetPointer());
	    csAddr.setPort(csIP.m_port);
	    NTS_callSignalAddress(csAddr);
	    break;
        }
    }
    
    // Get the timeToLive out of the rrq. If they do not want to
    // do keep alive, we will set this value to 0.
    //
    if ( rrq.HasOptionalField(H225_RegistrationRequest::e_timeToLive) )
    {
    	NTS_timeToLive(rrq.m_timeToLive);
    }

    NTS_updateLastRegistration();

    return ( NTS_valid() );
}


bool	    	
EndpointInfo::valid() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_valid() );
}


bool	    	
EndpointInfo::NTS_valid() const
{
    // If endpoint type is a terminal or undefined, there must be 
    // a phone number to be valid.
    //
    if  ( (myEndpointType == TERMINAL_TYPE || myEndpointType == UNDEFINED_TYPE)
        &&  !myAlias.hasPhone() 
        )
    {
        return ( false );
    }
    
    // Must have an endpoint identifier to be valid.
    //
    if ( myEndpointIdentifier.size() == 0 )
    {
        return ( false );
    }
    
    // Must either be statically registered or have a RAS address
    // to be valid.
    //
    if( !myStaticRegistration && myRasAddress.isAny() )
    {
        return ( false );
    }
    
    // Must have a call signaling address to be valid.
    //
    if ( myCallSignalAddress.isAny() )
    {
        return ( false );
    }
    
    return ( true );
}


void	    	
EndpointInfo::endpointIdentifier(const string & endptId)
{
    Lock    lock(myMutex); (void)lock;
    NTS_endpointIdentifier(endptId);
}


void	    	
EndpointInfo::NTS_endpointIdentifier(const string & endptId)
{
    const string    fn("EndpointInfo::endpointIdentifier");
    VLog    	    log(fn);

    myEndpointIdentifier = endptId;

    VDEBUG(log) << fn << ": " << myEndpointIdentifier << VDEBUG_END(log);
}

string
EndpointInfo::endpointIdentifier() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_endpointIdentifier() );
}

const string &
EndpointInfo::NTS_endpointIdentifier() const
{
    return ( myEndpointIdentifier );
}


void            
EndpointInfo::endpointType(Type type)
{
    Lock    lock(myMutex); (void)lock;
    NTS_endpointType(type);
}
        

void            
EndpointInfo::NTS_endpointType(Type type)
{
    const string    fn("EndpointInfo::endpointType");
    VLog    	    log(fn);

    myEndpointType = type;

    VDEBUG(log) << fn << ": " << TypeName(myEndpointType)
                << VDEBUG_END(log);
}


const char *    
EndpointInfo::TypeName(Type type)
{
    static const char * typeName[TYPE_SIZE] =
    {
        "Undefined",
        "Terminal",
        "Gateway",
        "MCU"
    };
    return ( typeName[type >= TYPE_SIZE ? UNDEFINED_TYPE : type] );
};


EndpointInfo::Type            
EndpointInfo::endpointType() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_endpointType() );
}
        

EndpointInfo::Type            
EndpointInfo::NTS_endpointType() const
{
    return ( myEndpointType );
}


void	    	
EndpointInfo::alias(const H225_ArrayOf_AliasAddress & pAlias)
{
    Lock    lock(myMutex); (void)lock;
    NTS_alias(pAlias);
}


void
EndpointInfo::NTS_alias(const H225_ArrayOf_AliasAddress & pAlias)
{
    const string    fn("EndpointInfo::alias");
    VLog    	    log(fn);
    
    myAlias.alias(pAlias);

    VDEBUG(log) << fn << ": " << myAlias << VDEBUG_END(log);
}


H225_ArrayOf_AliasAddress
EndpointInfo::alias() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_alias() );
}


H225_ArrayOf_AliasAddress
EndpointInfo::NTS_alias() const
{
    return ( myAlias.alias() );
}


void	    	
EndpointInfo::prefix(const H225_GatewayInfo & gwInfo)
{
    Lock    lock(myMutex); (void)lock;
    NTS_prefix(gwInfo);
}


void	    	
EndpointInfo::prefix(const list<string> & prefixes)
{
    H225_ArrayOf_AliasAddress   aliases;
    
    for ( list<string>::const_iterator  i = prefixes.begin();
          i != prefixes.end();
          ++i
        )
    {
        PINDEX  size = aliases.GetSize();
        aliases.SetSize(size + 1);

        aliases[size].SetTag(H225_AliasAddress::e_e164);

        PASN_IA5String  &   alias = aliases[size];
        alias = i->c_str();
    }
    
    Lock    lock(myMutex); (void)lock;
    NTS_prefix(aliases);
}


void	    	
EndpointInfo::NTS_prefix(const H225_ArrayOf_AliasAddress & pPrefix)
{
    const string    fn("EndpointInfo::prefix");
    VLog    	    log(fn);
    
    myPrefix = pPrefix;

    VDEBUG(log) << fn << ": " << myPrefix << VDEBUG_END(log);
}


void
EndpointInfo::NTS_prefix(const H225_GatewayInfo & gwInfo)
{
    const string    fn("EndpointInfo::prefix");
    VLog    	    log(fn);

    // Gateway info is a sequence of SupportedProtocols.
    // Each SupportedProtocol is a choice of capabilities.
    // Each capability has a sequence of SupportedPrefix.
    // A SupportedPrefix has a single alias for the prefix.
    //
    // I wish I knew why they are trying to make this as hard
    // as TCS. I really should ask Glen F. about this. *grin*
    //
    // Right now we are only going to deal with the voice caps,
    // but ideally we'll accept any capability, since we don't
    // really care what they talk on the other side of the GW at 
    // this point. We just need to know where to route the call 
    // based on the alias. Also, we route based on e.164 only. 
    // Perhaps we could route based on other fields in the alias, 
    // but there are better solutions for that... i.e. DNS and
    // DNS SRV, etc.
    //
    for ( size_t i = 0; i < gwInfo.m_protocol.GetSize(); i++ )
    {
        const H225_SupportedProtocols   &   supportedProtocols 
                                                = gwInfo.m_protocol[i];

        switch ( supportedProtocols.GetTag() )
        {
            case H225_SupportedProtocols::e_voice:
            {
                const H225_VoiceCaps & voiceCaps = supportedProtocols;

                if ( voiceCaps.HasOptionalField(H225_VoiceCaps::e_supportedPrefixes) )
                {
                    for ( size_t i = 0; i < voiceCaps.m_supportedPrefixes.GetSize(); i++ )
                    {
                        PINDEX  size = myPrefix.GetSize();
                        myPrefix.SetSize(size + 1);
                        
                        myPrefix[size] = voiceCaps.m_supportedPrefixes[i].m_prefix;
                    }
                }
                break;
            }

            case H225_SupportedProtocols::e_h323:
            {
                const H225_H323Caps & h323Caps = supportedProtocols;

                if ( h323Caps.HasOptionalField(H225_H323Caps::e_supportedPrefixes) )
                {
                    for ( size_t i = 0; i < h323Caps.m_supportedPrefixes.GetSize(); i++ )
                    {
                        PINDEX  size = myPrefix.GetSize();
                        myPrefix.SetSize(size + 1);
                        
                        myPrefix[size] = h323Caps.m_supportedPrefixes[i].m_prefix;
                    }
                }
                break;
            }

            case H225_SupportedProtocols::e_h320:
            {
                const H225_H320Caps & h320Caps = supportedProtocols;

                if ( h320Caps.HasOptionalField(H225_H320Caps::e_supportedPrefixes) )
                {
                    for ( size_t i = 0; i < h320Caps.m_supportedPrefixes.GetSize(); i++ )
                    {
                        PINDEX  size = myPrefix.GetSize();
                        myPrefix.SetSize(size + 1);
                        
                        myPrefix[size] = h320Caps.m_supportedPrefixes[i].m_prefix;
                    }
                }
                break;
            }
            
            // Do these later... perhaps.
            //
            case H225_SupportedProtocols::e_h310:
            case H225_SupportedProtocols::e_h321:
            case H225_SupportedProtocols::e_h322:
            case H225_SupportedProtocols::e_h324:
            case H225_SupportedProtocols::e_t120_only:

            case H225_SupportedProtocols::e_nonStandardProtocol:
            case H225_SupportedProtocols::e_nonStandardData:
            default:
            {
                // Skip it by default
                //
                break;
            }
        }
    }

    VDEBUG(log) << fn << ": " << myPrefix << VDEBUG_END(log);
}


H225_ArrayOf_AliasAddress
EndpointInfo::prefix() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_prefix() );
}


H225_ArrayOf_AliasAddress
EndpointInfo::NTS_prefix() const
{
    return ( myPrefix );
}


void            
EndpointInfo::stripPrefix(bool strip_prefix)
{
    Lock    lock(myMutex); (void)lock;
    NTS_stripPrefix(strip_prefix);
}
        

void            
EndpointInfo::NTS_stripPrefix(bool strip_prefix)
{
    myStripPrefix = strip_prefix;
}


bool            
EndpointInfo::stripPrefix() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_stripPrefix() );
}


bool            
EndpointInfo::NTS_stripPrefix() const
{
    return ( myStripPrefix );
}


void            
EndpointInfo::staticRegistration(bool static_reg)
{
    Lock    lock(myMutex); (void)lock;
    NTS_staticRegistration(static_reg);
}
        

void            
EndpointInfo::NTS_staticRegistration(bool static_reg)
{
    myStaticRegistration = static_reg;
}


bool            
EndpointInfo::staticRegistration() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_staticRegistration() );
}


bool            
EndpointInfo::NTS_staticRegistration() const
{
    return ( myStaticRegistration );
}


void	    	
EndpointInfo::rasAddress(const IPAddress & rasAddr)
{
    Lock    lock(myMutex); (void)lock;
    NTS_rasAddress(rasAddr);
}


void	    	
EndpointInfo::NTS_rasAddress(const IPAddress & rasAddr)
{
    const string    fn("EndpointInfo::rasAddress");
    VLog    	    log(fn);

    myRasAddress = rasAddr;
    
    VDEBUG(log) << fn << ": " << myRasAddress << VDEBUG_END(log);
}


IPAddress
EndpointInfo::rasAddress() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_rasAddress() );
}


const IPAddress &
EndpointInfo::NTS_rasAddress() const
{
    return ( myRasAddress );
}


void	    	
EndpointInfo::callSignalAddress(const IPAddress & csAddr)
{
    Lock    lock(myMutex); (void)lock;
    NTS_callSignalAddress(csAddr);
}


void	    	
EndpointInfo::NTS_callSignalAddress(const IPAddress & csAddr)
{
    const string    fn("EndpointInfo::callSignalAddress");
    VLog    	    log(fn);

    myCallSignalAddress = csAddr;

    VDEBUG(log) << fn << ": " << myCallSignalAddress << VDEBUG_END(log);
}


IPAddress
EndpointInfo::callSignalAddress() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_callSignalAddress() );
}


const IPAddress &
EndpointInfo::NTS_callSignalAddress() const
{
    return ( myCallSignalAddress );
}


void	    	
EndpointInfo::timeToLive(u_int32_t ttl)
{
    Lock    lock(myMutex); (void)lock;
    NTS_timeToLive(ttl);
}


void	    	
EndpointInfo::NTS_timeToLive(u_int32_t ttl)
{
    const string    fn("EndpointInfo::timeToLive");
    VLog    	    log(fn);

    // We need to eventually convert this to milliseconds, which is signed,
    // so there is no use keeping a number greater than MAXINT/1000.
    //
    static u_int32_t	max_ttl = MAXINT / 1000;
    
    myTimeToLive = ( ttl > max_ttl ? max_ttl : ttl );

    VDEBUG(log) << fn << ": " << myTimeToLive << VDEBUG_END(log);
}


u_int32_t  	
EndpointInfo::timeToLive() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_timeToLive() );
}


u_int32_t  	
EndpointInfo::NTS_timeToLive() const
{
    return ( myTimeToLive );
}


void	    	
EndpointInfo::timeToLiveTimerEventId(FifoBase< Sptr<Event> >::EventId eventId)
{
    Lock    lock(myMutex); (void)lock;
    NTS_timeToLiveTimerEventId(eventId);
}


void	    	
EndpointInfo::NTS_timeToLiveTimerEventId(FifoBase< Sptr<Event> >::EventId eventId)
{
    const string    fn("EndpointInfo::timeToLiveTimerEventId");
    VLog    	    log(fn);

    myTimeToLiveTimerEventId = eventId;

    VDEBUG(log) << fn << ": " << myTimeToLiveTimerEventId << VDEBUG_END(log);
}


FifoBase< Sptr<Event> >::EventId 
EndpointInfo::timeToLiveTimerEventId() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_timeToLiveTimerEventId() );
}


FifoBase< Sptr<Event> >::EventId 
EndpointInfo::NTS_timeToLiveTimerEventId() const
{
    return ( myTimeToLiveTimerEventId );
}


void	    	
EndpointInfo::updateLastRegistration()
{
    Lock    lock(myMutex); (void)lock;
    NTS_updateLastRegistration();
}


void	    	
EndpointInfo::NTS_updateLastRegistration()
{
    gettimeofday(&myLastRegistration, 0);
    NTS_updateLastActivity();
}


timeval
EndpointInfo::lastRegistration() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_lastRegistration() );
}


const timeval &
EndpointInfo::NTS_lastRegistration() const
{
    return ( myLastRegistration );
}


void	    	
EndpointInfo::updateLastActivity()
{
    Lock    lock(myMutex); (void)lock;
    NTS_updateLastActivity();
}


void	    	
EndpointInfo::NTS_updateLastActivity()
{
    gettimeofday(&myLastActivity, 0);
}


timeval
EndpointInfo::lastActivity() const
{
    Lock    lock(myMutex); (void)lock;
    return ( NTS_lastActivity() );
}


const timeval &
EndpointInfo::NTS_lastActivity() const
{
    return ( myLastActivity );
}


u_int32_t   
EndpointInfo::admissionCount() const
{   
    Lock    lock(myMutex); (void)lock;
    return ( myAdmissions );
}


void
EndpointInfo::admit()
{
    Lock    lock(myMutex); (void)lock;
    ++myAdmissions;
}


void 
EndpointInfo::disengage()
{
    Lock    lock(myMutex); (void)lock;

    if ( myAdmissions > 0 )
    {
        --myAdmissions;
    }
}


void            
EndpointInfo::outOfResources(bool oor)
{
    Lock    lock(myMutex); (void)lock;
    myOutOfResources = oor;
}


bool            
EndpointInfo::outOfResources() const
{
    Lock    lock(myMutex); (void)lock;
    return ( myOutOfResources );
}
        

ostream &       
EndpointInfo::writeTo(ostream & out) const
{
    Lock    lock(myMutex); (void)lock;

    out << "endpointIdentifier=\"" << myEndpointIdentifier.c_str()
	<< "\", alias=\"" << myAlias
    	<< "\", rasAddress=\"" 
	<< myRasAddress.getIPAsString() << ':' << myRasAddress.getPort()
	<< "\", callSignalAddress=\""
	<< myCallSignalAddress.getIPAsString() << ':' << myCallSignalAddress.getPort()
        << "\", endpointType=\"" << TypeName(myEndpointType) << '"';

    size_t  size = myPrefix.GetSize();
    
    if ( size > 0 )
    {
        for ( size_t i = 0; i < size; i++ )
        {
            assert(myPrefix[i].GetTag() == H225_AliasAddress::e_e164);
            
            PASN_IA5String  &   e164 = myPrefix[i];

            if ( i == 0 )
            {
                out << ", prefix=\"" << e164.GetValue();
            }
            else
            {
                out << ", " << e164.GetValue();
            }
        }
        out << "\"";
    }
    out << ", admissions: " << myAdmissions
        << ", out of resources: " << ( myOutOfResources ? "true" : "false" );
    
    return ( out );
}
