/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const SipToH323AdmissionPolicy_cxx_Version = 
  "$Id: SipToH323AdmissionPolicy.cxx,v 1.8 2001/11/07 08:57:16 icahoon Exp $";


#include "SipToH323AdmissionPolicy.hxx"
#include "H323AdmissionEvent.hxx"
#include "Service.hxx"
#include "GatekeeperInfo.hxx"
#include "EndpointInfoMap.hxx"
#include "VLog.hxx"


using namespace Vocal::SipH323CSGW;
using Vocal::SipH323Info::EndpointInfo;
using Vocal::SipH323Info::EndpointInfoSet;
using Vocal::Transport::IPAddress;
using Vocal::Logging::VLog;


SipToH323AdmissionPolicy::SipToH323AdmissionPolicy(
    Service 	    	    &	service,
    RASGKProtocol   	    &	rasGkProtocol,
    GatekeeperInfo  	    &	gatekeeperInfo,
    EndpointInfoMap 	    &	endpointInfoMap
)
    :	DefaultAdmissionPolicy(service, rasGkProtocol, 
    	    	    	       gatekeeperInfo, endpointInfoMap),
    	EventObserver< Sptr<Event> >(service.getEventSubject()),
	myH323AdmissionEvent(0)
{
    myService.subscribe(*this);
}


SipToH323AdmissionPolicy::~SipToH323AdmissionPolicy()
{
    myService.unsubscribe(*this);
}


bool	    
SipToH323AdmissionPolicy::interestedIn(const Sptr<Event> & event)
{
    return ( myH323AdmissionEvent.dynamicCast(event) != 0 );
}


void	    
SipToH323AdmissionPolicy::onEvent(Sptr<Event>)
{
    const string    fn("SipToH323AdmissionPolicy::onEvent");
    VLog    	    log(fn);

    if ( myH323AdmissionEvent == 0 )
    {
    	assert(0);
    	return;
    }

    ReturnCode	rc = !SUCCESS;
    
    Sptr<EndpointInfo>	destEptInfo = 0;
    Sptr<EndpointInfo>  srcEptInfo = 0;

    // Lookup the destination endpoints' info by callSignalAddress first.
    //
    Sptr<IPAddress> destCSAddr = myH323AdmissionEvent->destCallSignalAddress();

    if ( destCSAddr != 0 )
    {
    	destEptInfo = myEndpointInfoMap.findByCallSignalAddress(*destCSAddr);

    	VDEBUG(log) << fn << ": searching by destination call signal address = "
	    	    << *destCSAddr << VDEBUG_END(log);
    }

    // Then try to lookup the destination endpoints' info by the alias address.
    //
    Sptr<H225_ArrayOf_AliasAddress> destAlias = myH323AdmissionEvent->destAlias();

    if ( destEptInfo == 0 && destAlias != 0 )
    {
    	destEptInfo = myEndpointInfoMap.findByAlias(*destAlias);    

    	VDEBUG(log) << fn << ": searching by alias address = "
	    	    << *destAlias << VDEBUG_END(log);    
    }

    // Finally try to lookup the destination endpoint by matching the alias
    // to a gateway. 
    //    
    if ( destEptInfo == 0 && destAlias != 0 )
    {
        destEptInfo = findGateway(*destAlias);
    }
    
    // Try to lookup the source endpoint by call signal address.
    //
    Sptr<IPAddress> srcCSAddr = myH323AdmissionEvent->srcCallSignalAddress();
				
    if ( srcCSAddr != 0 )
    {
    	srcEptInfo = myEndpointInfoMap.findByCallSignalAddress(*srcCSAddr);

    	VDEBUG(log) << fn << ": searching by source call signal address = "
	    	    << *srcCSAddr << VDEBUG_END(log);    
    }
    
    if ( destEptInfo != 0 )
    {
    	VDEBUG(log) << fn << ": destination endpoint found = " << *destEptInfo
	    	    << VDEBUG_END(log);
                    
        rc = SUCCESS;
    }
    
    if ( srcEptInfo != 0 )
    {
    	VDEBUG(log) << fn << ": source endpoint found = " << *srcEptInfo
	    	    << VDEBUG_END(log);

        rc = SUCCESS;
    }

    // If the dest endpoint is a terminal, and it is admitted in as many calls 
    // as it has registered aliases (which are the number of e.164 addresses 
    // present), we are going to reject the call^W^W^Wcomplain vigorously. 
    // This is a hack to work around a bug in the router/translator/cap map.
    //
    size_t          number_e164 = 0;
    
    if ( destEptInfo != 0 )
    {
        const size_t    number_aliases = destEptInfo->alias().GetSize();

        for ( size_t i = 0; i < number_aliases; ++i )
        {
            if ( destEptInfo->alias()[i].GetTag() == H225_AliasAddress::e_e164 )
            {
                ++number_e164;
            }
        }
    }

    if  (   destEptInfo != 0
        &&  destEptInfo->endpointType() == EndpointInfo::TERMINAL_TYPE
        &&  destEptInfo->admissionCount() >= number_e164
        )
    {
        /*
    	VERR(log)   << fn << ": Number of admissions ("
                    << destEptInfo->admissionCount()
                    << ") exceeds to number of registered phone numbers ("
                    << number_e164 
                    << "). Resource leakage likely."
	    	    << VERR_END(log);
        */
    }
    
    myH323AdmissionEvent->returnCode(rc);

    myH323AdmissionEvent->destEndpointInfo(destEptInfo);
    myH323AdmissionEvent->srcEndpointInfo(srcEptInfo);

    VDEBUG(log) << fn << ": returning admission event = "
	    	<< *myH323AdmissionEvent << VDEBUG_END(log);

    myH323AdmissionEvent->replyFifo().add(myH323AdmissionEvent);
    
    myH323AdmissionEvent = 0;
}


ostream &           
SipToH323AdmissionPolicy::writeTo(ostream & out) const
{
    return ( out << "SipToH323AdmissionPolicy" );
}


Sptr<EndpointInfo>  
SipToH323AdmissionPolicy::chooseGateway(
    const EndpointInfoSet & gateways
)
{
    return  (   gateways.size() > 0 
                ?   *(gateways.begin())
                :   Sptr<EndpointInfo>(0)
            );
}
