#ifndef INCLUDED_EventRefCnt_hxx
#define INCLUDED_EventRefCnt_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const EventRefCnt_hxx_Version = 
"$Id: EventRefCnt.hxx,v 1.9 2000/12/18 23:48:39 bko Exp $";

#ifndef DISABLE_MTHREAD
#include "VMutex.h"
#include "Lock.hxx"
#endif

namespace mte
{

    /// Memory allocator for reference counting events
    template<class Type>
    class EventRefCnt: public Type
    {
    public:
        EventRefCnt();
        
        /// Creates working copy of the event (implemented by top-level class)
        EventRefCnt<Type>* clone();

        /// Destroys working copy of the event
        virtual void destroy();

    protected:
        /// Direct call to copy constructor is not allowed: use clone()
        EventRefCnt(const EventRefCnt& other);

    private:
        /// Assignment operator is not allowed
        EventRefCnt operator = (const EventRefCnt&);

        /// Creates working copy of the event (implementation call-back)
        virtual EventRefCnt<Type>* clone_impl() = 0;

        int refCounter_;
#ifndef DISABLE_MTHREAD
        VMutex mutex_;
#endif
    };

    //----------------------------
    // Inline and template methods
    //----------------------------

    template<class Type>
    EventRefCnt<Type>::
    EventRefCnt()
        : Type()
    {
    }
    
    template<class Type>
    void EventRefCnt<Type>::
    destroy()
    {
        bool counterIsZero;
        
        {
#ifndef DISABLE_MTHREAD
            Vocal::Process::Lock lock(mutex_);
#endif
            -- refCounter_;
            counterIsZero = (refCounter_ <= 0);
        }

        if (counterIsZero)
        {
            delete this;
        }
    }

    template<class Type>
    EventRefCnt<Type>::
    EventRefCnt(const EventRefCnt& other)
        : Type(other)
    {
    }

    template<class Type>
    EventRefCnt<Type>* EventRefCnt<Type>::
    clone()
    {
        if (refCounter_ > 0)
        {
            {
#ifndef DISABLE_MTHREAD
                Vocal::Process::Lock lock(mutex_);
#endif
                ++ refCounter_;
            }

            return this;
        }
        else
        {
            EventRefCnt<Type>* copy = clone_impl();
            ++ copy->refCounter_;
            return copy;
        }
    }
    
}

#endif
