#ifndef INCLUDED_Feature_hxx
#define INCLUDED_Feature_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const Feature_hxx_Version = 
"$Id: Feature.hxx,v 1.10 2001/10/19 21:37:37 icahoon Exp $";


#include <cassert>
#include <cstdio>
#include <string>


#include "FeatureBase.hxx"
#include "VLog.hxx"


using Vocal::Logging::VLog;
using std::string;


namespace mte
{

    /// State machine feature
    template<class Pattern, class Derived>
    class Feature: public FeatureBase
    {
    public:
        typedef void (Pattern::* PState)(Derived);

        /// Construct state machine feature
        Feature();
        
        /// Process event
        virtual void process(Pattern& pattern, const char * label = 0);

        /// Change state
        void setState(PState state, const char * label = 0);

        /// Change state, no entry/exit handlers are called
        void setStateStealth(PState state, const char * label = 0) throw();

        /// Set entry handler
        void setEntry(Pattern* handler, const char * label = 0) throw();

        /// Set exit handler
        void setExit(Pattern* handler, const char * label = 0) throw();

        /// Set prefix handler
        void setPrefix(Pattern* handler, const char * label = 0) throw();

        /// Set postfix handler. Do not use if object may be destoyed in process_impl!
        void setPostfix(Pattern* handler, const char * label = 0) throw();

        PState getState() const throw();
        
    private:
        /// Process event (implementation call-back)
        virtual void process_impl(Pattern& pattern) = 0;
        //to be implemented as (pattern.*(getState()))(*this);

        PState state_;
        Pattern* entry_;
        Pattern* exit_;
        Pattern* prefix_;
        Pattern* postfix_;
        string  entryLabel_;
        string  exitLabel_;
        string  prefixLabel_;
        string  postfixLabel_;
    };
    
    //----------------------------
    // Inline and template methods
    //----------------------------

    template<class Pattern, class Derived>
    Feature<Pattern, Derived>::
    Feature()
        : entry_(0),
          exit_(0),
          prefix_(0),
          postfix_(0)
    {
    }
    
    template<class Pattern, class Derived>
    void Feature<Pattern, Derived>::
    process(Pattern& pattern, const char * label)
    {
        const string    fn("Feature::process");
        VLog            log(fn);
        
        assert(this);
        Pattern* postfix = postfix_;

        if (prefix_)
        {
            VDEBUG(log) << fn << ": process prefix pattern: " << prefixLabel_ 
                        << VDEBUG_END(log);
            
            process_impl(*prefix_);
        }

        if ( label )
        {
            VDEBUG(log) << fn << ": process pattern: " << label
                        << VDEBUG_END(log);
        }
        else
        {
            VDEBUG(log) << fn << ": process pattern: " << &pattern 
                        << VDEBUG_END(log);
        }
        
        process_impl(pattern);

        if (postfix)
        {
            VDEBUG(log) << fn << ": process postfix pattern: " << postfixLabel_
                        << VDEBUG_END(log);
                        
            process_impl(*postfix);
        }
    }

    template<class Pattern, class Derived>
    inline void Feature<Pattern, Derived>::
    setEntry(Pattern* pattern, const char * label) throw()
    {
        assert(this);

        entry_ = pattern;
        
        if ( label )
        {
            entryLabel_ = label;
        }
        else
        {
            char    buf[32];
            snprintf(buf, 32, "%p", pattern);
            
            entryLabel_ = buf;
        }
    }

    template<class Pattern, class Derived>
    inline void Feature<Pattern, Derived>::
    setExit(Pattern* pattern, const char * label) throw()
    {
        assert(this);

        exit_ = pattern;

        if ( label )
        {
            exitLabel_ = label;
        }
        else
        {
            char    buf[32];
            snprintf(buf, 32, "%p", pattern);
            
            exitLabel_ = buf;
        }
    }

    template<class Pattern, class Derived>
    inline void Feature<Pattern, Derived>::
    setPrefix(Pattern* pattern, const char * label) throw()
    {
        assert(this);

        prefix_ = pattern;

        if ( label )
        {
            prefixLabel_ = label;
        }
        else
        {
            char    buf[32];
            snprintf(buf, 32, "%p", pattern);
            
            prefixLabel_ = buf;
        }
    }

    template<class Pattern, class Derived>
    inline void Feature<Pattern, Derived>::
    setPostfix(Pattern* pattern, const char * label) throw()
    {
        assert(this);

        postfix_ = pattern;

        if ( label )
        {
            postfixLabel_ = label;
        }
        else
        {
            char    buf[32];
            snprintf(buf, 32, "%p", pattern);
            
            postfixLabel_ = buf;
        }
    }

    template<class Pattern, class Derived>
    inline Feature<Pattern, Derived>::PState Feature<Pattern, Derived>::
    getState() const throw()
    {
        return state_;
    }

    template<class Pattern, class Derived>
    void Feature<Pattern, Derived>::
    setState(PState state, const char * label)
    {
        const string    fn("Feature::setState");
        VLog            log(fn);

        assert(this);
        assert(state);
        
        if (exit_)
        {
            VDEBUG(log) << fn << ": exit action: " << exitLabel_ 
                        << VDEBUG_END(log);

            process_impl(*exit_);
        }

        state_ = state;

        VDEBUG(log) << fn << ": changing state to: " 
                    << ( label ? label : "-" ) 
                    << VDEBUG_END(log);

        if (entry_)
        {
            VDEBUG(log) << fn << ": entry action: " << entryLabel_ 
                        << VDEBUG_END(log);

            process_impl(*entry_);
        }
    }

    template<class Pattern, class Derived>
    inline void Feature<Pattern, Derived>::
    setStateStealth(PState state, const char * label) throw()
    {
        const string    fn("Feature::setStateStealth");
        VLog            log(fn);

        assert(this);
        assert(state);

        state_ = state;

        VDEBUG(log) << fn << ": changing state to: " 
                    << ( label ? label : "-" ) 
                    << VDEBUG_END(log);
    }

}

#endif
