
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const CAServerContainer_cxx_Version = "$Id: CAServerContainer.cxx,v 1.24 2002/02/25 05:57:18 sprajpat Exp $";


#include "CAServerContainer.hxx"
#include "Data.hxx"
#include "Common.h"
#include "cpLog.h"
#include "callAgent.hxx"
#include "CAServer.hxx"
#include "StartupSequence.hxx"
#include "ReadLockHelper.hxx"
#include "WriteLockHelper.hxx"
#include "MSServer.hxx"
#include "ProvisionInterface.hxx"
#include "VIoException.hxx"

static const string GlobConfigfilename("GlobalConfigData");

Sptr < CAServerContainer > CAServerContainer::theCAContainer = 0;

/*
 * Singleton of the CAServerContainer
 * @return the CAServerContainer object.
 */
CAServerContainer&
CAServerContainer::instance()
{
    if ( theCAContainer == 0 )
    {
        theCAContainer = new CAServerContainer;
    }
    return *theCAContainer;
}

/*
 * Default Constructor
 */
CAServerContainer::CAServerContainer()
        : BaseServerContainer(), startupFlag(false)
{
    // register interest in Global Config file
    // ProvisionInterface& pif = ProvisionInterface::instance();
    try
    {
     ProvisionInterface::instance().registerForUpdate
      (updateCAHeartbeatParms, GlobConfigfilename, ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s",
              e.getDescription().c_str());
    }

    Data host(ca.sipMyHost), port(ca.sipMyPort);
    Data caId(host + ":" + port);
    addServer(caId, SERVER_CA);

    if (ca.otherCaHost != 0)
    {
        host = ca.otherCaHost;
        port = ca.otherCaPort;
        Data otherCa(host + ":" + port);
        addServer(otherCa, SERVER_CA);
    }

    itsMsList = new vector < Data > ;

    displayLists();
}

/*
 * Copy constructor
 * $param rhs the object to be copied
 */
CAServerContainer::CAServerContainer( const CAServerContainer& rhs )
{
    if (this != &rhs)
        copyRhsToThis( rhs );
}

/*
 * Assignment Operator
 * @param rhs the object contains data to be copied
 * @return CAServerContainer& the copied object
 */
CAServerContainer&
CAServerContainer::operator=( const CAServerContainer& rhs )
{
    if ( this != &rhs )
    {
        BaseServerContainer::operator= (rhs);
        copyRhsToThis( rhs );
    }
    return *this;
}

/*
 * Copy the data from one object to self
 * @param rhs the object contains data to be copied
 */
void
CAServerContainer::copyRhsToThis( const CAServerContainer& rhs )
{}

/*
 * Default destructor
 */
CAServerContainer::~CAServerContainer( )
{}

/*
 * Process the heartbeat it receives from the Multi cast group
 * @param key the server id
 * @return bool true if the id is in the list, false otherwise
 */
bool
CAServerContainer::processHeartbeat( const Data& key)
{
    Sptr < Server > serverPtr;

    if (search( key, serverPtr) )
    {
        ReadLockHelper tableHelper(tableLock);

        if ( serverPtr == 0 )
        {
            cpLog( LOG_DEBUG, "The serverPtr pointer is NULL !!");
            assert( serverPtr != 0 );
            return ( false );
        }

        switch (serverPtr->getType() )
        {
            case SERVER_CA:
            {
                Sptr < CAServer > caServer;
                caServer.dynamicCast(serverPtr);
                if ( caServer == 0 )
                {
                    cpLog( LOG_DEBUG, "The caServer pointer is NULL !!");
                    assert (caServer != 0);
                    return ( false );
                }
                caServer->processHeartbeat(key);
            }
            break;
            case SERVER_MS:
            {
                Sptr < MSServer > msServer;
                msServer.dynamicCast(serverPtr);
                if (msServer == 0)
                {
                    cpLog (LOG_DEBUG, "The msServer pointer is NULL !!");
                    assert (msServer != 0);
                    return false;
                }
                msServer->processHeartbeat(key);
            }
            break;
            default:
            break;
        };

        return true;
    }

    return false;

}

/*
 * Detect servers in the container up or down
 */
void
CAServerContainer::doHouseKeeping()
{
    Data serverId;
    Sptr < Server > serverPtr;

    cpLog( LOG_DEBUG_HB, "");
    cpLog( LOG_DEBUG_HB, "****** CA Server Container: do HouseKeeping *****");

    ReadLockHelper tableHelper(tableLock);

    if ( !serverList.empty() )
    {
        for ( TableIter i = serverList.begin(); i != serverList.end(); ++i)
        {
            serverId = i->first;
            serverPtr = i->second;

            if ( serverPtr == 0 )
            {
                cpLog( LOG_DEBUG, "The serverPtr pointer is NULL !!");
                assert( serverPtr != 0 );
                continue;
            }
            if ( serverPtr->getStatus() != Inactive )
                cpLog( LOG_DEBUG_HB, "%s", serverId.logData() );

            switch (serverPtr->getType() )
            {
                case SERVER_CA:
                {
                    Sptr < CAServer > caServer;
                    caServer.dynamicCast(serverPtr);
                    if ( caServer == 0 )
                    {
                        cpLog( LOG_DEBUG, "The caServer pointer is NULL !!");
                        assert (caServer != 0);
                        continue;
                    }
                    caServer->doHouseKeeping(serverId);
                }
                break;
                case SERVER_MS:
                {
                    Sptr < MSServer > msServer;
                    msServer.dynamicCast(serverPtr);
                    assert (msServer != 0);
                    msServer->doHouseKeeping(serverId);
                }
                break;
                default:
                break;
            }
        }
    }
    else
    {
        cpLog( LOG_DEBUG_HB, "nothing in server container");
    }

    cpLog( LOG_DEBUG_HB, "****** finished HouseKeeping *****");
    cpLog( LOG_DEBUG_HB, "");
}

/*
 * Add a new server object into the server container list
 * @param key server id
 * @param serverType either callAgent or marshal
 */
void
CAServerContainer::addServer(const Data& key, const ServerType serverType)
{
    WriteLockHelper tableHelper(tableLock);
    switch (serverType)
    {
        case SERVER_CA:
        {
            cpLog (LOG_DEBUG, "Added CA = %s", key.logData());
            Sptr < CAServer > caServer = new CAServer(key);
            assert (caServer != 0);
            serverList[key] = caServer;
        }
        break;
        case SERVER_MS:
        {
            cpLog (LOG_DEBUG, "Added MS = %s", key.logData());
            Sptr < MSServer > msServer = new MSServer(key);
            assert (msServer != 0);
            serverList[key] = msServer;
            // add to the ms list
            itsMsList->push_back(key);
        }
        break;
        default:
        break;
    }

    return ;
}

/*
 * Utility function to print out the list of all 
 * callAgent in the server container list
 */
void
CAServerContainer::displayLists()
{
    cpLog( LOG_DEBUG, "----------- Call Agent ----------");
    Data caId;

    ReadLockHelper tableHelper(tableLock);

    for (TableIter i = serverList.begin();
            i != serverList.end();
            ++i)
    {
        caId = i->first;
        cpLog( LOG_DEBUG, "CallAgent: %s", caId.logData() );
    }

}

/*
 * Get the list of all active callAgent server in the system
 * @param type callAgent type
 * @param listofActiveServers returned list of active callAgent
 */
void
CAServerContainer::getListofActive( const ServerType type,
                                    Sptr < vector < Data > > & listofActiveServers )
{
    ReadLockHelper tableHelper(tableLock);

    if ( listofActiveServers.getPtr() == 0 )
    {
        cpLog( LOG_DEBUG, "The listofActiveServers contains a NULL pointer !!");
        assert( listofActiveServers.getPtr() != 0 );
        return ;
    }
    for (TableIter i = serverList.begin();
            i != serverList.end();
            ++i)
    {
        Sptr < Server > aServer = i->second;

        if ( aServer == 0 )
        {
            cpLog( LOG_DEBUG, "The aServer contains a NULL pointer !!");
            assert( aServer != 0 );
            continue;
        }

        if ((aServer->getType() == SERVER_CA) &&
                (aServer->getStatus() == Active))
            listofActiveServers->push_back(i->first);
    }

    return ;

}

/*
 * Process the startup sequence
 * @param isUp indicated the server is up or down
 */
void
CAServerContainer::startStartupSequence(bool isUp)
{
    if (! startupFlag)
    {
        startupSequence(isUp);
    }

}

/*
 * Set the startup flag to true
 */
void
CAServerContainer::setStartupFlag()
{
    startupFlag = true;
}

/*
 * Update the heartbeat parameters from PS  
 */
void
CAServerContainer::updateHBParmsFromPS()
{
    cpLog( LOG_DEBUG, "updating Heartbeat Parms in ServerContainer");

    ProvisionInterface& pif = ProvisionInterface::instance();
    try
    {
        WriteLockHelper tableHelper(tableLock);

        /// get pointer to global config object
        VGlobalConfigData globalConfigData;
        pif.getConfigData(globalConfigData);

        int newValue = globalConfigData.getMaxMissedHB();
        int oldValue = serverList.begin()->second->getMaxMissedHB();
        cpLog( LOG_DEBUG, "maxMissedHB old: %d new: %d", oldValue, newValue );

        if (( !serverList.empty() ) && ( oldValue != newValue ))
        {
            Data serverId;
            Sptr < Server > serverPtr;

            for ( TableIter i = serverList.begin(); i != serverList.end(); ++i)
            {
                serverId = i->first;
                serverPtr = i->second;
                serverPtr->setMaxMissedHB( newValue );
            }
        }
    }
    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed to get data from Provisioning Server, reason %s", e.getDescription().c_str());
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
    }
}


/*
 * Global Procedure: The Provisioning Server is notifying us of a change in Heartbeat Parms.
 */
void
updateCAHeartbeatParms(const string& data, const string& fileName, bool)
{
    // update Heartbeat Parms
    CAServerContainer::instance().updateHBParmsFromPS();
}



// End of File
