
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const MessageHandler_cxx_Version = "$Id: MessageHandler.cxx,v 1.12 2001/02/07 02:35:11 hlo Exp $";

#include <cassert>
#include "MessageHandler.hxx"
#include "cpLog.h"
#include "callAgent.hxx"
#include "endpoint.hxx"


/*
 * Default constructor
 */
MessageHandler::MessageHandler() : itsCmd(0)
{}

/*
 * Default destructor
 */
MessageHandler::~MessageHandler()
{}

/*
 * Received and Process the message
 * @param cmd the received message
 */
void
MessageHandler::processMsg(Sptr < MgcCommand > cmd)
{
    if ( cmd == 0 )
    {
        cpLog( LOG_DEBUG, "The cmd contains a NULL pointer !!");
        assert (cmd != 0);
        return ;
    }
    cpLog (LOG_DEBUG, "processMsg");

    itsCmd = cmd;

    string epId = itsCmd->getEndpointId();

    // returned ep is used in QueueInEventData only
    switch (getType(epId))
    {
        case SINGLE:  // single ep
        queueInEventData();
        break;
        case WILDCARD:  // wildcard
        processWildcard();
        break;
        case NOMATCH:  // no ep
        default:
        cpLog (LOG_WARNING, "getAsynMessage==> Unknown endpoint id %s", epId.c_str());
        break;
    }

}

/*
 * Return the type of the message to detect the
 * wildcard if it is in the message
 * @param epId the message string
 * @return RSIPType the type of message
 */
RSIPType
MessageHandler::getType(const string epId)
{
    if (epId.find("*@") == 0)
        return WILDCARD;

    epIter = epMap.find (epId);
    if (epIter == epMap.end())
        return NOMATCH;

    return SINGLE;

}

/*
 * Insert the message into the queue Event
 */
void
MessageHandler::queueInEventData()
{
    assert (itsCmd != 0);

    cpLog (LOG_DEBUG, "queueInEventData");

    QueueEventData* eventData = newQueueEventData ();
    if ( eventData == 0 )
    {
        cpLog( LOG_DEBUG, "Ran out of QueueEventData in the pool !!");
        assert( eventData != 0 );
        return ;
    }
    eventData->type = StackEvent;
    eventData->cmd = itsCmd;
    eventData->ep = epIter->second;

    if ( epIter->second == 0 )
    {
        cpLog( LOG_DEBUG, "The epIter->second is a NULL pointer !!");
        assert( epIter->second != 0 );
        return ;
    }
    if ( epIter->second->iq == 0 )
    {
        cpLog( LOG_DEBUG, "The epIter->second->iq is a NULL pointer !!");
        assert( epIter->second->iq != 0 );
        return ;
    }
    (void) epIter->second->iq->add (eventData);

}

/*
 * Insert into the queue event all ep which
 * match the host ip address
 */
void
MessageHandler::processWildcard()
{
    if ( itsCmd == 0 )
    {
        cpLog( LOG_DEBUG, "The itsCmd is a NULL pointer !!");
        assert (itsCmd != 0);
        return ;
    }

    string epId = itsCmd->getEndpointId();
    cpLog (LOG_DEBUG, "processWildcard = %s", epId.c_str());

    for (epIter = epMap.begin(); epIter != epMap.end(); ++epIter)
    {
        if (epId.find("*@*") == 0)
            queueInEventData();
        else if (addressMatched(epId, epIter->first))
            queueInEventData();
    }

    return ;
}

/*
 * Find out if the addresses matched from 2 id
 * @param id1 the first id to compared
 * @param id2 the second id to compared
 * @return int true if they are equal, false otherwise
 */
int
MessageHandler::addressMatched (const string id1, const string id2)
{
    assert (! id1.empty());
    assert (! id2.empty());

    string::size_type addrPos1 = id1.find("@");
    string::size_type addrPos2 = id2.find("@");

    string addr1, addr2;
    addr1.assign(id1.c_str(), addrPos1);
    addr2.assign(id2.c_str(), addrPos2);

    return (addr1 == addr2);

}
