/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LSApiServer_cxx_Version = 
"$Id: LSApiServer.cxx,v 1.9 2001/06/29 21:02:44 bko Exp $";


#include <unistd.h>
#include <iostream>
#include <stdlib.h> 
#include <strstream.h>
#include "LSSessionManager.hxx"
#include "LSLocator.hxx"
#include "LSApiServer.hxx"
#include "LSSendOnlySession.hxx"
#include "LSReceiveOnlySession.hxx"
#include "LSSendReceiveSession.hxx"
#include "Tcp_ServerSocket.hxx"
#include "TRIPOpenMsg.hxx"
#include "TRIPTransceiver.hxx"
#include "TRIPErrorHandler.hxx"
#include "TRIPNotificationMsg.hxx"
#include "TRIBAgent.hxx"
#include <algorithm>
#include "cpLog.h"

LSApiServer* LSApiServer::_instance = 0;

extern "C" {
    void LSApiServer_destroy()
    {
        LSApiServer::destroy();
    }
};

LSApiServer&
LSApiServer::instance(int port)
{
    if (!_instance)
    {
        _instance = new LSApiServer(port);
        cpLog(LOG_INFO, "Started LSApiServer on port (%d)", port);
    }
    return (*_instance);
}

LSApiServer&
LSApiServer::instance()
{
    assert(_instance != 0);
    return (*_instance);
}

void 
LSApiServer::destroy(void)
{
    delete LSApiServer::_instance;
    LSApiServer::_instance = 0;
}

LSApiServer::LSApiServer(int port)
{
    _port = port;
    _shutdown = false;
    itsProxyConnectedFlag = false;
    if (atexit(LSApiServer_destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register with atexit()");
    };

    _serverSocket = new TcpServerSocket(port);
    _thread = new VThread();
    _thread->spawn(LSApiServer::run, this);
}

LSApiServer::~LSApiServer()
{
    _shutdown = true;
    _serverSocket->close();
    if(_thread.getPtr() != 0)
    {
        _thread->join();
        _thread->exit();
    }
}

void*
LSApiServer::run(void* args)
{
    LSApiServer* self = static_cast<LSApiServer*>(args);
    self->processMsg();
    return (0);
}

void
LSApiServer::processMsg()
{
    //Currently the logic assumes that A Proxy Server will be pulling
    //information from the Location Server.However if it changes
    //into a PUSH model, where Location server will be pushing data
    //into Proxy, the logic need to be redesigned to fit the requirement
    for (; ; )
    {
        if(_serverSocket->getServerConn().isReadReady(1))
        {
            Connection conn;
            _serverSocket->accept(conn);
            cpLog(LOG_DEBUG, "Accepted connection (%s)", 
                              conn.getDescription().c_str());
            _connectionList.push_back(conn);
            setProxyConnected(true);
            continue;
        }
        if(_shutdown ) break;

        if (_connectionList.size() > 0)
        {
            for(ConnectionList::iterator itr = _connectionList.begin();
                                itr != _connectionList.end(); itr++)
            {
                try
                {
                    if(!(*itr).isLive())
                    {
                        removeConnection((*itr));
                        break;
                    }
                    if((*itr).isLive() && (*itr).isReadReady(1))
                    {
                        processRequest((*itr));
                        if (_connectionList.size() == 0)
                        {
                            setProxyConnected(false);
                            break;
                        }
                    }
                }
                catch(VNetworkException& e)
                {
                    cpLog(LOG_ALERT, "Network error, reason:%s",
                             e.getDescription().c_str());
                    removeConnection((*itr));
                    if (_connectionList.size() == 0)
                    {
                        setProxyConnected(false);
                    }
                    break;
                }
                catch(VException& e)
                {
                    cpLog(LOG_ALERT, "Caught exception, reason:%s",
                             e.getDescription().c_str());
                }
            }
        }
    }
}

void
LSApiServer::processRequest(Connection& conn) throw (VNetworkException&)
{
    cpLog(LOG_DEBUG, "Processing request on conn (%s)",
                                 conn.getDescription().c_str());
    TRIPTransceiver trans;

    Sptr<TRIPMsg> msg = trans.receiveMsg(conn.getConnId());
    assert(msg != 0);
    if (msg->getType() == TRIP_OP_OPEN)
    {
        processOpenMsg(msg, conn);
    }
    else if (msg->getType() == TRIP_OP_NOTIF)
    {
        processNotificationMsg(msg, conn);
    }
}

void
LSApiServer::processNotificationMsg(Sptr<TRIPMsg> msg, Connection& conn)
{
    cpLog(LOG_DEBUG, "LSApiServer::processNotificationMsg");    
    if(msg == 0)
    {
        assert(0);
         return;
    }
    Sptr<TRIPNotificationMsg> notify;
    notify.dynamicCast(msg);
    assert(notify != 0);
    removeConnection(conn);
}

void
LSApiServer::processOpenMsg(Sptr<TRIPMsg> msg, Connection& conn)
{
    cpLog(LOG_DEBUG, "LSApiServer::processOpenMsg");
    if(msg == 0)
    {
        assert(0);
        return;
    }

    Sptr<TRIPOpenMsg> openMsg;
    openMsg.dynamicCast(msg);
    assert(openMsg != 0);
    u_int16_t itad = openMsg->getItad();
    if (itad != LSSessionManager::instance().getItad())
    {
        cpLog(LOG_DEBUG, "LSApiServer::request itad does not match");
        return;
    }

    //Match up required Route Capabilities with those available
    //If all OK store the required protocal and Address family for each
    //connection to be used when updates for a particular protocol occur
    Sptr<TRIPRouteCapability> routeCap = openMsg->getRouteCapability();
    assert(routeCap != 0);
    LSSessionManager& mgr = LSSessionManager::instance();
    const TRIPRouteCapabilityList& routeCapList =
                                            routeCap->getRouteCapabilityList();   
    ProtoKeyList protoList;
    for(TRIPRouteCapabilityList::const_iterator itr = routeCapList.begin();
                                            itr != routeCapList.end(); itr++)
    {
        TRIPAddrFamily aFamily = (*itr)->getAddrFamily();
        TRIPAppProto proto = (*itr)->getAppProtocol();
        if (! ((mgr.isAddrFamilySupported(aFamily)) &&
            (mgr.isProtocolSupported(proto))))
        {
            cpLog(LOG_ALERT, "Unsupported Route Capability");
            TRIPData data("Unsupported Route Capability");
            TRIPNotificationMsg msg(TRIP_ERR_OPEN, O_UNSUPP_CAPABILITY, data);
            msg.writeData(conn.getConnId());
            removeConnection(conn);
            return;
        }
        TRIPProtoKey protoKey(proto, aFamily);
        if (itsConnListMap.count(protoKey))
        {
            Sptr<ConnectionList> connList = itsConnListMap[protoKey];
            ConnectionList::const_iterator where = find(connList->begin(),
                                            connList->end(), conn);
            if (where == connList->end())
            {
                connList->push_back(conn);
            }
        }
        else
        {
            Sptr<ConnectionList> connList = new ConnectionList();
            connList->push_back(conn);
            itsConnListMap.insert(pair< TRIPProtoKey, 
                                Sptr <ConnectionList> >(protoKey, connList));
        }
        protoList.push_back(protoKey);
    }

    //Received OPEN msg from proxy - need to update proxy with Route data
    initialiseProxyRouteSet(conn, protoList);

}

void
LSApiServer::initialiseProxyRouteSet(Connection& conn, ProtoKeyList& pList)
{
    TRIBAgent& tribAgent = TRIBAgent::instance();
    Sptr <TRIPTrieCache> locTrib = tribAgent.getLocTrib();
    locTrib->readLock();

    for (ProtoKeyList::const_iterator itr = pList.begin();
                                            itr != pList.end(); itr++)
    {
        Sptr<TRIPRouteSet> proxySet = locTrib->getSortedRouteSet((*itr));

        if (proxySet.getPtr())
        {
            Sptr <UpdateMsgList> msgList = getProxyMsgList(*proxySet, true);  
            if (msgList.getPtr())
            {
                sendProxyUpdates(msgList, conn);
            }
        }
    }

    locTrib->unlock();
}

void
LSApiServer::sendProxyUpdates(Sptr <UpdateMsgList> msgList, Connection& conn)
{
    if (!(msgList.getPtr()))
    {
        return;
    }
    for (UpdateMsgList::const_iterator itr = msgList->begin();
                                        itr != msgList->end(); itr++)
    {
        sendUpdate(*(*itr), conn);
    }
}

void
LSApiServer::sendAllProxiesUpdates(Sptr <UpdateMsgList> msgList, 
                                                const TRIPProtoKey& protoKey)
{
    if (!(msgList.getPtr()))
    {
        return;
    }
    for (UpdateMsgList::const_iterator itr = msgList->begin();
                                        itr != msgList->end(); itr++)
    {
        Connection conn;
        if (itsConnListMap.count(protoKey))
        {
            Sptr <ConnectionList> connList = itsConnListMap[protoKey];
            if (connList.getPtr())
            {
                for (ConnectionList::const_iterator connItr = connList->begin();
                                         connItr != connList->end(); connItr++)
                {
                    conn = (*connItr);
                    sendUpdate(*(*itr), conn);
                }
            }
        }
    }
}


void
LSApiServer::sendUpdate(const TRIPUpdateMsg& uMsg, Connection& conn) 
                                            throw (VNetworkException&)
{
    if(conn.isLive())
    {
        uMsg.writeData(conn.getConnId());
    }
}


void
LSApiServer::updateProxies(Sptr <TRIPWorkSet> workSet)
{
    cpLog(LOG_DEBUG, "LSApiServer::updateProxies");
    if (!(workSet.getPtr()))
    {
        assert(0);
    }
    const RouteSetMap& workSetMap = workSet->getRouteSet();
    if (workSetMap.size() == 0)
    {
        return;
    }
    try
    {
        TRIPRouteSet routeSet;
        for (RouteSetMap::const_iterator wItr = workSetMap.begin();
                                wItr != workSetMap.end(); wItr++)
        {
            TRIPProtoKey protoKey = (*wItr).first;
            routeSet = (*wItr).second;
            Sptr <UpdateMsgList> msgList = getProxyMsgList(routeSet, false);
            if (msgList.getPtr())
            {
                sendAllProxiesUpdates(msgList, protoKey);
            }
        }
    }
    catch(VMissingDataException& e)
    {
        cpLog(LOG_DEBUG, "(%s)", e.getDescription().c_str());
    }
}


Sptr <UpdateMsgList>
LSApiServer::getProxyMsgList(const TRIPRouteSet& routeSet, bool initial)
{
    if (routeSet.size() == 0)
    {
        return 0;
    }
    u_int32_t prevTranId = 0;
    bool msgToSend = false;
    bool withdrawn = false;
    Sptr <UpdateMsgList> retList = new UpdateMsgList();
    Sptr<TRIPUpdateMsg> uMsg;
    Sptr<TRIPWithdrawnRoute> wRoute;
    Sptr<TRIPReachableRoute> rRoute;
    Sptr<TRIPNextHopServer> nextHop;
    Sptr<TRIPAdvertisePath> aPath = new TRIPAdvertisePath();
    Sptr<TRIPRoutedPath> rPath = new TRIPRoutedPath();

    for(TRIPRouteSet::iterator itr = routeSet.begin();
                                    itr != routeSet.end(); itr++)
    {
        if (!((*itr)->isChanged(RouteChangeBitMask(T_PROXY_BIT))))
        {
            continue;
        }
        withdrawn = false;
        if ((*itr)->isWithdrawn())
        {
            withdrawn = true;
        }
        if (initial)
        {
            if (withdrawn) continue;
        }
        if (((*itr)->getTransactionId() == prevTranId) && (msgToSend))
        {
            TRIPRoutes tRoute((*itr)->getAddrFamily(), (*itr)->getAppProtocol(),                               TRIPData((*itr)->getPrefix()));
            if (withdrawn)
            {
                *wRoute << tRoute;
            }
            else
            {
                *rRoute << tRoute;
            }
        }
        else
        {
            prevTranId = (*itr)->getTransactionId();
            if (msgToSend)
            {
                uMsg = new TRIPUpdateMsg();
                uMsg->addWithdrawnRoute(wRoute);
                uMsg->addReachableRoute(rRoute);
                uMsg->addAttribute(nextHop);
                uMsg->addAttribute(aPath);
                uMsg->addAttribute(rPath);
                retList->push_back(uMsg);
            }

            wRoute = new TRIPWithdrawnRoute();
            rRoute = new TRIPReachableRoute();
            nextHop = 0;
            nextHop.dynamicCast((*itr)->getAttribute(TRIP_NEXT_HOP_SERVER));
            if (!(nextHop.getPtr()))
            {
                cpLog(LOG_ALERT, "Bad NextHop Server");
                throw TRIPBadDataException("Bad NextHop Server", __FILE__,
                                                __LINE__, TRIP_ERR_UPDATE);
            }
            msgToSend = true;
            TRIPRoutes tRoute((*itr)->getAddrFamily(), (*itr)->getAppProtocol(),                               TRIPData((*itr)->getPrefix()));
            if (withdrawn)
            {
                *wRoute << tRoute;
            }
            else
            {
                *rRoute << tRoute;
            }
        }
    }
    if (nextHop.getPtr())
    {
        uMsg = new TRIPUpdateMsg();
        uMsg->addWithdrawnRoute(wRoute);
        uMsg->addReachableRoute(rRoute);
        uMsg->addAttribute(nextHop);
        uMsg->addAttribute(aPath);
        uMsg->addAttribute(rPath);
        retList->push_back(uMsg);
    }
    cpLog(LOG_DEBUG, "getFloodMsgList - Number of flood msgs is (%d) ",
                          retList->size());
    return retList;
}

void
LSApiServer::removeConnection(Connection& conn)
{
    if (conn.isLive())
    {
        conn.close();
    }
    for (ConnectionListMap:: const_iterator itr = itsConnListMap.begin();
                                            itr != itsConnListMap.end(); itr++)
    {
        Sptr <ConnectionList> connList = (*itr).second;
        ConnectionList::iterator where = find(connList->begin(),
                                            connList->end(), conn);
        if (where != connList->end())
        {
            connList->erase(where);
        }
    }

    for(ConnectionList::iterator itr2 = _connectionList.begin();
                                 itr2 != _connectionList.end(); itr2++)
    {
        if((*itr2) == conn)
        {
            _connectionList.erase(itr2);
            break;
        }
    }
}
